# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
"""
LSA backend python module
Host common type and helpers

"""

import threading
import logging
import taskmgr

__all__ = ['BackupTask', 'RestoreTask', 'ExtTimer']


class BackupTask(taskmgr.Task):
    """
    Task used to perform backup
    """
    def __init__(self, share_name):
        taskmgr.Task.__init__(self, share_name, 'backup')

    def begin(self):
        taskmgr.Task.begin(self)
        self._add_state(taskmgr.TaskStateChange(taskmgr.TaskState.STARTED))

    def end(self, status):
        taskmgr.Task.end(self, status)
        self._add_state(taskmgr.TaskStateChange(taskmgr.TaskState.STOPPED))


class RestoreTask(taskmgr.Task):
    """
    Task used to perform backup restoration
    """
    def __init__(self, share_name):
        taskmgr.Task.__init__(self, share_name, 'restore')

    def begin(self):
        taskmgr.Task.begin(self)
        self._add_state(taskmgr.TaskStateChange(taskmgr.TaskState.STARTED))

    def end(self, status):
        taskmgr.Task.end(self, status)
        self._add_state(taskmgr.TaskStateChange(taskmgr.TaskState.STOPPED))


class ExtTimer(threading.Thread):
    """
    Timer class which call a method repeatedly
    Compared to existing python timer, this one execute
    the job in infinite loop
    """

    __logger = logging.getLogger('lsa.backend.ExtTimer')

    def __init__(self, interval, method):
        """
        Initialize a new timer
        param
        interval : execution inteval  in seconds as float
        method: a callable to be executed
        """
        threading.Thread.__init__(self)
        if not callable(method):
            raise TypeError('not a callable')
        self.method = method
        if not type(interval) == float or interval == 0:
            raise TypeError('invalid interval')
        self.interval = interval

        # private : use to stop the timer
        self.__evt = threading.Event()

    def run(self):
        while True:
            if self.__evt.wait(self.interval):
                # we are out not because of timeout
                # i.e somebody ask us to stop
                break
            else:
                try:
                    self.method()
                except Exception, e:
                    ExtTimer.__logger.warning('error while running  timer: %s' % str(e))

    def stop(self):
        self.__evt.set()


class ExecuteAfterThread(threading.Thread):
    """
    Provides a way to execute a function (callback) when a thread terminates
    """
    __logger = logging.getLogger('lsa.backend.ExecuteAfterThread')

    def __init__(self, thread, callback, args):
        """
            thread: thread object
            callback: function to be executed when thread terminates
            args: tuple representing the arguments of callback
        """
        threading.Thread.__init__(self, target=self.__callback)
        self.callback = callback
        self.args = args
        self.thread = thread

    def __callback(self):
        ExecuteAfterThread.__logger.debug('ExecuteAfterThread started')
        self.thread.join()
        self.callback(*self.args)
        ExecuteAfterThread.__logger.debug('ExecuteAfterThread completed')


if __name__ == '__main__':
    import backend
    import threading
    # ExecuteAfterThread example

    def test_callback(s):
        print s

    def printer():
        import time
        for _ in range(3):
            time.sleep(1.0)
            print "thread"

    thread = threading.Thread(target=printer)
    a = backend.ExecuteAfterThread(thread, test_callback, ('blah',))
    thread.start()
    a.start()
