#__copyright__ = "Copyright 2018, Oracle.  All rights reserved."
#__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."

'''
This module contains the active directory related functions. Also manages /etc/hosts and /etc/resolv file
'''

import shutil
from command_util import execute_command
from common import CMD_CHATTR, ETC_HOSTS, ETC_RESOLVE_CONF, CMD_REALM, SMB_CONF, SMB_AD_CREDENTIALS_CONF
from log_util import get_logger, set_log_config
from utility import get_hostname_short, get_host_ip_address
from validation import is_valid_ipv4_address
import tempfile
import os
from kv_store import KeyValueStore

LOG = get_logger('lsa.ad_util')


class ADCredentialsKey(object):
    '''
    Constants for AD credentials keys
    '''
    # Key for AD server IP address
    ADS_IP = 'ads_ip'
    # Key for AD fqdn
    ADS_FQDN = 'ads_fqdn'
    # Key for AD user(delegated permission)
    ADS_USER = 'ads_user'

    @staticmethod
    def get_all_keys():
        return [ADCredentialsKey.ADS_IP, ADCredentialsKey.ADS_FQDN, ADCredentialsKey.ADS_USER]


def store_ad_credentials(fqdn, ip, user):
    '''
    Add or modify the AD credentials in the config file
    parameter:
        fqdn: Fully qualified domain name as string
        ip: IP address of the AD server as string
        user: user name of the AD server as string
        psd: password of the user as string
    raise: StandardError if fails to add
    '''
    items = {ADCredentialsKey.ADS_IP: ip,
             ADCredentialsKey.ADS_FQDN: fqdn,
             ADCredentialsKey.ADS_USER: user}
    KeyValueStore(SMB_AD_CREDENTIALS_CONF).add_keys(items)


def get_ad_credentials():
    '''
    Get AD credentials as dictionary items
    format:{'ads_fqdn': 'ads-srv.somedomain.com', 'ads_ip': '100.100.100.100', 'ads_user': 'testuser', 'ads_pass': 'xxxxx'}
    raise: StandardError if fails
    '''
    return KeyValueStore(SMB_AD_CREDENTIALS_CONF).get_all()


def remove_ad_credentials():
    '''
    Remove the AD credentials from the config file. Typically this function
    is called upon leaving the active domain.
    raise: StandardError if fails
    '''
    KeyValueStore(SMB_AD_CREDENTIALS_CONF).remove_keys(ADCredentialsKey.get_all_keys())


def add_dns_resolver(ip):
    '''
    Add a name server entry to the  /etc/resolv.conf file
    Entry which is a line in the file, format 'nameserver 192.168.1.1'
    if the entry is already there do nothing other wise append 'nameserver IP' line to the end of the file.
    Comment out all the entried as we only want to keep the AD server entry
    param:
        ip: Ip address of the host or server
    raise StandardError if failed to update the file
    '''
    temp_file = '/tmp/resolv.conf'
    main_file = ETC_RESOLVE_CONF
    bak_resolv = main_file + '.bak'
    try:
        fp = open(main_file, 'r')
        lines = fp.readlines()
    except (IOError, OSError) as e:
        LOG.error('Failed to read the file %s err: %s' % (main_file, str(e)))
        raise StandardError('Failed to read the file %s, err: %s' % (main_file, str(e)))
    finally:
        if fp:
            fp.close()
    updated_lines = []
    for line in lines:
        if line == '\n' or line is None:
            continue
        if line.startswith('#'):
            updated_lines.append(line)
            continue
        # if ip address is already in the file we do not do anything. anyway wee add the ip to the end and comment out anything other
        # than that. Once the LSA server is added to the domain or failed to add we restore the resolv file always.
        if ip in line:
            parts = line.split()
            if parts[0] == 'nameserver' and parts[1] == ip:
                continue
        updated_lines.append('#' + line)
    try:
        fp = tempfile.NamedTemporaryFile(delete=False)
        temp_file = fp.name
        for line in updated_lines:
            fp.write(line)
        fp.write('nameserver ' + ip + '\n')
        fp.flush()
    except (IOError, OSError) as e:
        LOG.error('Failed to write to file %s err: %s' % (temp_file, str(e)))
        raise StandardError('Failed to write to file %s, err: %s' % (temp_file, str(e)))
    finally:
        if fp:
            fp.close()
    shutil.copyfile(main_file, bak_resolv)
    shutil.move(temp_file, main_file)


def hosts_conf_add_host(fqdn, ip):
    '''
    Update the /etc/hosts file  with the specified FQDN and IP address
    This function will add an entry to the file if already not there
    If the fqdn is already there and the IP address is the different than the previous one
    update the entry with new value.
    param:
        fwdn: fully qualified domain name of the host
        ip: ip address of the host
    raise StandardError of failed to update the file
    '''
    if fqdn.count('.') < 2:
        raise StandardError('FQDN is not in valid format')
    if is_valid_ipv4_address(ip) is False:
        raise StandardError('IP address is not in valid format')

    hostname, domain, ext = fqdn.split('.')[:3]
    main_file = ETC_HOSTS
    bak_hosts = main_file + '.bak'
    temp_file = '/tmp/hosts'
    try:
        fp = open(main_file)
        lines = fp.readlines()
    except (IOError, OSError) as e:
        LOG.error('Failed to read the /etc/hosts file, err: %s' % str(e))
        raise StandardError('Failed to read the /etc/hosts file, err: %s' % str(e))
    finally:
        if fp:
            fp.close()
    updated_lines = []
    for line in lines:
        line = line.strip()
        if line == '\n' or line is None:
            continue
        parts = line.split()
        if line.startswith('#'):
            updated_lines.append(line)
            continue
        if len(parts) != 3:
            continue
        if parts[0] == ip and parts[1] == fqdn and parts[2] == hostname:
            return
        if parts[0] != ip and (parts[1] == fqdn or parts[1] == hostname or parts[2] == fqdn or parts[1] == hostname):
            updated_lines.append('#' + line)
            continue
        if parts[0] == ip and (parts[1] != fqdn or parts[2] != hostname):
            updated_lines.append('#' + line)
            continue
        updated_lines.append(line)
    updated_lines.append(ip + ' ' + fqdn + ' ' + hostname)
    shutil.copyfile(main_file, bak_hosts)
    try:
        fp = tempfile.NamedTemporaryFile(delete=False)
        temp_file = fp.name
        for item in updated_lines:
            fp.write("%s\n" % item)
        fp.flush()
    except (IOError, OSError) as e:
        LOG.error('Failed to write to file %s' % temp_file)
        raise StandardError('Failed to write to file %s' % temp_file)
    finally:
        if fp:
            fp.close()
    shutil.move(temp_file, main_file)


def remove_immutable_etc_resolv():
    '''
    Remove the read only restriction from the /etc/resolv file. i.e make the file editable
    raise StandardError if failed to modify the permission of the file
    '''
    cmd = [CMD_CHATTR, '-i', ETC_RESOLVE_CONF]
    out, err, code = execute_command(cmd)
    if code != 0:
        LOG.exception('Failed to change permission of  %s err: %s' % (ETC_RESOLVE_CONF, str(err)))
        raise StandardError('Failed to change permission of  %s err: %s' % (ETC_RESOLVE_CONF, str(err)))


def restore_conf_file(orig_file):
    '''
    Restore a config file from its backup if it does exist
    param:
        orig_file: original config file as string
    output: nothing on success and also on failure as it does not do anything if the file does not exist
    '''
    bak_conf = orig_file + '.bak'
    if os.path.exists(bak_conf):
        shutil.copyfile(bak_conf, orig_file)


def add_this_host_to_active_domain(ip, fqdn, user, password):
    '''
    Add this host to the active domain
    param:
        ip: IP address of the  active directory server
        fqdn: fully qualified domain name of the active directory server
        user: user name of the active directory delegated user
        password: password of the user
    raise StandardError if failed to join the  active directory
    '''
    if fqdn.count('.') < 2:
        raise StandardError('FQDN is not in valid format')
    if is_valid_ipv4_address(ip) is False:
        raise StandardError('IP address is not in valid format')
    if user is None or password is None:
        raise StandardError('User/Password must not be empty')

    # split fqdn and retrieve the domain  name with extension
    domain = fqdn.split('.', 1)[1]
    # local host (lsa server) fqdn
    local_fqdn = get_hostname_short() + '.' + domain
    # local host IP
    host_ip = get_host_ip_address()
    remove_immutable_etc_resolv()
    # Add AD server entry to /etc/hosts file
    hosts_conf_add_host(fqdn, ip)
    # add local host entry to /etc/hosts file
    hosts_conf_add_host(local_fqdn, host_ip)
    # update /etc/reolv.conf file to add name server entry with AD IP address
    add_dns_resolver(ip)
    cmd = [CMD_REALM, 'join', '-v', '--user', user, '--client-software=winbind', fqdn]
    try:
        out, err, code = execute_command(cmd, input=password)
        if code != 0:
            restore_conf_file(ETC_RESOLVE_CONF)
            restore_conf_file(SMB_CONF)
            LOG.error('Failed to add this server to the domain(%s), please check the credentials: %s' % (fqdn, err))
            raise StandardError('Failed to add this server to the domain(%s), please check the credentials: %s' % (fqdn, err))
    except Exception as e:
        restore_conf_file(ETC_RESOLVE_CONF)
        restore_conf_file(SMB_CONF)
        LOG.error('Failed to add this server to the domain(%s), please check the credentials: %s' % (fqdn, str(e)))
        raise StandardError('Failed to add this server to the domain(%s), please check the credentials: %s' % (fqdn, str(e)))

    restore_conf_file(ETC_RESOLVE_CONF)
    store_ad_credentials(fqdn, ip, user)


def try_leave_domain(fqdn):
    '''
    Try to leave active directory domain from this host
    param:
        fqdn: Fully qualified domain name of the AD server
    raise StandardError if failed to leave the  active directory domain
    '''
    if is_host_ad_member(fqdn) is True:
        leave_domain(fqdn)


def is_host_ad_member(fqdn):
    '''
    Check if this hist is already joined in the active directory
    param:
        fqdn: fully qualified domain name of the active directory server
    output: True if  already joined otherwise false
    '''
    if fqdn.count('.') < 2:
        raise StandardError('FQDN is not in valid format')
    domain = fqdn.split('.', 1)[1]
    cmd = [CMD_REALM, 'list']
    filtered = 'domain-name: ' + domain
    out, error, code = execute_command(cmd)
    if code != 0:
        return False
    if out is None:
        return False
    if filtered in out:
        return True
    return False


def leave_domain(fqdn):
    '''
    Remove the host from the domain
    param:
        fqdn: fully qualified domain name of the active directory server
    raise StandardError if failed to leave the  active directory domain
    '''
    if fqdn.count('.') < 2:
        raise StandardError('FQDN is not in valid format')
    domain = fqdn.split('.', 1)[1]
    cmd = [CMD_REALM, 'leave', domain.upper()]
    out, err, code = execute_command(cmd)
    if code != 0:
        LOG.exception('Failed to leave domain %s err: %s' % (domain, str(err)))
        raise StandardError('Failed to leave  %s err: %s' % (domain, str(err)))


def remove_comments_all(file_name):
    '''
    raise StandardError if failed
    '''
    bak_file = file_name + '.bak'
    try:
        fp = open(file_name)
        lines = fp.readlines()
    except (IOError, OSError) as e:
        LOG.error('Failed to read the %s file, err: %s' % (file_name, str(e)))
        raise StandardError('Failed to read the %s file, err: %s' % (file_name, str(e)))
    finally:
        if fp:
            fp.close()
    updated_lines = []
    for line in lines:
        if line == '\n':
            continue
        line = line.strip()
        line = line.strip('\n')
        if line.startswith('#') and 'GENERATED BY' not in line.upper():
            updated_lines.append(line[1:])
        else:
            updated_lines.append(line)
    try:
        fp = tempfile.NamedTemporaryFile(delete=False)
        temp_file = fp.name
        for item in updated_lines:
            fp.write("%s\n" % item)
        fp.flush()
    except (IOError, OSError) as e:
        LOG.error('Failed to write to file %s' % temp_file)
        raise StandardError('Failed to write to file %s' % temp_file)
    finally:
        if fp:
            fp.close()
    shutil.copyfile(file_name, bak_file)
    shutil.move(temp_file, file_name)


def remove_entry_from_resolv(ip):
    '''
    Remove an entry(line) from the file if amtches the IP.
    raise StandardError if failed
    '''
    if ip is None:
        return
    temp_file = '/tmp/resolv.conf'
    main_file = ETC_RESOLVE_CONF
    bak_resolv = main_file + '.bak'
    try:
        fp = open(main_file, 'r')
        lines = fp.readlines()
    except (IOError, OSError) as e:
        LOG.error('Failed to read the file %s err: %s' % (main_file, str(e)))
        raise StandardError('Failed to read the file %s, err: %s' % (main_file, str(e)))
    finally:
        if fp:
            fp.close()
    updated_lines = []
    for line in lines:
        if line == '\n' or line is None:
            continue
        line = line.strip()
        line = line.strip('\n')
        if line.startswith('#'):
            updated_lines.append(line)
            continue
        # if IP address is already listed we are done
        if ip in line:
            parts = line.split()
            if parts[0] == 'nameserver' and parts[1] == ip:
                continue
        updated_lines.append(line)
    print updated_lines
    try:
        #fp = open(temp_file, 'w')
        fp = tempfile.NamedTemporaryFile(delete=False)
        temp_file = fp.name
        for line in updated_lines:
            fp.write('%s\n' % line)
        fp.flush()
    except (IOError, OSError) as e:
        LOG.error('Failed to write to file %s err: %s' % (temp_file, str(e)))
        raise StandardError('Failed to write to file %s, err: %s' % (temp_file, str(e)))
    finally:
        if fp:
            fp.close()
    shutil.copyfile(main_file, bak_resolv)
    shutil.move(temp_file, main_file)


def remove_entry_from_hosts(fqdn):
    '''
    Remove an entry from /etc/hosts file with matching fqdn
    raise StandardError if failed
    '''
    hostname, domain, ext = fqdn.split('.')[:3]
    main_file = ETC_HOSTS
    bak_hosts = main_file + '.bak'
    temp_file = '/tmp/hosts'
    LOG.error('Removing #%s# from hosts file' % fqdn)
    try:
        fp = open(main_file)
        lines = fp.readlines()
    except (IOError, OSError) as e:
        LOG.error('Failed to read the /etc/hosts file, err: %s' % str(e))
        raise StandardError('Failed to read the /etc/hosts file, err: %s' % str(e))
    finally:
        if fp:
            fp.close()
    updated_lines = []
    #print lines
    for line in lines:
        line = line.strip()
        line = line.strip('\n')
        if line == '\n' or line is None:
            continue
        parts = line.split()
        if line.startswith('#'):
            updated_lines.append(line)
            continue
        if len(parts) != 3:
            updated_lines.append(line)
            continue
        if parts[1] == fqdn and parts[2] == hostname:
            continue
        updated_lines.append(line)
    shutil.copyfile(main_file, bak_hosts)
    print updated_lines
    try:
        fp = tempfile.NamedTemporaryFile(delete=False)
        temp_file = fp.name
        for item in updated_lines:
            fp.write('%s\n' % item)
        fp.flush()
    except (IOError, OSError) as e:
        LOG.error('Failed to write to file %s' % temp_file)
        raise StandardError('Failed to write to file %s' % temp_file)
    finally:
        if fp:
            fp.close()
    shutil.move(temp_file, main_file)


def get_ip_address_from_fqdn(fqdn):
    '''
    Get the IP address of a host by FQDN
    output: IP address of the  host, None if entry not found
    raise StandardError if failed
    '''
    hostname, domain, ext = fqdn.split('.')[:3]
    main_file = ETC_HOSTS
    try:
        fp = open(main_file)
        lines = fp.readlines()
    except (IOError, OSError) as e:
        LOG.error('Failed to read the /etc/hosts file, err: %s' % str(e))
        raise StandardError('Failed to read the /etc/hosts file, err: %s' % str(e))
    finally:
        if fp:
            fp.close()
    #print lines
    for line in lines:
        if line.startswith('#'):
            continue
        parts = line.split()
        if len(parts) < 3:
            continue
        if parts[1] == fqdn and parts[2] == hostname:
            return parts[0]
    return None


def restore_hosts_resolv(fqdn, realm):
    '''
    Restore the /etc/hosts and /etc/resolv file back to previous state after leaving the domain.
    We uncomment the entries and remove the hosts from /etc/hosts file that end with ad domain name.
    param:
    raise StandardError if failed
    '''
    domain = fqdn.split('.', 1)[1]
    if domain.upper() != realm:
        raise StandardError('realm and upper cased domain are not same.')
    # local host (lsa server) fqdn
    local_fqdn = get_hostname_short() + '.' + domain
    remove_comments_all(ETC_HOSTS)
    remove_comments_all(ETC_RESOLVE_CONF)
    ip = get_ip_address_from_fqdn(fqdn)
    remove_entry_from_hosts(fqdn)
    remove_entry_from_hosts(local_fqdn)
    remove_entry_from_resolv(ip)


if __name__ == '__main__':
    set_log_config()
