# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
from common import (EXPORTFS, EXPORTS_FILE, BACKUP_EXPORT,
                    TEMP_EXPORT_BACK, SMB_CONF,
                    SYSCTL, SAMBA_SERVICES, CHCON,
                    SEMANAGE, RESTORECON, SMB_EXPORT_CONF, CMD_SMBCONTROL)
from command_util import execute_command, command_output_lines
import shutil
import os
import StringIO
import socket
from ConfigParser import ConfigParser

from log_util import get_logger, set_log_config
from validation import is_valid_netif
from nfs_util import get_nfs_entries_from_line
from samba_util import SambaExportEntry, SambaConfKey
from utility import get_full_mount_path, CustomConfigParser
from service_util import SystemdService
from utility import parse_boolean
from ad_util import (add_this_host_to_active_domain, restore_hosts_resolv,
                     try_leave_domain, is_host_ad_member, get_ad_credentials,
                     ADCredentialsKey, remove_ad_credentials)

LOG = get_logger('lsa.config_manager')


class NfsConfigManager(object):
    '''
    This module manages NFS config file and etab exports.
    Using this module you can create/modify/delete NFS exports.
    '''
    '''
    Usage:

    create an NFS export:
    share = ShareEntry(name)
    export = NfsExportEntry(hosts, path, options)
    NfsConfigManager().add_export(share, export)
    modify an nfs export:
    NfsConfigManager().update_export(share, export)
    delete an nfs export:
    NfsConfigManager().remove_export(share, export)
    delete nfs export by share name:
    NfsConfigManager().unexport_share(share)

    '''
    def __init__(self, export_file=EXPORTS_FILE):
        """
        NFS export file: default /.lsa/etc/nfs/exports.d/lsa.exports
        """
        self.export_file = export_file

    def _get_filtered_lines(self):
        '''
        Get the lines from export file. We skip comments and new lines.
        catch: IOError, OSError
        Raise: StandardError
        output: list of lines containing nfs exports
        '''
        org_lines = []
        filtered_lines = []
        bfo = None
        try:
            bfo = open(self.export_file)
            org_lines = bfo.readlines()
        except (IOError, OSError) as e:
            raise StandardError('Failed to read the configuration file: %s' % str(e))
        finally:
            if bfo:
                bfo.close()

        for line in org_lines:
            line = line.strip()
            # skip comment or new line
            if line.startswith('#') or line == '\n':
                continue
            filtered_lines.append(line)
        return filtered_lines

    def _get_config_line(self, share):
        '''
        Get a particular line from export file matching the share name
            parameter: ShareEntry object
        raise:
            StandardError if failed to access the exports file
        output: a config line as string
                format: /shares/test1 x.x.x.x(options) y.y.y.y(options) ... ...
                if the share is not exported this function will return none
                and the caller is responsible for handling the case
        '''
        for line in self._get_filtered_lines():
            path = line.split()[0]
            if path == share.get_path():
                return line
        return None

    def _flush_config_lines(self, lines):
        """
        Write all the export entries to export file from in-memory list.
        parameter:
            lines: lines of NFS config
        raises: OSError, IOError if failed to access the config file
                sshutil.Error if failed to copy the backup config file
        """
        LOG.debug("Writing NFS exports")
        bak_file = TEMP_EXPORT_BACK
        # create new file
        bfo = None
        try:
            bfo = open(bak_file, "w")
            for line in lines:
                bfo.write(line)
                bfo.write('\n')
        except (OSError, IOError) as e:
            LOG.error('Failed to write entries to file: \n' + str(e))
            raise Exception('Failed to write nfs exports: ' + str(e))
        finally:
            if bfo:
                bfo.close()
            if os.path.exists(BACKUP_EXPORT):
                os.remove(BACKUP_EXPORT)
        # create a backup file just in-case
        LOG.debug('backup %s to %s' % (self.export_file, BACKUP_EXPORT))
        shutil.copyfile(self.export_file, BACKUP_EXPORT)
        LOG.debug('rename %s to %s' % (bak_file, self.export_file))
        shutil.move(bak_file, self.export_file)

    def _add_export_into_line(self, share, line, export):
        '''
        Add an nfs export to a config line and return the updated line
        The input parameter line is already a formated line, we just add new host(options) at the end
        parameter:
            share: ShareExport entry
            line: Current line as String
            export: NfsExportEntry object
        output: updated config line as string
        raise: StandardError if export already exists
        '''
        path = line.split()[0]
        if path != share.get_path():
            raise StandardError('Cannot add export for other share.')
        exports = get_nfs_entries_from_line(line)
        if export in exports:
            raise StandardError('Export entry already exists')
        return line + ' ' + repr(export)

    def _update_export_into_line(self, share, line, export):
        '''
        Update a particular export inside a export line
         The input parameter line is already a formated line, we just update host(options) inside the line
        parameter:
            share: ShareExport object
            line: export line(format supported in nfs export config file) i.e  '/access_point hosts1(options) host2(options) ... ...'
            export: NfsExportEntry object
        output: updated config line as string
        '''
        exports = get_nfs_entries_from_line(line)
        if export not in exports:
            raise StandardError('Nfs export entry not found for modification')
        for idx, obj in enumerate(exports):
            if export.get_id() == obj.get_id():
                exports[idx] = export
        return self._format_config_line(share, exports)

    def _remove_export_from_line(self, share, line, export):
        '''
        Remove an export from a config line
        parameter:
            share: ShareEntry object
            line: export line(format supported in nfs export config file) i.e  '/access_point hosts1(options) host2(options) ... ...'
            export: NfsExportEntry object
        '''
        exports = get_nfs_entries_from_line(line)
        if export not in exports:
            raise StandardError('Export not found in the config file for removing')
        exports.remove(export)
        return self._format_config_line(share, exports)

    def _format_config_line(self, share, export_entries):
        '''
        Build a config line supported in nfs export file from share object and list of export entries
        parameter:
            share: ShareEntry object
            export_entries: list of NfsExportEntry objects
        output: export line(format supported in nfs export config file) i.e  '/access_point hosts1(options) host2(options) ... ...'
        '''
        if not export_entries:
            return ''
        output = StringIO.StringIO()
        output.write(share.get_path())
        for entry in export_entries:
            output.write(' ')
            output.write(repr(entry))
        config_line = output.getvalue()
        output.close()
        return config_line

    def get_exports(self, share):
        '''
        Gets the list of NFS export for a share
        This method will read from configuration all the export for this
        particular share
        param:
          share: ShareEntry object
        returns:
          list of exports as List of NfsExportEntry
        raises:
          StandardError : issue accessing the configuration
        '''
        #  get the particular export line from config file
        line = self._get_config_line(share)
        if line is not None:
            return get_nfs_entries_from_line(line)
        return []

    def get_export(self, share, export_id):
        '''
        Gets the NfsExport for a share and export ID
        parameter:
            share: ShareEntry object
            export_id: Nfs export identifier (hosts)
        output: NfsExportEntry object, None if the export is not found
        StandardError : issue accessing the configuration
        '''
        exports = self.get_exports(share)
        for export in exports:
            if export.get_id() == export_id:
                return export

    def add_export(self, share, export):
        '''
        Adds an nfs export to configuration file
        parameter:
            share: ShareEntry object
            export: NfsExportEntry object
        raise: StandardError if failed to access the config file or export already exists
        '''
        self.add_exports(share, [export])

    def add_exports(self, share, exports):
        '''
        Adds a list of Nfs exports into the config file
        parameter:
            share: ShareEntry object
            exports: list of NfsExportEntry object
        raise: StandardError if failed to access the config file or export already exists
        '''
        lines = self._get_filtered_lines()
        line_exist = False
        for idx, line in enumerate(lines):
            path = line.split()[0]
            if path == share.get_path():
                updated_line = line
                for export in exports:
                    updated_line = self._add_export_into_line(share, updated_line, export)
                lines[idx] = updated_line
                line_exist = True
                break
        if line_exist is False:
            lines.append(self._format_config_line(share, exports))
        self._flush_config_lines(lines)

    def update_export(self, share, export):
        '''
        Updates an export from config file
        parameter:
            share: ShareEntry object
            export: NfsExportEntry object
        raise: StandardError if failed to access the config file or export not found in the config file
        '''
        lines = self._get_filtered_lines()
        for idx, line in enumerate(lines):
            path = line.split()[0]
            if path == share.get_path():
                lines[idx] = self._update_export_into_line(share, line, export)
                break
        self._flush_config_lines(lines)

    def remove_export(self, share, export):
        '''
        Remove an export from config file
        parameter:
            share: ShareEntry object
            export: NfsExportEntry object
        raise: StandardError if failed to access the config file or export not found in the config file
        '''
        lines = self._get_filtered_lines()
        for idx, line in enumerate(lines):
            path = line.split()[0]
            if path == share.get_path():
                updated_line = self._remove_export_from_line(share, line, export)
                # line might be empty
                if updated_line:
                    lines[idx] = updated_line
                else:
                    lines.pop(idx)
                break
        self._flush_config_lines(lines)

    def unexport_share(self, share):
        '''
        remove  all the nfs exports associated with the share
        parameter:
            share: ShareEntry object
        raise: StandardError if failed to access the config file
            Exception caught by upper layer(priv_exec)
        '''
        lines = self._get_filtered_lines()
        for idx, line in enumerate(lines):
            path = line.split()[0]
            if path == share.get_path():
                lines.pop(idx)
                break
        self._flush_config_lines(lines)


class NFSManager(object):
    '''
    This is an exportfs wrapper which utilizes exportfs command
    Usage:
        # add an export
        NFSManager.add_export(share, exp)
        # remove an export
        NFSManager.remove_export(share, exp)
    '''
    def remove_export_list(self, share, exports):
        """
        Unexport  multiple NFS exports.
        parameter:
            share: ShareEntry object
            entries: list NfsExportEntry object
        raises:
          StandardError: operation has failed to add export to etab using sxportfs command
        """

        for export in exports:
            if self._is_export_active(share, export):
                # avoid calling this when export not actually active
                # as it may raise an error
                self.remove_export(share, export)

    def _is_export_active(self, share, exp_entry):
        """
        check that an exportis actually active on the system
        According to the specified host,the export may be set in the conf
        but not actually exported (unresolvable host etc...)
        Checks that the export is actually exported
        parameter:
            share: ShareEntry object
            exp_entry: NfsExportEntry object
        Returns:
            True if exported, False otherwise
        """
        LOG.debug("_is_export_active: checking export active on %s %s" % (share.get_path(), exp_entry.get_hosts()))
        # exportfs -s output is like
        # <path> <hosts>(<options>)
        lines = command_output_lines([EXPORTFS, '-s'])
        for line in lines:
            if not line:
                continue
            splitted_line = line.split()
            # no sanity: we trust exportfs(8) output
            if splitted_line[0] == share.get_path():
                # now check that the same hosts
                _host = splitted_line[1].split('(')[0]
                # host is '*' or host or IP.
                # string comparison is not enough as we cannot trust how exportfs(8) has resolved it
                # only hostname, fqn, ip address ...
                # i.e if string match we return True otherwise we have to check further
                if _host == exp_entry.get_hosts():
                    LOG.debug("_is_export_active: export found")
                    return True
                # perform further check for "host" equality: if they resolve as same IP they are the same
                if socket.gethostbyname(exp_entry.get_hosts()) == socket.gethostbyname(_host):
                    LOG.debug("_is_export_active: export found")
                    return True

        LOG.debug("_is_export_active: export not found")
        return False

    def modify_export(self, share, exp_entry):
        """
        Modifies an existing NFS export
        parameter:
            share: ShareEntry object
            exp_entry: NfsExportEntry object
        raises:
          StandardError: operation has failed.
        """
        self.remove_export(share, exp_entry)
        self.add_export(share, exp_entry)

    def remove_export(self, share, exp_entry):
        """
        Unexport a share using exportfs command
        parameter:
            share: ShareEntry object
            exp_entry: NfsExportEntry object
        raises:
          StandardError: operation has failed to add export to etab
        """
        param = exp_entry.get_hosts() + ":" + share.get_path()
        LOG.debug("Removing entry from etab: " + param)
        out, err, code = execute_command([EXPORTFS, '-u', param])
        if code != 0:
            LOG.exception("Failed to remove entry from etab")
            raise StandardError("Failed to remove entry from etab: %s" % str(err))

        LOG.debug("Removing entry from etab done: " + param)

    def add_export(self, share, exp_entry):
        """
        Add an export entry to etab
        parameter:
            share: ShareEntry object
            exp_entry: NfsExportEntry object
        raise:
          StandardError: if exportfs failed and return code other than zero
        """
        param = exp_entry.get_hosts() + ":" + exp_entry.get_path()
        options = exp_entry.get_options_as_string()
        LOG.debug("Adding entry to etab:" + param)
        out, err, code = execute_command([EXPORTFS, param, '-o', options])
        if code != 0:
            LOG.exception("Failed to add entry to  etab: %s" % str(err))
            raise StandardError("Failed to add entry to etab: %s" % str(err))

        LOG.debug("Adding entry to etab done:" + param)


class SambaManager(object):
    '''
    Manage runtime configuration of samba exports
    Usage:
    code = SambaManager().remove_exports()
    # add selinux rule
    SambaManager.add_export(share)
    '''
    def remove_export_list(self, share, exports):
        """
        Flush samba runtime configuration from removed exports
        parameter:
            share: ShareEntry object
            exports: list of SambaExportEntry object
        Returns:  code ad integer. 0 means success and other than 0 means failure.
        """
        result = 0
        for export in exports:
            result = result + self.remove_export(share, export)
        return result

    def modify_export(self, share, export):
        """
        Modifies an existing SMB export.
        We rely on the configuration to be updated and ask daemons to reload        parameter:
            share: ShareEntry object
            export: SambaExportEntry object
        Returns:  0  on  success and other than 0 on failure.
        """
        LOG.debug('SambaManager.modify_export called for %s %s' % (str(share), str(export)))
        out, err, code = execute_command([CMD_SMBCONTROL, 'smbd', 'close-share', export.get_resource()])
        if code != 0:
            LOG.warning("modify_export: close of share failed: %s" % str(err))
        else:
            self._reload()
            LOG.debug('SambaManager.modify_export smbcontrol called ok')
        return code

    def remove_export(self, share, export):
        """
        Flush samba runtime configuration from removed exports
        parameter:
            share: ShareEntry object
            export: SambaExportEntry object
        Returns:  code ad integer. 0 means success and other than 0 means failure.
        """
        LOG.debug('SambaManager.remove_export called for %s %s' % (str(share), str(export)))
        out, err, code = execute_command([CMD_SMBCONTROL, 'smbd', 'close-share', export.get_resource()])
        if code != 0:
            LOG.warning("close of share failed: %s" % str(err))
        else:
            self._reload()
            LOG.debug('SambaManager.remove_export smbcontrol called ok')
        return code

    def _reload(self):
        """
        Ask Samba daemons to gracefully reload the configuration
        raises:
           StandardError: if reload has failed
        """
        out, err, code = execute_command([CMD_SMBCONTROL, 'all', 'reload-config'])
        if code != 0:
            LOG.exception('Cannot reload samba config err: %s' % str(err))
            raise StandardError('Samba config reload failed (%d,%s)' % (code, str(err)))
        return code

    def reload_configuration(self):
        """
        Ask Samba daemons to gracefully reload the configuration
        Usually called after modification done through SambaConfigManager
        raises:
           StandardError: if reload has failed
        """
        LOG.debug('reload_configuration called')
        return self._reload()

    def restart_samba_services(self):
        """
        Restart samba related services
        Returns:  code an integer. 0 means success and other than 0 means failure.
        """
        LOG.debug('restart_samba_services')

        cmd = [SYSCTL, SystemdService.RESTART] + SAMBA_SERVICES

        out, err, code = execute_command(cmd)

        LOG.debug('restart_samba_services (%d,%s,%s)' % (code, out, err))

        return code

    def add_export(self, share, export):
        '''
        Add required selinux rules during a samba export creation. Helpful link: https://selinuxproject.org/page/SambaRecipes
        We need to add three selinux rules.
            1. chcon -t samba_share_t /access_point
            2. semanage fcontext -a -t samba_share_t "/access_point(/.*)?"
            3. semanage fcontext -a -t oublic_content_rw_t "/access_point(/.*)?"
            4. restorecon -R -v /access_point
        Parameters:
            share: ShareEntry object
            export: SambaExportEntry object
        Output: Nothing.
        Raise StandardError if it fails to add a selinux rule.
        '''
        path = get_full_mount_path(share.get_name())
        cmd = [CHCON, '-t', 'samba_share_t', path]
        out, err, code = execute_command(cmd)
        if code != 0:
            LOG.exception('Selinux rule for samba, Chcon failed for access point %s, err: %s' % (path, str(err)))
            raise StandardError('Selinux rule for samba, Chcon failed for access point %s err: %s' % (path, str(err)))

        cmd = [SEMANAGE, 'fcontext', '-a', '-t', 'samba_share_t', path + '(/.*)?']
        out, err, code = execute_command(cmd)
        if code != 0:
            LOG.exception('Selinux rule for samba, Semanage failed for access point %s err: %s' % (path, str(err)))
            raise StandardError('Selinux rule for samba, Semanage failed for access point %s err: %s' % (path, str(err)))
        # turn on anonymous read/write
        cmd = [SEMANAGE, 'fcontext', '-a', '-t', 'public_content_rw_t', path + '(/.*)?']
        out, err, code = execute_command(cmd)
        if code != 0:
            LOG.exception('Selinux rule for samba, Semanage failed for access point %s err: %s' % (path, str(err)))
            raise StandardError('Selinux rule for samba, Semanage failed for access point %s err: %s' % (path, str(err)))

        cmd = [RESTORECON, '-R', '-v', path]
        out, err, code = execute_command(cmd)
        if code != 0:
            LOG.exception('Selinux rule for samba, Restorecon failed for access point %s err: %s' % (path, str(err)))
            raise StandardError('Selinux rule for samba, Restorecon failed for access point %s err: %s' % (path, str(err)))

        # now ask smb daemon to reload the config
        self._reload()


class SambaConfigManager(object):
    '''
    This class manages the config file modoficaions for add/update or deleting the exports:
    Usage:
        # add an export
        SambaConfigManager().add_export(share, export)
    '''
    # sections that we need to skip.
    skipped_sections = ['global', 'printers', 'homes']

    def __init__(self, conf_file=SMB_CONF):
        LOG.debug('new SambaConfigManager on %s' % conf_file)
        self.config_file = conf_file

    def _set_global_options(self, options, ad_enabled):
        """
        Update global configurable options in smb.conf
        param:
            options: dictionary contains key value pair of smb configuration
            ad_enabled: if active directory is enabled i.e smb access control is active directory enabled
        Supported options: 'workgroup', 'netbios name', 'interfaces', 'ad_enabled', 'fqdn', 'ip_address', 'user_name', 'user_password'
        If options['interfaces'] is not empty, then the 'bind interfaces only' option will be implicitly set as 'yes'
        example1:
            options['workgroup'] = 'workgroup_test1'
            options['netbios name'] = 'netbios_test1'
            options['interfaces'] =
                in this case, the 'bind interfaces only' option will be set as 'no' implicitly`
            options['ad_enabled'] = True/False ; whether want to enable Active directory based access control
            options['ip_address'] = x.x.x.x; IP address of the AD server
            options['fqdn'] = 'host.domain.com'; fully qualified domain name of the AD server
            options['user_name'] = 'user name'; delegated user name of the AD server
            options['user_password'] = 'user password'; password of the delegated user

        example2:
            options['workgroup'] = 'workgroup_test2'
            options['netbios name'] = 'netbios_test2'
            options['interfaces'] = 'lo, eno1'
                in this case, the 'bind interfaces only' option will be set as 'yes' implicitly
            options['ad_enabled'] = True/False; whether want to enable Active directory based access control
            options['ip_address'] = x.x.x.x; IP address of the AD server
            options['fqdn'] = 'host.domian.com'; fully qualified domain name of the AD server
            options['user_name'] = 'user name'; deleted user name of the AD server
            options['user_password'] = 'user password'; password of the delegated user
        """
        default_keys = ['security', 'netbios name', 'interfaces', 'workgroup', 'map to guest', 'include']
        if not os.path.exists(self.config_file):
            LOG.warning('No samba configuration found. samba stack probably not installed')
            # we could continue and create the configuration file but
            # there's no point if samba feature is not there.
            # if someone install samba we gonna the configuration anyway
            raise StandardError("missing Samba configuration")
        fo = None
        try:
            configParser = CustomConfigParser(allow_no_value=True)
            configParser.read_file(self.config_file)
            if ad_enabled is False:
                # try to restore the samba config file if we want to go back to default user based access control.
                security = configParser.get('global', 'security')
                if security == 'ads':
                    fqdn = configParser.get('global', 'password server')
                    realm = configParser.get('global', 'realm')
                    restore_hosts_resolv(fqdn, realm)
                    #remove ad related keys from samba config file
                    global_keys = configParser.options('global')
                    for key in global_keys:
                        if key not in default_keys:
                            configParser.remove_option('global', key)
                    try_leave_domain(fqdn)
                    remove_ad_credentials()
                # if we already joined to AD following options are already set by realm join command, so no need to update for AD
                configParser.set('global', 'workgroup', options['workgroup'])
                configParser.set('global', 'security', 'user')
            else:
                configParser.set('global', 'vfs objects', 'acl_xattr')
                configParser.set('global', 'map acl inherit', 'yes')
                configParser.set('global', 'store dos attributes', 'yes')

            configParser.set('global', 'netbios name', options['netbios name'])
            configParser.set('global', 'interfaces', options['interfaces'])
            configParser.set('global', 'map to guest', 'Bad User')
            configParser.set('global', 'include', SMB_EXPORT_CONF)
            if not options['interfaces']:
                configParser.set('global', 'bind interfaces only', 'no')
            else:
                configParser.set('global', 'bind interfaces only', 'yes')
            fo = open(self.config_file, 'w')
            configParser.write(fo)
        except(IOError, OSError) as e:
            raise StandardError("Failed to update samba global options %s " % str(e))
        finally:
            if fo:
                fo.close()

    def _validate_samba_options(self, options):
        """
        Check whether the input of the options are valid or not.
        So far, only check whether the specified 'interfaces' option is a valid netif
        raise ValueError if one option value is not valid
        raise KeyError is one option is missing
        """
        if 'workgroup' not in options:
            raise KeyError("missing workgroup option")
        if 'netbios name' not in options:
            raise KeyError("missing netbios name option")
        if 'interfaces' not in options:
            raise KeyError("missing interfaces option")
        # netbios name can not be empty
        if options['netbios name'] == '':
            raise KeyError('netbios name can not be empty')
        # Note: options['interfaces'] is empty means the user doesn't specify, this is fine
        # but, empty interface in a specified string, say 'lo,,eth1', is not valid
        if options['interfaces'] != '':
            for ifname in options['interfaces'].split(","):
                if not is_valid_netif(ifname.strip()):
                    raise ValueError("Invalid interface: " + ifname)

    def get_samba_options(self):
        """
        API to get samba global options.
        Read configurable options from global section in smb.conf file.
        Retrieved options:
            workgroup: The workgroup samba server will appear to be in when queried by clients
            netbios name: The share name clients can see
            interfaces: The network interfaces that Samba will use for browsing, name registration and other
                        netbios over tcp/ip traffic.
        example:
            options['workgroup'] = 'workgroup_test1'
            options['netbios name'] = 'netbios_test'
            options['interfaces'] = 'lo, eno1'
        """
        options = {}

        if not os.path.exists(self.config_file):
            LOG.warning('No samba configuration found. samba stack probably not installed')
            # we could continue and create the configuration file but
            # there's no point if samba feature is not there.
            # if someone install samba we gonna the configuration anyway
            raise StandardError("missing Samba configuration")

        try:
            configParser = CustomConfigParser(defaults={'workgroup': '', 'netbios name': '', 'interfaces': ''}, allow_no_value=True)
            configParser.read_file(self.config_file)
            options['workgroup'] = configParser.get('global', 'workgroup')
            options['netbios name'] = configParser.get('global', 'netbios name')
            options['interfaces'] = configParser.get('global', 'interfaces')
            options['ad_enabled'] = False
            security = configParser.get('global', 'security')
            if security == 'ads':
                options['ad_enabled'] = True
                ad_items = get_ad_credentials()
                if ad_items:
                    options['ip_address'] = ad_items.get(ADCredentialsKey.ADS_IP)
                    options['fqdn'] = ad_items.get(ADCredentialsKey.ADS_FQDN)
                    options['user_name'] = ad_items.get(ADCredentialsKey.ADS_USER)
        except (IOError, OSError) as e:
            raise StandardError("Failed to read samba conf file %s: %s" % (self.config_file, str(e)))
        LOG.debug('get_samba_options => %s' % str(options))
        return options

    def set_samba_options(self, options):
        '''
        API to set samba global options.
        The original conf file is backed up first before the options are updated.
        If update fails, the old ones are recovered
        raise StandardError if update fails
        raise ValueError if invalid options
        Raise shutil.Error if copy fails
        '''
        conf_bak = self.config_file + '.bak'
        self._validate_samba_options(options)
        ad_enabled = parse_boolean(options.get('ad_enabled'))
        # backup smb.conf first in case we need to recover later
        try:
            shutil.copy(self.config_file, conf_bak)
        except IOError, e:
            raise StandardError('cannot make a backup copy of samba configuration [%s]' % str(e))
        if ad_enabled:
            if is_host_ad_member(options.get('fqdn')):
                raise StandardError('This host is already a member of the AD. Please leave the domain first')
            # if active directory is enabled we need to  make the samba config file empty.
            # make the config file empty
            try:
                os.remove(SMB_CONF)
                open(SMB_CONF, 'w').close()
            except (IOError, OSError) as e:
                LOG.error('Failed to modify samba configuration file, err: %s' % str(e))
                raise StandardError('Failed to modify samba configuration file, err: %s' % str(e))
            try:
                add_this_host_to_active_domain(options.get('ip_address'), options.get('fqdn'), options.get('user_name'), options.get('user_password'))
            except Exception as e:
                raise StandardError('Failed to add this LSA server to AD %s, err %s' % (options.get('fqdn'), str(e)))
        self._set_global_options(options, ad_enabled)
        LOG.debug('Setting samba global options done')

    def _update_config_parser(self, share, exp, create_section=False):
        '''
        Update the config parser with export attributes and return the parser.
        parameter:
            share: ShareEntry object
            exp: SambaExportEntry object
            create_section: if a section creation is needed.
        raise: Standard error for duplicate section in case of creation.
        '''
        config_parser = CustomConfigParser(allow_no_value=True)
        config_parser.read_file(self.config_file)
        if create_section is True:
            try:
                config_parser.add_section(exp.resource)
            except ConfigParser.DuplicateSectionError as e:
                raise StandardError('Can not add duplicate section to the config file i.e resource/section already exists: %s' % str(e))
        path = get_full_mount_path(share.get_name())
        if exp.comment:
            config_parser.set(exp.resource, SambaConfKey.COMMENT, exp.comment)
        config_parser.set(exp.resource, SambaConfKey.GUEST_OK, str(exp.guest_ok.enabled()))
        config_parser.set(exp.resource, SambaConfKey.STRICT_SYNC, str(exp.strict_sync.enabled()))
        config_parser.set(exp.resource, SambaConfKey.SYNC_ALWAYS, str(exp.sync_always.enabled()))
        config_parser.set(exp.resource, SambaConfKey.BROWSABLE, str(exp.browsable.enabled()))
        config_parser.set(exp.resource, SambaConfKey.READ_ONLY, str(exp.read_only.enabled()))
        config_parser.set(exp.resource, SambaConfKey.PATH, path)
        config_parser.set(exp.resource, SambaConfKey.DIRECTORY_MASK, exp.directory_mask)
        config_parser.set(exp.resource, SambaConfKey.CREATE_MASK, exp.create_mask)
        #config_parser.set(exp.resource, SambaConfKey.FORCE_USER, exp.force_user)
        if exp.hosts_allow is not None:
            config_parser.set(exp.resource, SambaConfKey.HOSTS_ALLOW, exp.hosts_allow)
        if exp.hosts_deny is not None:
            config_parser.set(exp.resource, SambaConfKey.HOSTS_DENY, exp.hosts_deny)
        return config_parser

    def remove_export(self, exp):
        '''
        Delete a samba export from smb.conf file
        parameter:
            exp: SambaExportEntry
        raise: Standard error if failed to find/delete the export or failed to write the export to the config file.
        '''
        config_parser = CustomConfigParser(allow_no_value=True)
        config_parser.read_file(self.config_file)
        if config_parser.has_section(exp.get_resource()) is False:
            raise StandardError('Resource %s not found for deletion' % exp.get_resource())
        config_parser.remove_section(exp.get_resource())
        self._write_samba_export_to_conf(config_parser)

    def remove_exports_by_share(self, share):
        '''
        Delete all samba exports matching with share name
            parameter:
                share_name: name of the share
        raise: KeyEerror if path does not exist in the options(unusual case)
               StandardError if failed to write the config file
               Exception caught by upper layer(priv_exec)
         '''
        config_parser = CustomConfigParser(allow_no_value=True)
        config_parser.read_file(self.config_file)
        resources = []
        for section in config_parser.sections():
            # skip global, printers and home
            if section in SambaConfigManager.skipped_sections:
                continue
            path = config_parser.get(section, 'path')
            if not path:
                raise KeyError('Unusual case, path must be an attribute in samba export')
            name = os.path.basename(os.path.normpath(path))
            if name == share.get_name():
                resources.append(section)
        if resources:
            for resource in resources:
                config_parser.remove_section(resource)
            self._write_samba_export_to_conf(config_parser)

    def _write_samba_export_to_conf(self, config_parser):
        '''
        Write the in memory config to file. Here we try to update the file first by creating a backup,
        Caller should check by restarting the services if the service status is running we keep the updated config file,
        otherwise we restore the previous consistent config file from backup. The reason behind the failure might be
        invalid modification of the config parameters.
        parameter:
            config_parser:  CustomConfigParser object
        raise: StandardError if failed to write the config file
        '''
        conf_bak = self.config_file + '.bak'
        try:
            shutil.copy(self.config_file, conf_bak)
        except IOError, e:
            raise StandardError('cannot make a backup copy of samba configuration [%s]' % str(e))
        try:
            fo = open(self.config_file, 'w')
            config_parser.write(fo)
        except (IOError, OSError) as e:
            LOG.exception('Failed to write samba config file: %s' % str(e))
            raise StandardError('Failed to write samba config file: %s' % str(e))
        finally:
            fo.close()

    def revert_smb_conf_modifications(self):
        '''
        Restore the samba config file from previously created backup
        raise: OSError if failed to rename the backup file
        '''
        conf_bak = self.config_file + '.bak'
        os.rename(conf_bak, self.config_file)

    def get_export(self, share, export_id):
        '''
        Gets a particular samba export by share and export_id
        parameter:
            share: ShareEntry object
            export_id: export identifier
        return: SmabaExport entry object, None if not found
        '''
        LOG.debug('get_export share name [%s], export [%s]' %
                  (share.get_name(), str(export_id)))
        exports = self.get_exports(share)
        for exp in exports:
            if exp.get_id() == export_id:
                return exp
        LOG.debug('get_export share name [%s], export [%s]: not found ' %
                  (share.get_name(), str(export_id)))
        return None

    def get_exports(self, share):
        '''
        Retrieve all the samba exports from config file
        share: a ShareEntry instance
        output: list of SambaExportEntry objects, might me empty
        raise: Exception raised by upper layer(priv_exec) if fails.
        '''
        LOG.debug('get_exports called, config file=%s', self.config_file)

        if not os.path.exists(self.config_file):
            LOG.warning('missing samba configuration file [%s]' % self.config_file)
            return []

        config_parser = CustomConfigParser(allow_no_value=True)
        config_parser.read_file(self.config_file)
        exports = []
        for section in config_parser.sections():
            '''
            We don't care about global, homes, printer section since that section is not an export entry
            '''
            if section in SambaConfigManager.skipped_sections:
                continue
            options = dict(config_parser.items(section))
            path = options.get('path')
            name = os.path.basename(os.path.normpath(path))
            if name != share.get_name():
                continue
            exp = SambaExportEntry(section, **options)
            exports.append(exp)
        LOG.debug('get_exports => %s', str(exports))
        return exports

    def add_export(self, share, exp):
        '''
        Create a samba export on the system.
        parameter:
            share: ShareEntry object
            exp: SambaExportEntry object
        raise: StandardError of failed to access the config file or resource already exists
               Exception raised by upper layer(priv_exec)
        '''

        LOG.debug('new samba export created [%s]' % str(exp))
        config_parser = self._update_config_parser(share, exp, True)
        self._write_samba_export_to_conf(config_parser)

    def modify_export(self, share, exp):
        '''
        Modify a samba export on the system.
        parameter:
             share: ShareEntry object
             exp: SambaExportEntry object
        raise: StandardError of failed to access the config file, resource not found.
               Exception caught by upper layer(priv_exec)
        '''

        LOG.debug('export to modify [%s]' % str(exp))
        config_parser = self._update_config_parser(share, exp)
        self._write_samba_export_to_conf(config_parser)


if __name__ == '__main__':
    set_log_config()
