#!/usr/bin/bash
# Copyright (c) 2017, 2018, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.

: '
    This script adjusts the NFS settings and manages services
'

CUR_DIR="$(dirname "$0")"
source $CUR_DIR/globals.sh
source $CUR_DIR/common.sh

: '
    restart NFS services
'
function restart_nfs_services() {
    $SYSCTL restart nfs-server
    check_exit "nfs-server start"
    return $?
};

: '
    Modify some of the NFS configurations for better performance
'
function adjust_nfs_configs() {

    # NFS setting requirements aligned with EBS on Oracle cloud
    #  CR 25794589
    # TODO : make this available through UI for those
    #        who do not have the same requirements.

    out=0
    echo "Inside adjust_nfs_configs"
    # Make a backup of the config file
    echo "make backup of $NFS_SYSCONFG to $NFS_SYSCONFG_BAK"
    $CP -f $NFS_SYSCONFG $NFS_SYSCONFG_BAK
    if [ $? -ne 0 ]; then
       echo "Cannot make backup copy of $NFS_SYSCONFG, skipping adjustment"
       return 1
    fi
    echo "setting mountd port to 20048"
    #remove existing property if found
    sed  -i '/^#\?MOUNTD_PORT=.*$/d' $NFS_SYSCONFG
    out=$((out+$?))
    #set our own value
    echo MOUNTD_PORT=20048 >> $NFS_SYSCONFG

    echo "setting statd port to 662"
    #remove existing property if found
    sed  -i '/^#\?STATD_PORT=.*$/d' $NFS_SYSCONFG
    out=$((out+$?))
    #set our own value
    echo STATD_PORT=662 >> $NFS_SYSCONFG

    echo "setting NFSD proc count to 48"
    #remove existing property if found
    sed  -i '/^#\?RPCNFSDCOUNT=.*$/d' $NFS_SYSCONFG
    out=$((out+$?))
    #set our own value
    echo RPCNFSDCOUNT=48 >> $NFS_SYSCONFG

    echo "Disabling NFSv2"
    #remove existing property if found
    sed  -i '/^#\?MOUNTD_NFS_V2.=.*$/d' $NFS_SYSCONFG
    out=$((out+$?))
    #set our own value
    echo MOUNTD_NFS_V2=\"no\" >> $NFS_SYSCONFG

    sed -i -e '/^RPCNFSDARGS/d' $NFS_SYSCONFG
    out=$((out+$?))
    echo RPCNFSDARGS=\"--no-udp -N 2\" >> $NFS_SYSCONFG

    sed -i -e '/^RPCMOUNTDOPTS/d' $NFS_SYSCONFG
    out=$((out+$?))
    echo RPCMOUNTDOPTS=\"--no-udp\" >> $NFS_SYSCONFG

    echo "Enabling NFSv3"
    #remove existing property if found
    sed  -i '/^#\?MOUNTD_NFS_V3.=.*$/d' $NFS_SYSCONFG
    out=$((out+$?))
    #set our own value
    echo MOUNTD_NFS_V3=\"yes\" >> $NFS_SYSCONFG
    echo "Adjusting nfs sysconfig done"
    if [ "$out" == "0" ]; then
	return 0
    fi
    return 1
};
: '
    Setup NFS exports file
    1. replace /etc/exports.d by a symlink to /.lsa/etc/nfs/exports.d
    2. setup SELinux rules around export files
'
function setup_nfs_exports_config() {
    if [ ! -d ${CONF_MNT}/etc/nfs ]
    then
	echo "Creating layout ${CONF_MNT}/etc/nfs"
	${MKDIR} --parents ${CONF_MNT}/etc/nfs
	if [ $? -ne 0 ]
	then
	    echo "Cannot create directory ${CONF_MNT}/etc/nfs" >&2
	    return 1
	fi
    fi
    if [ ! -d ${EXPORTFS_DIR} ]
    then
	echo "Creating export layout ${EXPORTFS_DIR}"
	${MKDIR} --parents ${EXPORTFS_DIR}
	if [ $? -ne 0 ]
	then
	    echo "Cannot create directory ${CONF_MNT}/etc/nfs" >&2
	    return 1
	fi
    fi
    if [ ! -f ${EXPORTS_FILE} ]
    then
	echo "Creating export file ${EXPORTS_FILE}"
	${TOUCH} ${EXPORTS_FILE}
	if [ $? -ne 0 ]
	then
	    echo "Cannot create ${EXPORTS_FILE}" >&2
	    return 1
	fi
    fi
    if [ ! -d /etc/exports.d.bck ]
    then
	echo "backing up /etc/exports.d/ to /etc/exports.d.bck"
	${MV} /etc/exports.d /etc/exports.d.bck
	if [ $? -ne 0 ]
	then
	    echo "Cannot backup /etc/exports.d/" >&2
	    return 1
	fi
    fi
    if [ ! -h /etc/exports.d ]
    then
	echo "Linking /etc/exports.d to ${CONF_MNT}/etc/nfs/exports.d"
	${LN} --symbolic ${CONF_MNT}/etc/nfs/exports.d /etc/
	if [ $? -ne 0 ]
	then
	    echo "Cannot link /etc/exports.d to ${CONF_MNT}/etc/nfs/exports.d" >&2
	    return 1
	fi
    fi
    echo "set SELinux read permission on export files"
    $SEMANAGE fcontext -a -t public_content_t ${EXPORTS_FILE}
    if [ $? -ne 0 ]
    then
	echo "cannot semanage ${EXPORTS_FILE}"
	return 1
    fi
    $RESTORECON -F -R -v ${EXPORTS_FILE}
    if [ $? -ne 0 ]
    then
	echo "cannot restorecon ${EXPORTS_FILE}"
	return 1
    fi

    return 0
}
: '
  make sure all is exported
'
function sync_nfs_exports() {
    echo "Exporting all (NFS) exported shares"
    ${EXPORTFS} -a
    if [ $? -ne 0 ]
    then
	echo "cannot export all filesystems"
	return 1
    fi

    return 0
}

: '
    Setup all NFS related configurations and services.
    This method might by called by other scripts who has a reference to this script.
'
function main_setup_nfs(){

    echo "Starting NFS  configurations and services setup"
    out=0
    adjust_nfs_configs
    out=$((out+$?))
    restart_nfs_services
    out=$((out+$?))
    if [ "$out" == "0" ]; then
	echo "NFS setup was successful"
	return 0
    fi
    echo "NFS setup failed"
    return 1
};

script_name=$($BASENAME ${0#-}) #- needed if sourced no path
this_script=$($BASENAME ${BASH_SOURCE})

# If running this script should call the main function,
# Other wise script is used as a source
if [[ ${script_name} = ${this_script} ]] ; then
    out=0
    main_setup_nfs
    out=$((out+$?))
    setup_nfs_exports_config
    out=$((out+$?))
    if [ "$out" == "0" ]; then
	echo "NFS setup was successful"
	exit 0
    fi
    echo "NFS setup failed"
    exit 1
fi
