#!/usr/bin/bash
# Copyright (c) 2017, 2018, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.

: '
    This script creates raid device and/or volume group. Can be
    summed up by following tasks
    1. Find local NVMe devices
    2. Select a raid level
    3. Create raid
    4. Create volume group
    5. Validate and check the storage (raid, volume group)
    6. Write status of the storage to log file
'
CUR_DIR="$(dirname "$0")"
source $CUR_DIR/globals.sh
source $CUR_DIR/common.sh

: '
    Create the logical volume used to host configuration
    This function may be called in case of pool import
    some actions are skipped if items already exist
    Return 0 on success , 1 otherwise
'
function create_conf_filesystem () {
    echo "Creating configuration logical volume"
    does_lv_exist ${VG_NAME} ${CONF_LV_NAME}
    if [ $? -eq 0 ]
    then
	${LVCREATE} --size ${CONF_LV_SIZE} --monitor y --name ${CONF_LV_NAME} ${VG_NAME}
	if [ $? -ne 0 ]
	then
	    echo "Configuration logical volume (${VG_NAME}/${CONF_LV_NAME}) creation failed" >&2
	    return 1
	fi
	echo "Configuration logical volume created"
    else
	echo "logical volume ${VG_NAME}/${CONF_LV_NAME} already exists"
    fi

    # now check we have a filesystem on it
    fs=`${LSBLK} --noheadings -o FSTYPE  /dev/${VG_NAME}/${CONF_LV_NAME} 2>/dev/null`
    if [ -z "$fs" ]
    then
	echo "Creating configuration filesystem"
	${MKXFS} -f /dev/${VG_NAME}/${CONF_LV_NAME}
	if [ $? -ne 0 ]
	then
	    ${LVREMOVE} -f ${VG_NAME}/${CONF_LV_NAME}
	    if [ $? -ne 0 ]
	    then
		echo "WARNING: cannot remove __lsa lv " >&2
	    fi
	    echo "Configuration filesytem creation failed" >&2
	    return 1
	fi
    else
	echo "filesystem already created"
    fi
    if [ ! -d ${CONF_MNT} ]
    then
	${MKDIR} --parents ${CONF_MNT}
	if [ $? -ne 0 ]
	then
	    echo "Cannot create directory ${CONF_MNT}" >&2
	    return 1
	fi
    fi
    ${MOUNT} /dev/${VG_NAME}/${CONF_LV_NAME} ${CONF_MNT}
    if [ $? -ne 0 ]
    then
	echo "Cannot mount configuration lv" >&2
	${LVREMOVE} -f ${VG_NAME}/${CONF_LV_NAME}
	if [ $? -ne 0 ]
	then
	    echo "WARNING: cannot remove __lsa lv " >&2
	fi
	echo "Configuration filesytem creation failed" >&2
	return 1
    fi

    # add entry in fstab
    # example :  '/dev/lsa/foo /shares/foo xfs defaults 0 0'
    # check if we use remote storage
    _rtype=`get_pv_devices_type ${VG_NAME}`
    if [ "${_rtype}" == "LOCAL" ]
    then
	printf "%s\t%s\txfs\tdefaults\t0 0\n" /dev/${VG_NAME}/${CONF_LV_NAME} ${CONF_MNT} >> ${FSTAB_FILE}
    else
	printf "%s\t%s\txfs\tdefaults,_netdev,nofail\t0 0\n" /dev/${VG_NAME}/${CONF_LV_NAME} ${CONF_MNT} >> ${FSTAB_FILE}
    fi
    echo "Configuration filesystem created"
}

: '
    Create the configuration filesystem layout
    Return 0 on success , 1 otherwise
'
function create_conf_layout () {
    if [ ! -d ${CONF_MNT}/etc/ ]
    then
	${MKDIR} ${CONF_MNT}/etc/
	if [ $? -ne 0 ]
	then
	    echo "Cannot create directory ${CONF_MNT}/etc/" >&2
	    return 1
	fi
    else
	echo "Directory ${CONF_MNT}/etc/ already exits"
    fi

    return 0
}
: '
    Create the volume group "lsa"
    Parameter:
	devices: list of space-separated paths (e.g., "/dev/md0" or "/dev/sdb /dev/sdc")
'
function create_volume_group() {
    devs=(${@:1})
    out=0
    for ((c=0; c<${#devs[@]}; c++))
    do
	devstr="${devs[$c]}"
	$PVCREATE -v $devstr
	check_exit "Physical volume creation on $devstr"
	out=$((out+$?))
    done
    $VGCREATE -v $VG_NAME ${devs[@]}
    check_exit "Volume group ($VG_NAME) creation"
    out=$((out+$?))
    if [ "$out" == "0" ]; then
	echo "${FUNCNAME[0]} OK"
	return 0
    fi
    echo "${FUNCNAME[0]} failed"
    return 1
};

: '
    Create raid on the specified list of disks
    Parameters:
	raid_level: raid level (0, 1, 6, or 10)
	devices: list of space-separated paths (e.g., "/dev/sdb /dev/sdc")
'
function create_raid() {
    now=$(date "+%Y-%m-%d %H:%M:%S")
    echo "${now} Inside ${FUNCNAME[0]}"
    if [ $# -lt 2 ]; then
	echo "Error: ${FUNCNAME[0]} needs at least 2 parameters."
	return 1
    fi
    raid_level=$1
    shift
    devs=(${*})
    num_devs=${#devs[@]}

    if [ $raid_level -ne 0 ] && \
       [ $raid_level -ne 1 ] && \
       [ $raid_level -ne 6 ] && \
       [ $raid_level -ne 10 ]; then
	echo "${FUNCNAME[0]} failed: raid_level=$raid_level, not supported"
	return 1
    fi

    if [ $raid_level -eq 0 ]; then
	# have to use --force to create raid0 with 1 disk
	mdadm_options="--force"
    fi
    if [ $raid_level -eq 1 ]; then
	# use --metadata=1.2 to avoid a question, and this array can not be
	# used for /boot unless the boot-loader understands md/v1.x metadata
	mdadm_options="--metadata=1.2"
    fi

    $MDADM --create $RAID_DEV --level=$raid_level $mdadm_options --raid-devices=$num_devs ${devs[@]}
    check_exit "raid $raid_level creation"
    out=$?

    echo 'DEVICE' ${devs[@]} >> $MD_CONF
    echo $($MDADM --detail --scan $RAID_DEV) >> $MD_CONF
    now=$(date "+%Y-%m-%d %H:%M:%S")
    if [ "$out" == "0" ]; then
	echo "${now} ${FUNCNAME[0]} done"
	return 0
    else
	echo "${now} ${FUNCNAME[0]} failed"
	return 1
    fi
};

: '
    Get the list of available disks in the system.
    Return a list of disks. For example, sda sdb ... sdx
'
function get_available_devs() {
    devs=$($LSBLK -o KNAME,TYPE | $AWK '$2 == "disk" {print  $1}')
    echo $devs
};

: '
    Get the list of local NVMe devices in the system.
    Echo a list of paths (e.g., "/dev/nvme0n1 /dev/nvme1n1")
    Returns 0 if disks found , 1 otherwise
'
function find_local_nvme_devices() {
    devs=$(get_available_devs)
    OIFS=$IFS
    IFS=" "
    devs=($devs)
    final=()
    val=0
    for ((c=0; c <${#devs[@]}; c++))
    do
	devstr="${devs[$c]}"
	# skip non-nvme devices
	if [[ $devstr != nvme* ]]; then
	    continue
	fi

	totalPartitions=0
	dev="/dev/${devs[$c]}"
	ROT=1
	rot_path="/sys/block/${devs[$c]}/queue/rotational"
	if [ -f $rot_path ]; then
	    ROT=$(cat $rot_path)
	fi
	# Exclude system disk and raid configured disk which are used by OS
	totalPartitions=$(lsblk -l $dev -o TYPE|grep -E 'part|raid'|wc -l)
	# Check if this isn't rotational device and doesn't have partitions
	if [ $totalPartitions -eq 0 ] && [ $ROT -eq 0 ]; then
	    final[$val]=$dev
	    val=$val+1
	fi
    done

    no_disk=${#final[@]}

    echo ${final[@]}
    if [ $no_disk -ge 1 ]; then
	return 0
    else
	return 1
    fi
};

: '
    Select a raid level based on the number of devices.
    Parameters:
	num_devs: number of devices
    Return:
	raid_level: raid level (1, 6, 10, or $RAID_LEVEL_NONE)
'
function select_raid_level() {
    num_devs=$1
    if [ $num_devs -ge 5 ]; then
	raid_level=6
    elif [ $num_devs -ge 3 ]; then
	raid_level=10
    elif [ $num_devs -eq 2 ]; then
	raid_level=1
    else
	raid_level=$RAID_LEVEL_NONE
	echo "${FUNCNAME[0]} failed: num_devs=$num_devs"
    fi
    return $raid_level
};

: '
  This function creates the /.lsa filesystem used to store
  the exports configuration
  Return 0 on success , 1 otherwise
'
function create_storage_configuration() {

    create_conf_filesystem
    if [ $? -ne 0 ]; then
	echo "${FUNCNAME[0]} failed."
	return 1
    fi

    create_conf_layout
    if [ $? -ne 0 ]; then
	echo "${FUNCNAME[0]} failed."
	return 1
    fi

    return 0
};

: '
    This function checks if volume group lsa has been set up, if so returns error.
    Otherwise creates raid on local nvme devices and creates a volume group on it,
    or creates a volume group on remote block volumes (or a single local nvme device) directly.
    Parameters:
	raid_level: raid level (0, 1, 6, 10, or $RAID_LEVEL_NONE)
	devices: list of space-separated paths (e.g., "/dev/sdb /dev/sdc")
'
function create_storage() {
    if [ $# -lt 2 ]; then
	echo "Error: ${FUNCNAME[0]} needs at least 2 parameters."
	return 1
    fi
    raid_level=$1
    shift
    devices=(${*})

    # If volume group lsa exists, return error from here.
    does_vg_exist $VG_NAME
    if [ $? -eq 1 ]; then
	echo "Error: VG $VG_NAME existed. Exiting."
	return 1
    else
	echo "VG $VG_NAME not found. Trying to create it."
    fi

    if [ $raid_level -ne $RAID_LEVEL_NONE ]; then
	# Check if raid device /dev/md0 exists and is active.
	$MDADM --detail --brief $RAID_DEV | $GREP ARRAY | $GREP -v INACTIVE >/dev/null 2>&1
	if [ $? -ne 0 ]; then
	    echo "$RAID_DEV not found. Trying to create it."
	    create_raid $raid_level ${devices[@]}
	    if [ $? -eq 0 ]; then
		echo "$RAID_DEV created."
	    else
		echo "${FUNCNAME[0]} failed at $RAID_DEV creation."
		return 1
	    fi
	else
	    echo "Error: $RAID_DEV existed. Exiting."
	    return 1
	fi
	devices=($RAID_DEV)
    fi
    echo "Now trying to create the volume group."
    create_volume_group ${devices[@]}
    if [ $? -ne 0 ]; then
	echo "${FUNCNAME[0]} failed."
	return 1
    fi

    echo "Storage pool creation OK"
    return 0
};

: '
    Checks the raid device entry in /etc/mdadm.conf file for consistency
'
function check_raid_device() {
    echo "Checking raid configuration"

    # Check if mdadm configuration file exists
    if [ -f $MD_CONF ]; then
	total_entries=$($GREP -c "ARRAY $RAID_DEV" $MD_CONF)
	echo "Total entries found in configuration: $total_entries"
	# There must be one configuration entry for now
	if [ $total_entries -gt 1 ]; then
	    echo "More than one configuration found in $MD_CONF, fix it manually."
	    return 1
	elif [ $total_entries -eq 0 ]; then
	    # This case is unexpected. Exit with 1
	    echo "$MD_CONF file found but no configuration. Contact with Administrator."
	    return 1
	fi
    else
	echo "$MD_CONF file not found."
	return 1
    fi
    return 0
};

: '
    Validates the raid device after setup or in a pre-configured system
'
function validate_raid_device() {
    # Here we have an active raid configuration, parse the configuration and write to the log file

    active=()
    faulty=()
    spare=()
    aval=0
    fval=0
    sval=0
    raid_status="unknown"
    raid_dev=0

    echo "Reading mdstat from proc file"

    if [ -f $MDSTAT_PROC ]; then
	echo "Reading $MDSTAT_PROC"
	while IFS= read -r line;
	do
	    if [[ $line  == "md0"* ]]; then
		dev=$(cut -d':' -f1 <<<"$line")
		rest=$(cut -d':' -f2 <<<"$line")
		OIFS=$IFS
		IFS=" "
		parts=($rest)
		length=${#parts[@]}
		size=$((length-2))
		raid_status=${parts[0]}
		raid_dev=${parts[1]}
		for ((c=2; c<length; c++))
		do
		    val=$val+1
		    str="${parts[$c]}"
		    status=${str:(-3)}
		    dv=$(cut -d'[' -f1 <<<"$str")
		    if [[ $status == "(F)" ]]; then
			faulty[$fval]=$dv
			fval=$fval+1
		    elif [[ $status == "(S)" ]]; then
			spare[$sval]=$dv
			sval=$sval+1
		    else
			active[$aval]=$dv
			aval=$aval+1
		    fi
		done
	    fi
	done < $MDSTAT_PROC
    fi

    IFS=$'\n' sorted=($(sort <<<"${active[*]}"))
    echo "Writing raid information"
    echo "raid_status:$raid_status"
    echo "num_devs:$size"
    echo "active_devs:${sorted[@]}"
    echo "faulty_devs:${faulty[@]}"
    echo "spare_devs:${spare[@]}"
    total=("${active[@]}" "${faulty[@]}" "${spare[@]}")
    IFS=$'\n' total=($(sort <<<"${total[*]}"))
    echo "all_devs:${total[@]}"
    echo "init_checked_by_backend:no"
    echo "Writing raid information done."
    return 0
};

: '
  checks that a LVM volume group exists
  return 1 if exists, 0 otherwise
  parameter:
   $1 : vg name
'
function does_vg_exist() {
    $VGS --quiet -o vg_name $1 >/dev/null 2>&1
    if [ $? -eq 0 ]
    then
	return 1
    fi
    return 0
}

: '
  checks that a LVM logical volume exists
  return 1 if exists, 0 otherwise
  parameter:
   $1 : vg name
   $2 : lv name
'
function does_lv_exist() {
    $LVS --quiet -o lv_name ${1}/${2} >/dev/null 2>&1
    if [ $? -eq 0 ]
    then
	return 1
    fi
    return 0
}

: '
  maintain /etc/fstab in sync with LV found in lsa VG
'
function sync_fstab() {
    echo "syncing fstab"

    # check if we use remote storage because we need to add netdev options then
    _rtype=`get_pv_devices_type ${VG_NAME}`

    ${LVS} --no-headings -o lv_name,path $VG_NAME 2>/dev/null | while read _lv _lvpath
    do
	${AWK} -v _p=${_lvpath} '$1 == _p {exit 12}' < ${FSTAB_FILE}
	if [ $? -ne 12 ]
	then
	    echo "${_lvpath} not found , add it to fstab"
	    if [ "${_rtype}" == "LOCAL" ]
	    then
	       printf "%s\t%s\txfs\tdefaults\t0 0\n" ${_lvpath} ${SHARE_POINT}/${_lv} >> ${FSTAB_FILE}
	    else
	       printf "%s\t%s\txfs\tdefaults,_netdev,nofail\t0 0\n" ${_lvpath} ${SHARE_POINT}/${_lv} >> ${FSTAB_FILE}
	    fi
	fi
    done
}
: '
  Make sure all shares have mount point and are mounted
'
function mount_all() {
    echo "mounting all shares"
    ${LVS} --no-headings -o lv_name $VG_NAME 2>/dev/null | while read _lv
    do
	if [ ${_lv} == ${CONF_LV_NAME} ]
	then
	    continue
	fi
	_do_mount=0
	if [ ! -d ${SHARE_POINT}/${_lv} ]
	then
	    _do_mount=1
	    echo "creating ${SHARE_POINT}/${_lv}"
	    ${MKDIR} ${SHARE_POINT}/${_lv}
	    if [ $? -ne 0 ]
	    then
		_do_mount=0
		echo "WARNING: cannot create mount point ${SHARE_POINT}/${_lv}"
	    else
		${CHOWN} ${SHARE_DIR_USER_OWNERSHIP}:${SHARE_DIR_GROUP_OWNERSHIP} ${SHARE_POINT}/${_lv}
		if [ $? -ne 0 ]
		then
		   echo "WARNING: cannot chown'ed mount point ${SHARE_POINT}/${_lv}"
		fi
		${CHMOD} ${SHARE_DIR_ACCESS_RIGHT} ${SHARE_POINT}/${_lv}
		if [ $? -ne 0 ]
		then
		   echo "WARNING: cannot chmod'ed mount point ${SHARE_POINT}/${_lv}"
		fi
	    fi
	else
	    # check if /shares/<share name> is mounted (i.e found within /etc/mtab)
	    ${AWK} -v _p=${SHARE_POINT}/${_lv} '$2 == _p {exit 12}' < ${PROC_MOUNT}
	    if [ $? -ne 12 ]
	    then
		echo "${_lv} not mounted , mount it now"
		_do_mount=1
	    fi
	fi
	if [ $_do_mount == 1 ]
	then
	    ${MOUNT} ${SHARE_POINT}/${_lv}
	    if [ $? -ne 0 ]
	    then
		echo "WARNING: mount of ${SHARE_POINT}/${_lv} failed"
	    fi
	fi
    done
}

: '
    This method might be called by other scripts which have a reference to this script.
'
function main_setup_storage() {
    echo "Setting up storage (raid, volume group)"
    if [ $# -lt 2 ]; then
	echo "Error: ${FUNCNAME[0]} needs at least 2 parameters."
	echo "Storage setup (raid, volume group) failed"
	return 1
    fi
    raid_level=$1
    shift
    devices=(${*})
    echo "${FUNCNAME[0]}: raid_level=$raid_level, devices=${devices[@]}"
    out=0
    create_storage $raid_level ${devices[@]}
    out=$((out+$?))
    if [ $raid_level -ne $RAID_LEVEL_NONE ]; then
	check_raid_device
	out=$((out+$?))
	validate_raid_device
	out=$((out+$?))
    fi
    if [ $out -ne 0 ]; then
	echo "Storage setup (raid, volume group) failed"
	return 1
    fi

    create_storage_configuration
    if [ $? -ne 0 ]
    then
	echo "configuration storage setup failed"
	return 1
    fi
    return 0
};

: '
  Gets the type of device used as PV for a given VG
  Returns LOCAL or REMOTE
  Param:
	$1 : VG name
  Note: on OCI all remote storage seems to have "BlockVolume" as model
	We rely on this to make distinction
'
function get_pv_devices_type () {
    # we assume all device are the same type within a VG
    # check the first one
    _dev=`${VGS} --noheadings --unbuffered --rows --aligned --options pv_name $1 2>/dev/null | awk '{ print $1 }'`
    _model=`${LSBLK}  --list --noheadings --output MODEL ${_dev} 2>/dev/null`
    # strip blanks
    _model=`echo ${_model}`
    if [ "${_model}" == "BlockVolume" ]
    then
	echo "REMOTE"
    else
	echo "LOCAL"
    fi
}

script_name=$($BASENAME ${0#-}) #- needed if sourced no path
this_script=$($BASENAME ${BASH_SOURCE})

# If running this script should call the main function,
# Otherwise script is used as a source
if [[ ${script_name} = ${this_script} ]]; then
    out=0
    main_setup_storage $@
    out=$((out+$?))

    if [ "$out" == "0" ]; then
	echo "Storage setup (raid, volume group) was successful"
	exit 0
    fi
    echo "Storage setup (raid, volume group) failed"
    exit 1
fi
