# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
'''
Key value store utility class

'''
import shutil
import os
from utility import CustomConfigParser
from log_util import get_logger, set_log_config

LOG = get_logger('lsa.samba_util')

DEFAULT_SECTION = 'global'


class KeyValueStore(CustomConfigParser):
    '''
    Simple key value store. This class store key value pair to the file specified in the
    constructor parameter. An unique ID can also be provided to make it unique section.
    If the section id is not provided it uses default section as global
    '''
    def __init__(self, file_name, id_name=DEFAULT_SECTION):
        CustomConfigParser.__init__(self)
        self.file_name = file_name
        self.section = id_name
        if not os.path.exists(file_name):
            open(file_name, 'w').close()
        self.read_file(self.file_name)

    def add_key(self, key, val):
        '''
        Add a key value pair to the file
        parameter:
        key: name of the key as string
        val: value of the key as string
        '''
        self.add_keys({key: val})

    def add_keys(self, items):
        '''
        Add dictioanry items
        raise: StandardError if fails
        parameter:
        items: dictionary items
        '''
        if self.has_section(self.section) is False:
            self.add_section(self.section)
        for key, value in items.iteritems():
            self.set(self.section, key, value)
        self._flush()

    def _flush(self):
        '''
        Flush the config items from in memory to file
        raise: StandardError if fails
        '''
        conf_bak = self.file_name + '.bak'
        try:
            shutil.copy(self.file_name, conf_bak)
        except IOError, e:
            raise StandardError('cannot make a backup copy of samba configuration [%s]' % str(e))
        try:
            fo = open(self.file_name, 'w')
            self.write(fo)
        except (IOError, OSError) as e:
            LOG.exception('Failed to write samba config file: %s' % str(e))
            raise StandardError('Failed to write samba config file: %s' % str(e))
        finally:
            fo.close()

    def remove_key(self, key):
        '''
        Remove a key from the file
        raise: StandardError if fails
        '''
        self.remove_keys([key])

    def remove_keys(self, list_of_keys):
        '''
        Remove a list of keys from the file
        raise: StandardError if fails
        '''
        try:
            for key in list_of_keys:
                if self.has_option(self.section, key):
                    self.remove_option(self.section, key)
        except Exception as e:
            raise StandardError('Failed to remove the keys: %s' % str(e))
        self._flush()

    def get_value(self, key):
        '''
        Get the value of the key from the file
        '''
        try:
            return self.get(self.section, key)
        except Exception as e:
            raise StandardError('Failed to get the key: %s' % str(e))

    def get_all(self):
        try:
            return dict(self.items(self.section))
        except Exception as e:
            raise StandardError('Failed to get all the keys of the section(%s): %s' % (self.section, str(e)))


if __name__ == '__main__':
    set_log_config()
