# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
"""
NFS util:
This file manages all the NFS configuration, operations(add/modify/remove/list share).

Usage:

List current shares:
nfs_manager = NfsExports()
shares = nfs_mngr.get_nfs_entries() # This will give you the list of shares
each item would be an object of NfsExportEntry class

Add Shares:
nfs_manager = NfsExports()
nfs_manager.add_entry(name, hosts, options)
name = name of the share/directory
hosts = ip address/or host name of the client
options = comma separated option for example ro,rw,sync etc

Modify Shares:
Input: name, hosts, options
nfs_manager = NfsExports()
nfs_manager.remove_entry(name)
nfs_manager.add_entry(name, hosts, options)
Delete Shares:
nfs_manager = NfsExports()
nfs_manager.remove_nfs_export_by_name(name)
"""
from backend.common import SYSCTL
from backend.command_util import execute_command
from backend.service_util import SystemdService
from backend.log_util import get_logger, set_log_config
from backend.validation import is_valid_file_name, is_valid_share_client
import json

from backend.export_base import ExportBase
from backend.option_util import BooleanOption, BooleanOptionWithPrefix
from backend.utility import parse_boolean
import copy

LOG = get_logger('lsa.nfs_util')


SERVICE_NFS_IDMAPD = "nfs-idmapd"
SERVICE_NFS_MOUNTD = "nfs-mountd"
SERVICE_NFS_SERVER = "nfs-server"

NFS_SERVICES = [SERVICE_NFS_IDMAPD, SERVICE_NFS_MOUNTD, SERVICE_NFS_SERVER]

__all__ = ['restart_nfs_services', 'NfsExportEntry']


def restart_nfs_services():
    """
    Restart NFS related services
    i.e perform reload-or-restart on nfs-idmapd, nfs-mountd, nfs-server
    raise:
        StandardError : restart (systemctl command) has failed
    """
    LOG.debug('restartNFSServices')
    # Note: we do not use SystemdService class as we must call the restart only once
    #      and let dependencies do the work. Cannot aford multiple restart
    #      We assume restart is synchronous

    cmd = [SYSCTL, SystemdService.RELOAD_OR_RESTART] + NFS_SERVICES

    out, err, code = execute_command(cmd)

    LOG.debug('restartNFSServices (%d,%s,%s)' % (code, out, err))

    if code != 0:
        LOG.warning('restartNFSServices failure %s' % err)
        raise StandardError('cannot restart NFS stack: %s' % err)

    LOG.debug('restart_nfs_services done')


class NfsExportEntry(ExportBase):
    """
    This class represents NFS export entry
    hosts: host names share is exported to
    path: path of the share
    options: export options
    """
    def __init__(self, hosts, path, **options):
        '''
        Parameter:
            hosts: hosts to be shared with
            path: path/dir(access point) of the share
            options: dictionary items
                {read_only: true/false,
                 sync: true/false,
                 root_squash: true/false,
                 hide: true/false,
                 subtree_check: true/false,
                 secure: true/false
                }
        '''
        ExportBase.__init__(self, path, 'nfs', **options)
        if hosts is None:
            raise TypeError('hosts(%s) must be valid item' % hosts)
        if not is_valid_share_client(hosts):
            raise ValueError("Invalid client specification: " + hosts)
        if not is_valid_file_name(path):
            raise ValueError("Invalid path: " + path)
        self.sync = BooleanOption('sync', False)
        self.subtree_check = BooleanOptionWithPrefix('subtree_check')
        self.secure = BooleanOptionWithPrefix('secure', 'in')
        self.hide = BooleanOptionWithPrefix('hide', 'no', True)
        self.root_squash = BooleanOptionWithPrefix('root_squash')
        self.hosts = hosts
        if options:
            self.populate_options_from_dict(options)

        LOG.debug('new export instance: %s' % str(self))

    @classmethod
    def create_from_line(cls, path, line):
        '''
        Create the object from a line getting from export file
        Line format: /shares/blah x.x.x.x(ro, secure, nohide, no_subtree_check)
        At first we retrieve the parts from the line by splitting. Then create the object
        and assign the attribute.
        '''
        line = line.lstrip()

        if len(line) > 0:
            parts = line.split('(')
            hosts = parts[0]
            remaining = parts[1].replace(")", "")
            ne = cls(hosts, path)
            ne.populate_options(remaining)
            return ne
        else:
            raise ValueError("Persistent line is empty, not creating NfsExportEntry object")

    def get_id(self):
        '''
        Get the ID(hosts) of the share that is exported
        '''
        return self.hosts

    def get_path(self):
        '''
        Get the path of the share that is exported
        '''
        return self.path

    def get_hosts(self):
        '''
        Get the hosts of the share exported to as nfs clients
        '''
        return self.hosts

    def _get_options(self):
        '''
        Get all the options except read_only. we don't include read_only here because its
        representation is a bit different True: ro, False: rw
        '''
        options = []
        options.append(self.sync)
        options.append(self.root_squash)
        options.append(self.subtree_check)
        options.append(self.secure)
        options.append(self.hide)
        return options

    def _set_option(self, key, val):
        '''
        Set an option
        parameter:
            key: name of the option
            val: value of the option True, False ( boolean or string or int)
        '''
        if val is None:
            raise KeyError('Value of %s can not be None' % key)
        enabled = parse_boolean(val)
        if key == 'read_only':
            if enabled:
                self.read_only.enable()
            else:
                self.read_only.disable()
        elif key == 'sync':
            if enabled:
                self.sync.enable()
            else:
                self.sync.disable()
        elif key == 'hide':
            if enabled:
                self.hide.enable()
            else:
                self.hide.disable()
        elif key == 'root_squash':
            if enabled:
                self.root_squash.enable()
            else:
                self.root_squash.disable()
        elif key == 'subtree_check':
            if enabled:
                self.subtree_check.enable()
            else:
                self.subtree_check.disable()
        elif key == 'secure':
            if enabled:
                self.secure.enable()
            else:
                self.secure.disable()
        else:
            raise KeyError('Key not found:' + key)

    def __repr__(self):
        """
        This function returns hosts(options) same as the format present in /.lsa/etc/nfs/exports.d/lsa.exports
        output format: x.x.x.x(ro, secure, root_squash) etc
        """
        return self.hosts + '(' + self.get_options_as_string() + ')'

    def get_options_as_string(self):
        """
        Get comma separated options
        Format: rw,sync,nohide,*
        """
        out = ""
        if self.read_only.enabled():
            out = out + 'ro,'
        else:
            out = out + 'rw,'
        for opt in self._get_options():
            if len(repr(opt)) > 0:
                out = out + repr(opt) + ','
        return out[:-1]

    def get_options_as_dict(self):
        '''
        Get the options as dictionary item
        '''
        opts = copy.deepcopy(self._get_options())
        out = {}
        out.update(self.read_only.to_dict())
        for opt in opts:
            out.update(opt.to_dict())
        return out

    def populate_options(self, raw_options):
        """
        parse the comma separated options and populate the dictionary
        input format: "ro,insecure,nohide,no_root_squash"
        """

        r_opts = raw_options.replace(' ', '')
        parts = r_opts.split(',')

        for part in parts:
            part = part.replace(' ', '')
            if part == 'ro':
                self.read_only.enable()
            elif part == 'rw':
                self.read_only.disable()
            elif part == 'sync':
                self._set_option('sync', True)
            elif part == 'no_subtree_check':
                self._set_option('subtree_check', False)
            elif part == 'subtree_check':
                self._set_option('subtree_check', True)
            elif part == 'insecure':
                self._set_option('secure', False)
            elif part == 'secure':
                self._set_option('secure', True)
            elif part == 'nohide':
                self._set_option('hide', False)
            elif part == 'no_root_squash':
                self._set_option('root_squash', False)
            elif part == 'root_squash':
                self._set_option('root_squash', True)

    def populate_options_from_dict(self, dict_options):
        '''
        Retrieve nfs options from dictionary
        '''
        for key, value in dict_options.iteritems():
            self._set_option(key, value)

    def __str__(self):
        return "id=[%s] entry=[%s]" % (self.get_id(), repr(self))

    def to_json(self):
        return json.dumps({
            'type': self.type,
            '__type__': self.__class__.__name__,
            'id': self.get_id(),
            'path': self.get_path(),
            'hosts': self.get_hosts(),
            'options': self.get_options_as_string()
        })

    def serialize(self):
        return dict(type=self.type,
                    __type__=self.__class__.__name__,
                    id=self.get_id(),
                    path=self.get_path(),
                    hosts=self.get_hosts(),
                    options=self.get_options_as_dict())

    @classmethod
    def deserialize(cls, js):
        """
        create an object from json format data
        parameter:
          input:
            js: json format
                {'path': '/shares/jsontest3',
                 'name': 'jsontest3',
                 'hosts': '192.168.1.1',
                 'options': 'rw,sync,no_root_squash,no_subtree_check,nohide'
                }
          output: NfsExportEntry object
        """
        deserializeJS = json.loads(js)
        ne = cls(deserializeJS.get('hosts'), deserializeJS.get('path'))
        ne.populate_options(deserializeJS.get('options'))
        return ne

    def __eq__(self, other):
        """
        Check for equality by testing entry name equality
        name uniquely identify an entry
        """
        if id(self) == id(other):
            return True
        if other is None:
            return False

        if isinstance(other, NfsExportEntry):
            return self.path == other.path and self.get_hosts() == other.get_hosts()
        if isinstance(other, basestring):
            return self.__eq__(NfsExportEntry(other))

    def __ne__(self, other):
        """
        see __eq__
        """
        return not self.__eq__(other)


def get_nfs_entries_from_line(line):
    '''
    Convert a  export line to list of NfsExportEntry objects
    parameter:
        line: export line as string , format: /access_point hosts1(options) host2(options) ... ...
    output: lust of NfsExportEntry objects
    raise: StandardError if failed to parse the line
    '''
    # Removing white spaces from beginning
    line = line.strip()
    # skip the line if starts with #
    if line.startswith('#'):
        return []
    parts = line.split()
    path = parts[0]
    num_items = len(parts)
    entries = []
    for i in range(1, num_items):
        entries.append(NfsExportEntry.create_from_line(path, parts[i]))
    return entries


def run_all_test():
    """
    Not implemented yet
    """


if __name__ == '__main__':
    set_log_config()
    run_all_test()
