# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
"""
Module that holds helper methods and classes for OCI object store support.
All operations a intent to be asynchronous and so performed in threads
Upload involve two layers
  - StreamMultiPartUploader is responsible to initiate and commit the uploads. it spawn StreamUploader instances
  - StreamUploader worker thread actually doing the upload
     see https://oracle-cloud-infrastructure-python-sdk.readthedocs.io/en/latest/api/index.html?highlight=upload_part#oci.object_storage.object_storage_client.ObjectStorageClient.upload_part

download (object retrieveal) is done by StreamDownloader instance

ex:

queue = StreamQueue(3)
taskID = task_manager.post('myShare','backup')
uploader = StreamMultiPartUploader(taskID, queue, oci.config.from_file(),
                        'mynameSpace', 'bucketName', 'objectName')
uploader.init_upload()
uploader.start()




queue = StreamQueue(3)
downloader = StreamDownloader(queue, oci.config.from_file(), 'mynameSpace',
                            'bucketName', 'objectName')
downloader.init_download()
downloader.start()




"""

import oci.object_storage
from oci.exceptions import ServiceError as OCIServiceError
from oci.exceptions import ClientError as OCIClientError

import threading
import cStringIO
import Queue
from backend.log_util import get_logger, set_log_config
from backend.common import OS_STREAM_READ_SIZE
from backend.lsa_types import (StreamPart, EOSPart, ErrorPart)
from backend.taskmgr import (TaskStatusCode, TaskStatus)

LOG = get_logger('lsa.object_store_util')


class StreamUploader(threading.Thread):
    """
    Worker thread around multipart upload API
    https://oracle-cloud-infrastructure-python-sdk.readthedocs.io/en/latest/api/index.html?highlight=upload_part#oci.object_storage.object_storage_client.ObjectStorageClient.upload_part

    This worker thread consume part from a StreamQueue and upload them using the mulipart upload API

    """

    """
    Creates a StreamUploader instance.
    Parameters:
     osClient: an instance of oci.object_storage.object_storage_client.ObjectStorageClient
     namespace_name: the namespace name of the remote object store
     bucket_name: the bucket name of the remote object store
     object_name: the object name to be uploaded to  the remote object store
     multi_upload_id: the id of the multipart upload
     queue: an instance of  StreamQueue. part to be uploaded are retrieved from it.
    """
    def __init__(self, osClient, namespace_name, bucket_name, object_name, multi_upload_id, queue):
        threading.Thread.__init__(self)
        self.object_storage_client = osClient
        self.namespace_name = namespace_name
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.multi_upload_id = multi_upload_id
        self.queue = queue
        self.do_stop = False
        self.upload_details = []
        self.ok = True

        self.name = 'StreamUploader{%d}' % id(self)

    def get_upload_details(self):
        """
        Returns list of CommitMultipartUploadPartDetails for uploaded parts
        """
        return self.upload_details

    def is_upload_successful(self):
        """
        Returns True if upload has been successfull
        """
        return self.ok

    def stop(self):
        """
        Request this worker thread to stop
        """
        self.do_stop = True

    def __str__(self):
        return 'StreamUploader(%d)#%s' % (id(self), self.object_name)

    def run(self):
        """
        starts the upload
        uploads go on until the queue becomes inactive and empty
        """
        total_size = 0
        LOG.debug('uploader starting')

        while True:
            if self.do_stop:
                LOG.debug('being asked to stop')
                break
            if self.queue.empty() and not self.queue.is_active():
                LOG.debug('queue inactive and empty, stopping now')
                break

            LOG.debug('waiting for a new part')
            try:
                # never block,always get out on timeout
                # with concurrency we may have entered the loop because the queue was (still) active
                # but not anymore
                new_part = self.queue.get(True, 30)
            except Queue.Empty, e:
                continue

            if isinstance(new_part, EOSPart):
                LOG.debug('new part of zero size, stopping now')
                self.queue.task_done()
                break
            LOG.debug('new part fetched %s ' % str(new_part))
            try:
                self.object_storage_client.upload_part(
                    self.namespace_name,
                    self.bucket_name,
                    self.object_name,
                    self.multi_upload_id,
                    new_part.get_part_num(),
                    new_part.get_buffer())
                LOG.debug('upload on part %s (size=%d) done' %
                          (new_part.get_part_num(),
                           new_part.get_size()))
                total_size += new_part.get_size()
                _d = oci.object_storage.models.CommitMultipartUploadPartDetails()
                _d.part_num = new_part.get_part_num()
                _d.etag = ''
                self.upload_details.append(_d)
            except OCIServiceError, e:
                LOG.exception('upload has failed')
                self.error = e.args[0]
                self.ok = False
            finally:
                new_part.cleanup()
                self.queue.task_done()
                LOG.debug('total size pushed [%d]' % total_size)


class StreamDownloader(threading.Thread):
    """
    Takes care of object download from OCI object store
    https://oracle-cloud-infrastructure-python-sdk.readthedocs.io/en/latest/api/index.html?highlight=get_object#oci.object_storage.object_storage_client.ObjectStorageClient.get_object

    This worker thread download an remote object and fill a StreamQueue

    """

    def __init__(self, queue, client, namespace_name, bucket_name, object_name):
        """
        Instanciates a new StreamDownloader
        Parameters:
          queue: an instance of  StreamQueue. downloaded parts are push to it.
          client: an OCI object storage client
          namespace_name: object store namespace name
          bucket_name: object store bucket name
          object_name: object store object name
        """
        threading.Thread.__init__(self)
        self.namespace_name = namespace_name
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.object_storage_client = client
        self.expected_size = 0.0
        self.queue = queue

        self.name = 'StreamDownloader{%s}' % object_name

        LOG.debug('new downloader for %s %s %s' %
                  (self.namespace_name, self.bucket_name, self.object_name))

    def init_download(self):
        """
        Init this instance for download
        i.e init the OCI object store client and check for object presence
        """
        try:
            # get object size. This also then checks that given object exists
            response = self.object_storage_client.list_objects(
                self.namespace_name,
                self.bucket_name,
                prefix=self.object_name,
                fields='name,size'
            )
            if len(response.data.objects) == 0:
                LOG.warning('no object found for [%s]' % self.object_name)
                raise StandardError('object not found')

            self.expected_size = float(response.data.objects[0].size)
            LOG.debug('object size to be downloaded [%d]' % self.expected_size)
        except (OCIServiceError, OCIClientError), e:
            raise StandardError('init of upload failed: %s' % e.message)

    def run(self):
        """
        Starts the worker thread
        """

        LOG.debug('starting')
        total_size = 0.0
        response = None
        try:
            # actually gets the object. Initiate the download
            response = self.object_storage_client.get_object(self.namespace_name,
                                                             self.bucket_name,
                                                             self.object_name)
        except OCIServiceError, e:
            self.queue.put(ErrorPart('download failed: %s' % e.messag))
            self.queue.set_active(False)
            raise StandardError('download failed: %s' % e.message)

        # perform 10MB read on the download stream
        for data in response.data.raw.stream(OS_STREAM_READ_SIZE, decode_content=False):
            # first check that queue is still active, i.e somebody actually wait
            if not self.queue.is_active():
                LOG.debug('queue no longer active, stop now')
                break
            # we do not care about numbering in download, use 1 all the time
            _s = len(data)
            total_size += float(_s)
            part = StreamPart(1, _s, cStringIO.StringIO(data))
            # feed the StreamQueue
            self.queue.put(part, True)
            LOG.debug('new part posted, %.1f pct downloaded', ((total_size / self.expected_size) * 100.0))
            if _s == 0:
                LOG.debug('end of stream reached')
                # end of stream, set the queue inactive
                self.queue.set_active(False)
                break

        LOG.debug('end of stream reached')
        self.queue.set_active(False)
        self.queue.put(EOSPart(), True)

        LOG.debug('total size downloaded [%d]' % total_size)


class StreamMultiPartUploader(threading.Thread):
    """
    Takes care of multipart upload
    Multipart upload is dedicated to sub uploaders.
    This will spawn StreamUploader instances and wait for completion of their upload

    """

    def __init__(self, task, queue, client, namespace_name, bucket_name, object_name):
        """
        Instances a new StreamMultiPartUploader
        Parameters:
          task: Instance of BackupTask
          queue: an instance of  StreamQueue. downloaded parts are push to it.
          client: an OCI object storage client
          namespace_name: object store namespace name
          bucket_name: object store bucket name
          object_name: object store object name
        """
        threading.Thread.__init__(self)
        self.namespace_name = namespace_name
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.multi_upload_id = None
        self.partCount = 0
        self.object_storage_client = client
        self.long_task = task

        self.queue = queue

        self.name = 'StreamMultiPartUploader{%s}' % object_name

        LOG.debug('new uploader for %s %s %s' %
                  (self.namespace_name, self.bucket_name, self.object_name))

    def init_upload(self):
        """
        Initialize this instance for upload
        Create the object store client and init the multipart upload
        """
        try:
            self.multiUploadDetails = oci.object_storage.models.CreateMultipartUploadDetails()
            self.multiUploadDetails.object = self.object_name
            response = self.object_storage_client.create_multipart_upload(self.namespace_name,
                                                                          self.bucket_name,
                                                                          self.multiUploadDetails)
            self.multi_upload_id = response.data.upload_id
        except (OCIServiceError, OCIClientError), e:
            raise StandardError('init of upload failed: %s' % e.message)

    def run(self):
        """
        starts worker thread
        """
        self.long_task.begin()
        self.upload()

    def upload(self):
        """
        Start the upload
        Creates StreamUploader instances and join them
        """
        self.uploaders = [StreamUploader(self.object_storage_client, self.namespace_name, self.bucket_name,
                                         self.object_name, self.multi_upload_id, self.queue),
                          StreamUploader(self.object_storage_client, self.namespace_name, self.bucket_name,
                                         self.object_name, self.multi_upload_id, self.queue),
                          StreamUploader(self.object_storage_client, self.namespace_name, self.bucket_name,
                                         self.object_name, self.multi_upload_id, self.queue)]
        for u in self.uploaders:
            u.daemon = True
            u.start()
            LOG.debug('uploader [%s] started' % str(u))

        LOG.debug('workers started, joining them')

        one_uploader_failed = False
        for u in self.uploaders:
            u.join()
            if not u.is_upload_successful():
                one_uploader_failed = True
            LOG.debug('uploader [%s] joined, ok? %s ' % (str(u), u.is_upload_successful()))

        # upload should be over now
        LOG.debug('all uploaders joined')
        try:
            self._finish_upload()
            if one_uploader_failed:
                self.long_task.end(TaskStatus(TaskStatusCode.FAILED,
                                              'backup failed: uploaed has failed'))
            else:
                self.long_task.end(TaskStatus(TaskStatusCode.SUCCESS,
                                              'backup completed'))
        except Exception, e:
            self.long_task.end(TaskStatus(TaskStatusCode.FAILED,
                                          'backup failed: %s' % str(e)))

    def _finish_upload(self):
        """
        Finish the upload
           1. shutdown the streamQueue
           2. commit the multipart upload
        """
        if self.queue:
            while not self.queue.empty():
                self.queue.get(True).cleanup()
            self.queue = None

        got_error = [u for u in self.uploaders if not u.is_upload_successful()]
        if len(got_error) > 0:
            LOG.debug('errors raised in uploaders, aborting the upload')
            self.object_storage_client.abort_multipart_upload(self.namespace_name,
                                                              self.bucket_name,
                                                              self.object_name,
                                                              self.multi_upload_id)
            LOG.debug('upload aborted')
        else:
            mpud = oci.object_storage.models.CommitMultipartUploadDetails()
            mpud.parts_to_commit = []
            mpud.parts_to_exclude = []
            for u in self.uploaders:
                mpud.parts_to_commit.extend(u.get_upload_details())
            try:
                self.object_storage_client.commit_multipart_upload(self.namespace_name,
                                                                   self.bucket_name,
                                                                   self.object_name,
                                                                   self.multi_upload_id,
                                                                   mpud)
                LOG.debug('upload committed')
            except OCIServiceError, e:
                raise StandardError('upload commit failed: %s' % e.message)


def run_all_test():
    pass


if __name__ == '__main__':
    set_log_config()

    run_all_test()
