# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
"""
Host helper for OCI service API configuration
"""

import configparser
import os
import oci
import oci.config
import oci.identity
from common import OCI_INSTANCE_SHAPE


class OCIConfigParser(configparser.SafeConfigParser):
        """
        Sub class of ConfigParser specialized to OCI configuration
        """

        LOCAL_USER = 'root'
        LOCAL_USER_HOME = os.path.join('/', LOCAL_USER)

        LOCAL_CONF_HOME = os.path.join(LOCAL_USER_HOME, '.oci')

        LOCAL_KEY_FILE = os.path.join(LOCAL_CONF_HOME, 'bmcs_api_key.pem')

        CONFIG_FILE = os.path.join(LOCAL_CONF_HOME, 'config')

        SECTION = 'DEFAULT'

        USER_OPT = 'user'
        TENANCY_OPT = 'tenancy'
        REGION_OPT = 'region'
        FINGERPRINT_OPT = 'fingerprint'

        KEY_FILE_OPT = 'key_file'
        KEY_FILE_PASS_OPT = 'pass_phrase'

        def __init__(self):
                configparser.SafeConfigParser.__init__(
                    self,
                    {
                        OCIConfigParser.KEY_FILE_OPT: OCIConfigParser.LOCAL_KEY_FILE,
                        OCIConfigParser.KEY_FILE_PASS_OPT: '',
                        OCIConfigParser.USER_OPT: '',
                        OCIConfigParser.TENANCY_OPT: '',
                        OCIConfigParser.REGION_OPT: '',
                        OCIConfigParser.FINGERPRINT_OPT: ''
                    })

        def load(self):
                """
                Loads from local config file
                """
                self.read(OCIConfigParser.CONFIG_FILE)

        def flush(self):
                """
                Write configuration to persistence
                """
                if not os.path.exists(OCIConfigParser.LOCAL_CONF_HOME):
                        os.makedirs(OCIConfigParser.LOCAL_CONF_HOME, 0600)

                _w = None
                try:
                    _w = open(OCIConfigParser.CONFIG_FILE, 'w')
                    self.write(_w)
                except IOError, e:
                    raise StandardError('cannot write configuration: %s' % str(e))
                finally:
                    if _w:
                        _w.close()

        def to_safe_map(self):
                """
                Returns a dictionary with configuration elements
                without any security information
                """
                result = dict(self.items(OCIConfigParser.SECTION))
                del result[OCIConfigParser.KEY_FILE_OPT]
                del result[OCIConfigParser.KEY_FILE_PASS_OPT]
                return result

        def validate(self):
                """
                1. Validate configuration against SDK requirement to make sure the format of config file is correct
                2. Validate by querying user info to make sure the content of the config file is correct
                """
                configDict = dict(self.items(OCIConfigParser.SECTION))
                try:
                        oci.config.validate_config(configDict)
                except oci.exceptions.InvalidConfig, e:
                        raise StandardError(str(e))
                try:
                        client = oci.identity.IdentityClient(config=configDict)
                        client.get_user(configDict[OCIConfigParser.USER_OPT])
                except Exception, e:
                        raise StandardError(str(e))

        def set_option(self, option, value):
                self.set(OCIConfigParser.SECTION, option, value)


class OCIInstanceHelper():

    def get_oci_shape(self):
        """
        Returns a name of OCI shape, such as 'BM.DenseIO1.36', or
        raises an error, mostly due to timeout
        """
        import urllib2
        try:
            response = urllib2.urlopen(OCI_INSTANCE_SHAPE, timeout=1)
        except urllib2.URLError as e:
            # most likely due to not in OCI
            raise StandardError("Failed to get OCI shape: %s" % str(e))
        return response.read()
