# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
"""
Raid Util:
This module manages all the raid retrieval functionalities
Usage:
ra = RaidManager()
ra.read_mdstat()
array = ra.get_raid_array_by_name("md0")

get raid state(string)

ra = RaidManager()
ra.read_mdstat()
array = ra.get_raid_array_by_name("md0")
return array.get_current_raid_status()
"""
import re
import cStringIO
import json
from backend.common import MDSTAT_PROC, MDADM
from backend.log_util import get_logger, set_log_config
from backend.lvm_util import get_volume_group_stats
from backend.command_util import command_output_lines

LOG = get_logger('lsa.raid_util')


def get_max_fault_by_level(level):
    """
    Get the maximum number of faulty devices a raid can tolerate
    parameter:
    level: int
        raid level. example: 0-6, 10 etc
    output: integer
    """

    if level == RaidLevel.LEVEL_0:
        return RaidMaxFault.MAX_RAID0_FAULTY_DEVS
    if level == RaidLevel.LEVEL_1:
        return RaidMaxFault.MAX_RAID1_FAULTY_DEVS
    if level == RaidLevel.LEVEL_5:
        return RaidMaxFault.MAX_RAID5_FAULTY_DEVS
    if level == RaidLevel.LEVEL_6:
        return RaidMaxFault.MAX_RAID6_FAULTY_DEVS
    if level == RaidLevel.LEVEL_10:
        return RaidMaxFault.MAX_RAID10_FAULTY_DEVS
    raise ValueError("raid level unsupported")


def is_raid_array_member(dev_name):
    """
    Checks that a device is member of an array
    or used to be a member of.
    Returns : True, if device is a member. False otherwise
    """
    lines = command_output_lines([MDADM, '--examine', '--brief', dev_name])
    return len(lines) != 0


def get_raid_state(state):
    """
    Get the readable state
    parameter:
    state: integer
        0, 1, 2, 4
    output: string
    """
    states = ["unavailable", "active", "degraded", "unknown", "faulty"]
    return states[state]


class RaidLevel:
    ''' Raid level related flag '''
    LEVEL_0 = 0
    LEVEL_1 = 1
    LEVEL_5 = 5
    LEVEL_6 = 6
    LEVEL_10 = 10


class RaidState:
    """ Raid state related FLAG """
    # Raid not setup or not found
    RAID_STATE_UNAVAILABLE = 0b0000
    # All the devices are active, no faulty device
    RAID_STATE_ACTIVE = 0b0001
    # Have faulty device but still functional
    RAID_STATE_DEGRADED = 0b0010
    # Failed state. Number of faulty device is more than tolerable
    RAID_STATE_FAILED = 0b0100
    # We only treat function state if it's either in active clean or active degraded
    RAID_STATE_FUNCTIONAL_MASK = 0b0011


class RaidMaxFault:
    """ Constants for maximum faulty device a raid can tolerate"""

    # Maximum number of faulty devices raid0 can tolerate
    MAX_RAID0_FAULTY_DEVS = 0
    # Maximum number of faulty devices raid1 can tolerate
    MAX_RAID1_FAULTY_DEVS = 1

    # Maximum number of faulty devices raid5 can tolerate
    MAX_RAID5_FAULTY_DEVS = 1
    # Maximum number of faulty devices raid6 can tolerate
    MAX_RAID6_FAULTY_DEVS = 2
    # Maximum number of faulty devices raid10 can tolerate. In our
    # case we configure raid 10 for highIO shapes. So maximum is 2.
    MAX_RAID10_FAULTY_DEVS = 2


class RaidManager(object):
    # match   md0 : active raid[0-6]  or raid 10
    ARRAY_START_RE = re.compile('^(md\d+) : (\w+) (raid)(\d{1,2})(\s+)')
    # match 125730816 blocks super 1.2 level 5, 512k chunk, algorithm 2 [4/3] [UU_U]
    ARRAY_STATS_RE = re.compile('^(\s*)(\d+) blocks (.+?)(\[)(\d+)(\/)(\d+)(\]) (\[)([U_]+)(\])')
    # match 584820224 blocks super 1.2 512k chunks
    ARRAY_STATS_RAID0_RE = re.compile('^(\s*)(\d+) blocks (.+?) chunks')

    def __init__(self, source=MDSTAT_PROC):
        """
        Instantiate a new RaidManager class
        param:
          source : source file (default=/proc/mdstat)
        """
        self.filename = source
        self.raid_arrays = []

    def _parse(self, data):
        """
        we look for group of lines like whats in /proc/mdstat
        raid 6:
            md0 : active raid6 sdf1[0] sde1[1] sdd1[2] sdc1[3] sdb1[4] sda1[5] hdb1[6]
                  1225557760 blocks level 6, 256k chunk, algorithm 2 [7/7] [UUUUUUU]
        raid 10:
            md0 : active raid10 sdb[0] sdd[2] sdc[1] sde[4]
                  83820544 blocks super 1.2 512K chunks 2 near-copies [4/4] [UUUU]

        i.e a block which start with md[0-10]*

        returns all array found
        """
        allArraysFound = []
        nextLineContainStats = False
        for line in data:
            match = RaidManager.ARRAY_START_RE.match(line)
            if match:
                # array size is given by next line
                nextLineContainStats = True
                array = RaidArray(match.group(1), match.group(2), int(match.group(4)))
                deviceTokens = line[match.end():].split(' ')
                for deviceToken in deviceTokens:
                    device = RaidBackDev.create_object_from_string(deviceToken)
                    if device:
                        array._add_device(device)
            else:
                if nextLineContainStats:
                    statsMatch = RaidManager.ARRAY_STATS_RE.match(line)
                    statsMatch0 = RaidManager.ARRAY_STATS_RAID0_RE.match(line)
                    if statsMatch:
                        # size reported in 1 KB block
                        size = int(statsMatch.group(2)) * 1024
                        num_raid_devs = int(statsMatch.group(5))
                        deviceStates = statsMatch.group(10)
                        num_active_devs = deviceStates.count('U')
                        num_failed_devs = deviceStates.count('_')
                        array._set_num_raid_devs(num_raid_devs)
                        array._set_num_active_devs(num_active_devs)
                        array._set_num_failed_devs(num_failed_devs)
                        array._set_size(size)
                        # now we can add it to the list
                        allArraysFound.append(array)
                    elif statsMatch0:
                        # size reported in 1 KB block
                        size = int(statsMatch0.group(2)) * 1024
                        num_raid_devs = 1
                        deviceStates = 'degraded'
                        num_active_devs = 1
                        num_failed_devs = 0
                        array._set_num_raid_devs(num_raid_devs)
                        array._set_num_active_devs(num_active_devs)
                        array._set_num_failed_devs(num_failed_devs)
                        array._set_size(size)
                        # now we can add it to the list
                        allArraysFound.append(array)
                    # go back to normal state, looking for next array
                    nextLineContainStats = False

        return allArraysFound

    def read_mdstat(self):
        """
        Load RaidArray object from source file
        returns self
        raises:
           Exception : error raised while reading the mdstat pseudo file
        """
        try:
            f = open(self.filename)
        except (OSError, IOError) as e:
            raise Exception("Failed to open the proc file:" + self.filename + " " + str(e))
        else:
            with f:
                self.raid_arrays = self._parse(f.readlines())
        return self

    def get_raid_array_by_name(self, raid):
        """
        return RaidArray object based on the raid device name(md0 or so)
        If nothing present return None
        parameter:
        raid : string
            raid device name
        output: RaidArray
        raise: LookupError if array not found
        """
        for ra in self.raid_arrays:
            if ra.dev == raid:
                return ra
        raise LookupError('Raid array not found')

    def get_raid_array_names(self):
        """
        returns all array names as list of string
        can be empty list
        """
        return [a.dev for a in self.raid_arrays]


class RaidArray(object):

    def __init__(self, dev, state, raidLevel):
        """
        We are maintaining three list here active, spare and faulty.
        and three other comma separated string represents the devices
        in each group active, spare, and faulty.
        """
        self.devices = []
        self.dev = dev
        # size in B
        self.size = 0
        if raidLevel not in [0, 1, 4, 5, 6, 10]:
            raise ValueError('invalid raid type [%s]' % raidLevel)
        self.level = raidLevel
        if state not in ['active', 'faulty']:
            raise ValueError('Invalid state [%s]' % state)
        self.status = state
        # Volume group usage (vg_size, vg_free as a dictionary)
        # {'vg_size': xxx, 'vg_free': xxx} in bytes
        self.vg_stats = None

    def get_devices(self):
        return self.devices

    def get_faulty_devices(self):
        return [device for device in self.devices if device.state == RaidBackDev.STATE_FAULTY]

    def get_active_devices(self):
        return [device for device in self.devices if device.state == RaidBackDev.STATE_ACTIVE]

    def get_spare_devices(self):
        return [device for device in self.devices if device.state == RaidBackDev.STATE_SPARE]

    def get_removed_devices(self):
        return [device for device in self.devices if device.state == RaidBackDev.STATE_REMOVED]

    def _set_size(self, size):
        # TODO : add assertion
        self.size = size

    def get_raid_level(self):
        '''
        Get the level of the raid with raid prefix
        Output: string, example: raid0, raid6, raid10
        '''
        return 'raid ' + str(self.level)

    def _set_num_raid_devs(self, raid_devs):
        """
        Set actual number of devices the raid was configured with
        parameter:
        raid_devs: int
            number of raid backed devices
        """
        self.num_raid_devs = raid_devs

    def _set_num_active_devs(self, active_devs):
        """
        Set the number of active device the raid currently has
        parameter:
        active_devs: int
            number of active devices
        """
        self.num_active_devs = active_devs

    def _set_num_failed_devs(self, failed_devs):
        """
        Set the number of failed devices (non active)
        parameter:
        failed_devs: int
            number of failed devices
        """
        self.num_failed_devs = failed_devs

    def retrieve_vg_stats(self):
        # must call with root
        self.vg_stats = get_volume_group_stats()

    def set_vg_stats(self, vg_stats):
        '''
        Set volume group stats as dictionary
        format: {'vg_size': xxxx, 'vg_free': xxxx}
        All the units in bytes
        '''
        self.vg_stats = vg_stats

    def get_vg_stats(self):
        '''
        Get volume group stats as dictionary
        format: {'vg_size': xxxx, 'vg_free': xxxx}
        All the units in bytes
        '''
        return self.vg_stats

    def is_functional(self):
        """
        Check if the raid array is actively running/functional.
        This function checks if the raid configuration
        can work or not with some faulty devices
        """
        return self.get_raid_state() & RaidState.RAID_STATE_FUNCTIONAL_MASK

    def get_raid_state(self):
        """
        Return the state of the raid array.
        parameter:
        output:
        0 : state unavailable
        1 : active clean
        2 : degraded
        3 : failed/faulty
        for the flag, see RaidState class
        """
        # Count the number of devices of each category
        num_faulty_devs = self.num_raid_devs - self.num_active_devs

        # Max faulty devices we can support for this raid array
        max_supported = get_max_fault_by_level(self.level)

        if (self.level == 'raid0' and self.num_raid_devs == 1):
            return RaidState.RAID_STATE_DEGRADED

        if num_faulty_devs == 0:
            return RaidState.RAID_STATE_ACTIVE
        if num_faulty_devs > max_supported:
            LOG.error("Number of faulty device(%s) exceeded maximum(%s) supported for %s" %
                      (num_faulty_devs, max_supported, self.level))
            return RaidState.RAID_STATE_FAILED

        if num_faulty_devs <= max_supported:
            return RaidState.RAID_STATE_DEGRADED

        return RaidState.RAID_STATE_UNAVAILABLE

    def get_current_raid_status(self):
        state = self.get_raid_state()
        return get_raid_state(state)

    def _add_device(self, device):
        """
        Add a backend device to the device list.
        parameter:
        device: RaidBackDev type object
        output: None or raise exception for unsupported object.
        """
        if not isinstance(device, RaidBackDev):
            raise ValueError('should be instance of RaidBackDev')
        self.devices.append(device)

    def __str__(self):
        output = cStringIO.StringIO()
        output.write('[name=')
        output.write(self.dev)
        output.write(', state=')
        output.write(self.status)
        output.write(', level=')
        output.write(self.level)
        output.write(', size(B)=')
        output.write(str(self.size))
        output.write(', device=')
        for d in self.devices:
            output.write(str(d))
            output.write(',')
        output.write(']')

        line = output.getvalue()
        output.close()

        return line


class RaidBackDev(object):
    """
    Raid back device,
    Hold the information of each NVMe device
    """

    #                   match sdb[0-9](W|S|F|R)
    ARRAY_MEMBER_RE = re.compile('(\w+)\[(\d+)\](\([WSFR]\))?')

    STATE_FAULTY = "faulty"
    STATE_ACTIVE = "active"
    STATE_SPARE = "spare"
    STATE_REMOVED = "removed"

    @classmethod
    def _get_device_state_from_flag(cls, flag):
        if flag is None or flag == 'active':
            return 'active'
        if flag == 'F' or flag == 'faulty':
            return 'faulty'
        if flag == 'R' or flag == 'removed':
            return 'removed'
        if flag == 'S' or flag == 'spare':
            return 'spare'
        if flag == 'W':
            return 'write-mostly'

    def __init__(self, name, index, state='active'):
        # TODO : add assertion
        self.name = name
        self.index = int(index)
        self.state = RaidBackDev._get_device_state_from_flag(state)

    @classmethod
    def create_object_from_string(cls, buffer):
        """
        instantiate a new device from line like sdd[2](F)
        """
        match = RaidBackDev.ARRAY_MEMBER_RE.match(buffer)
        if match:
            # for the flag we remove parenthesis:  '(F)' -> 'F'
            if match.group(3):
                statusFlag = match.group(3)[1:-1]
            else:
                statusFlag = None
            rad = cls(match.group(1), match.group(2), statusFlag)
            return rad
        return None

    def serialize(self):
        '''
        Represent the object as json format
        '''
        return json.dumps({'name': self.name,
                           'index': self.index,
                           'state': self.state
                           })

    @classmethod
    def deserialize(cls, js):
        '''
        Build an object from json string
        '''
        js = json.loads(js)
        return cls(js.get('name'), js.get('index'), js.get('state'))

    def __eq__(self, other):
        """
        Check for equality by testing name equality
        name uniquely identify a disk
        """
        if id(self) == id(other):
            return True
        if other is None:
            return False

        if isinstance(other, RaidBackDev):
            return self.name == other.name

        return False

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return "RaidBackDev {dev=%s, index=%s, state=%s}" % (self.name, self.index, self.state)


if __name__ == '__main__':
    set_log_config()
