# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
'''
This module manages shares and acts as a service layer for exports  managements.
usage:
import share_util

# get list of shares
shares = share_util.get_shares()
# get list of exports
exports = ShareEntry(share_name).get_exports()

'''
from validation import is_valid_file_name
import os
import pwd
import grp
import shutil
import StringIO
import json
import threading

import taskmgr
from taskmgr import (TaskStatusCode, TaskStatus)
from common import (SHARE_DIR_USER_OWNERSHIP, SHARE_DIR_GROUP_OWNERSHIP,
                    SHARE_DIR_ACCESS_RIGHT, LV_DEFAULT_SECTOR_SIZE,
                    SHARE_POINT
                    )
from utility import get_full_mount_path, get_full_lv_path, parse_boolean
from fstab import FstabManager
from lvm_util import (add_lv, remove_lv, remove_lv_fs_signature, get_lv_stats,
                      extend_lv, list_lvs_by_vg)
from fs_util import mount_lv, umount_lv, format_lv_with_xfs, resize_xfs, get_lv_fs, is_lv_mounted, get_share_stat
from log_util import get_logger, set_log_config
from option_util import BooleanOption

LOG = get_logger('lsa.share')

# list of logical volumes not considered as customer share
_blacklisted_names = ['__lsa']


def get_shares(names=[]):
    '''
    List the shares. If caller does not provide any parameter retrieve
    all the shares otherwise filter the list with provided names
    parameter:
        names: use an array as parameter ( list of share names)
    returns an array of ShareEntry. can be empty
    '''

    out = list_lvs_by_vg()
    shares = []
    if not out:
        return shares
    lines = out.split('\n')
    for line in lines:
        line = line.strip('')
        (name, size) = line.split()
        if names and name not in names:
            continue
        if name in _blacklisted_names:
            continue
        fs_stat = get_share_stat(name)
        share = ShareEntry(name, fs_stat['mount_path'])
        share.size = int(size)
        if fs_stat['fs'] != 'xfs':
            # may happen when share is in middle of creation
            # we cannot trust fs_stat['free'], fallback to size
            share.free = share.size
        else:
            share.free = fs_stat['free']
        share.inode32 = FstabManager.get_inode32_for_share(name)
        shares.append(share)
    return shares


class ShareEntry(object):
    '''
    Class for representing a share
        member variables: name, path, inode32, size, free
        name: name of the share
        path: full path or dir
        size: size of the share in bytes
        free: free space available in the share(basically logical volume)
    '''
    def __init__(self, name, path=None):

        # validate name / path and IP
        if path is None:
            path = os.path.join(SHARE_POINT, name)
        if not is_valid_file_name(path):
            raise ValueError("Invalid path: " + path)
        if not is_valid_file_name(name):
            raise ValueError("Invalid name: " + name)
        if name not in path:
            raise ValueError("Name must be in the path")
        self.name = name
        self.path = path
        self.inode32 = False
        self.size = 0
        self.free = 0

    def get_path(self):
        """
        Get full path of the nfs share
        /shares/dir
        """
        return self.path

    def get_name(self):
        """
        Gets the export entry name
        """
        return self.name

    def get_size(self):
        '''
        Get size of the share
        '''
        return self.size

    def get_inode32(self):
        '''
        Return true if the option inode32 is set for this entry
        '''
        return self.inode32

    def get_free(self):
        '''
        Get free/available space in human readable format
        '''
        return self.free

    def serialize(self):
        '''
        JSON representation of the object
        output: dictionary
        '''
        return json.dumps({
            'name': self.get_name(),
            'size': self.get_size(),
            'free': self.get_free(),
            'path': self.get_path(),
            'fs_options': {'inode32': self.get_inode32()}
        })

    def __str__(self):
        '''
        String representation of the object
        '''
        output = StringIO.StringIO()
        output.write('name: ' + self.name + '\n')
        output.write('path: ' + self.path + '\n')
        output.write('size: ' + str(self.size) + '\n')
        output.write('free: ' + str(self.free) + '\n')
        output.write('fs_options: { inode32: ' + str(self.inode32) + '}')
        contents = output.getvalue()
        output.close()
        return contents

    @classmethod
    def deserialize(cls, js):
        '''
        Create an object from JSON representation of the object
        '''
        deserializeJS = json.loads(js)
        share = cls(deserializeJS.get('name'), deserializeJS.get('path'))
        share.size = deserializeJS.get('size')
        share.free = deserializeJS.get('free')
        fs_opt = deserializeJS.get('fs_options')
        if fs_opt:
            share.inode32 = fs_opt.get('inode32')
        return share


class AsyncShareCreateTask(threading.Thread, taskmgr.Task):
    """
    Create an asynchronous task for helping create a share
    usage:
        task = backend.share_util.AsyncShareCreateTask(name, size, inode32)
        backend.taskmgr.get_task_manager().register_task(task)
        task.begin()
        return task.get_id()
    """

    def __init__(self, share_name, share_size, inode32):
        """
        create a new task for creating the directory for exporting NFS or configuring samba.
        This is the backup storage for a share either nfs or samba or both.
        parameter:
            share_name: name of the share
            share_size: size of the share in bytes
            inode32: true if the inode32 option need to be set to workaround
                     legacy application limitation
        """
        threading.Thread.__init__(self, name="__share_create_%s" % share_name)
        taskmgr.Task.__init__(self, share_name, 'share_create')
        self.share_name = share_name
        self.share_size = share_size
        self.inode32 = inode32
        self.dir_created = False
        self.lv_added = False
        self.lv_formatted = False
        self.lv_mounted = False
        self.fstab_updated = False

    def _cleanup(self):
        """
        undo everything done so far, in reverse order
        this private method 'never fails', as it catches and logs all exceptions
        """
        if self.fstab_updated:
            try:
                FstabManager.remove_fstab_entry(self.share_name)
            except Exception as e:
                LOG.error("Failed to remove fstab entry for share %s: %s" % (self.share_name, str(e)))
        if self.lv_mounted:
            try:
                umount_lv(self.share_name)
            except Exception as e:
                LOG.error("Failed to umount lv for share %s: %s" % (self.share_name, str(e)))
        if self.lv_formatted:
            try:
                remove_lv_fs_signature(self.share_name)
            except Exception as e:
                LOG.error("Failed to remove fs signature for share %s: %s" % (self.share_name, str(e)))
        if self.lv_added:
            try:
                remove_lv(self.share_name)
            except Exception as e:
                LOG.error("Failed to remove lv for share %s: %s" % (self.share_name, str(e)))
        if self.dir_created:
            try:
                shutil.rmtree(self.path)
            except Exception as e:
                LOG.error("Failed to remove directory for share %s: %s" % (self.share_name, str(e)))

    def _cleanup_end_task(self):
        self._cleanup()
        self.end(TaskStatus(TaskStatusCode.FAILED, 'share creation failed'))

    def begin(self):
        """
        begin the task started by AsyncShareCreateTask by
            calling parent class Task's begin(),
            changing the task's state to STARTED, and
            calling parent class Thread's start(), which calls run()
        """
        LOG.debug("to call taskmgr.Task.begin")
        taskmgr.Task.begin(self)
        self._add_state(taskmgr.TaskStateChange(taskmgr.TaskState.STARTED))
        LOG.debug("started the task")
        self.start()
        LOG.debug("started the thread")

    def run(self):
        """
        called by Thread.start(), and run in a new thread
        failure might be caused by: - not enough space available
                                    - share directory already exists on disk
                                    - not able to change ownership/mode of directory
        """
        self.path = get_full_mount_path(self.share_name)
        try:
            os.makedirs(self.path)
        except OSError as e:
            LOG.error("Failed to create the directory %s: %s" % (self.path, str(e)))
            self.end(TaskStatus(TaskStatusCode.FAILED, 'share creation failed'))
            return
        self.dir_created = True
        LOG.debug('created dir [%s]' % self.path)

        LOG.debug('Adding lv name: %s, size %s' % (self.share_name, self.share_size))
        try:
            add_lv(self.share_name, self.share_size)
            self.lv_added = True
        except Exception as e:
            LOG.error("Failed to add the lv for share %s: %s" % (self.share_name, str(e)))
            return self._cleanup_end_task()
        LOG.debug("added lv")

        try:
            format_lv_with_xfs(self.share_name)
            self.lv_formatted = True
        except Exception as e:
            LOG.error("Failed to format the lv for share %s: %s" % (self.share_name, str(e)))
            return self._cleanup_end_task()
        LOG.debug("formatted lv")

        inode32_opt = BooleanOption('inode32').disable()
        if self.inode32 is not None and parse_boolean(self.inode32):
            inode32_opt.enable()
        try:
            mount_lv(self.share_name, inode32_opt.enabled())
            self.lv_mounted = True
        except Exception as e:
            LOG.error("Failed to mount the lv for share %s: %s" % (self.share_name, str(e)))
            return self._cleanup_end_task()
        LOG.debug("mounted lv")

        try:
            FstabManager.add_fstab_entry(self.share_name, inode32_opt.enabled())
            self.fstab_updated = True
        except Exception as e:
            LOG.error("Failed to add fstab entry for share %s: %s" % (self.share_name, str(e)))
            return self._cleanup_end_task()
        LOG.debug("added fstab entry")

        try:
            os.chown(self.path,
                     pwd.getpwnam(SHARE_DIR_USER_OWNERSHIP).pw_uid,
                     grp.getgrnam(SHARE_DIR_GROUP_OWNERSHIP).gr_gid
                     )
        except OSError as e:
            LOG.error("Failed to change the directory ownership for share %s: %s" % (self.share_name, str(e)))
            return self._cleanup_end_task()

        try:
            os.chmod(self.path, SHARE_DIR_ACCESS_RIGHT)
        except OSError as e:
            LOG.error("Failed to change the directory mode for share %s: %s" % (self.share_name, str(e)))
            return self._cleanup_end_task()

        self.end(TaskStatus(TaskStatusCode.SUCCESS, 'share creation completed'))

    def end(self, status):
        """
        end the task started by AsyncShareCreateTask
        parameter:
            status: operation status as TaskStatus
        """
        LOG.debug("to call taskmgr.Task.end")
        taskmgr.Task.end(self, status)
        self._add_state(taskmgr.TaskStateChange(taskmgr.TaskState.STOPPED))
        LOG.debug("ended the task")

    def abort(self):
        """
        abort the task, i.e., abort the timer
        """
        self._cleanup()
        taskmgr.Task.abort(self)
        self._add_state(taskmgr.TaskStateChange(taskmgr.TaskState.ABORTED))
        LOG.debug("aborted the task")


def resize_share(share_name, new_share_size):
    '''
    Expand logical volume and file system for a share when user wants to increment it's size.
    parameter:
        share_name: name of the share as string
        new_share_size: new size of the share in bytes
    raise: IOError, ValueError
    '''
    lvstat = get_lv_stats(share_name)
    old_size = lvstat.get('lv_size')
    target_incr = new_share_size - old_size
    LOG.debug('Old size %s New Size %s' % (str(old_size), str(new_share_size)))

    if target_incr == 0:
        return
    if target_incr < 0:
        raise IOError('Reducing the share size is not supported')
    LOG.info('Target incr %s, g_free %s' % (target_incr, lvstat.get('vg_free')))

    # Target increment of size in bytes must be at least 512 bytes (lvextend limitation)
    if target_incr > 0 and target_incr < LV_DEFAULT_SECTOR_SIZE:
        LOG.error('Difference between old and new size must be greater than %d' % LV_DEFAULT_SECTOR_SIZE)
        raise ValueError('Difference between old and new size must be greater than %d' % LV_DEFAULT_SECTOR_SIZE)

    if int(lvstat.get('vg_free')) < target_incr:
        raise IOError('Insufficient  storage on the volume group(lsa). Please clean some space and try again.')

    # check if the LV is eligible for an extension
    if target_incr >= LV_DEFAULT_SECTOR_SIZE:
        extend_lv(share_name, new_share_size)
        # Now resize xfs
        # check if xfs file system exists
        if get_lv_fs(share_name):
            resize_xfs(share_name)


def share_available(share_name):
    """
    If the storage for a share is available ie. Logical volume, fstab entry exists for
    this share and the share is mounted
    parameter:
        share_name: name of the share as string
        output: True or False
    """
    mp = get_full_mount_path(share_name)
    if not os.path.exists(mp):
        return False
    lv_path = get_full_lv_path(share_name)
    if not os.path.exists(lv_path):
        return False
    if not FstabManager.share_exist_in_fstab(share_name):
        return False
    if not is_lv_mounted(share_name):
        return False

    return True


def delete_share(share_name):
    """
    Delete the directory from back end. This should be done after removing the entry from
    /etc/exports
    parameter:
        share_name: name of the share as string
        output: nothing
        raise: raised by higher level ( by caller)
    """
    # umount the dir first
    umount_lv(share_name)
    # remove the fstab entry from /etc/fstab file
    FstabManager.remove_fstab_entry(share_name)
    # remove FS signature from LV
    remove_lv_fs_signature(share_name)
    # remove the logical volume
    remove_lv(share_name)
    path = get_full_mount_path(share_name)
    # remove the directory
    if os.path.exists(path):
        shutil.rmtree(path)


if __name__ == '__main__':
    set_log_config()
