# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
"""
ssh authorized key utility module:
this module provide helper functions to retrieve, validate and write ssh public keys to the authorized_keys file.
We can assume the opc user is in the LSA server, and user and home directory
are already there.
Usage:
#update public ssh keys:
import ssh_util
# Validate the key
ssh_util.validate_pub_ssh_key(ssh_key)
# Check for duplicate key
ssh_util.ssh_key_exist(ssh_key)
# each line contains a ssh key
ssh_util.update_ssh_keys(ssh_keys)
"""


import os
import shutil
from backend.common import OPC_AUTHORIZED_KEY_FILE, BAK_OPC_KEY_FILE, TEMP_SSH_KEY_FILE
from backend.log_util import get_logger, set_log_config
import base64
import struct

LOG = get_logger('lsa.ssh_util')


def get_ssh_keys():
    '''
    Get the ssh keys from authorized_key file
    raise: IOError on failed to read the key file
    '''
    keys = ''
    if not os.path.exists(OPC_AUTHORIZED_KEY_FILE):
        return keys
    try:
        fo = open(OPC_AUTHORIZED_KEY_FILE, 'r')
        keys = fo.read()
    except IOError as e:
        raise StandardError('Failed to read key file: %s' % str(e))
    finally:
        if fo:
            fo.close()
    return keys


def _validate_ssh_pub_key(ssh_pubkey):
    '''
    Validate a public ssh key, if it can be decoded and meets certain criteria.
        The data portion of the keyfile is decoded from base64, or it fails
        with a base64.binascii.Error
        The first 4 bytes are unpacked (an int), and returns 7, which is the
        length of the following string (I guess this could change, but we only
        support ssh-rsa which 7 characters).
        Ref: https://docs.moodle.org/dev/SSH_key,
             https://stackoverflow.com/questions/2494450/ssh-rsa-public-key-validation-using-a-regular-expression
     parameter:
        ssh_pubkey: public ssh key
        raise: Standard/Key error for failure.
    '''
    LOG.debug('ssh_pubkey = %s' % ssh_pubkey)
    key_type, key_string, comment = ssh_pubkey.split()

    try:
        data = base64.decodestring(key_string)
    except Exception as e:
        LOG.error('SSH key error, decode failed: %s' % str(e))
        raise StandardError('SSH key error, decode failed: %s' % str(e))
    try:
        int_len = 4
        # this should return 7
        str_len = struct.unpack('>I', data[:int_len])[0]
    except Exception as e:
        raise StandardError('Unpack failed')
    if int(str_len) != 7:
        raise KeyError('Invalid ssh key')

    decoded_type = data[int_len:int_len + str_len]
    if decoded_type != key_type:
        LOG.error('Invalid ssh key: key type and decoded key type are not same')
        raise KeyError('Invalid ssh key: key type and decoded key type are not same')


def _validate_ssh_pub_keys(ssh_pubkeys):
    '''
    Validate the ssy keys sepearted by new line
    param:
        ssh_pubkeys: keys as string
    raise: standard error if validation fails
    '''
    if len(ssh_pubkeys) == 0:
        return
    try:
        keys = ssh_pubkeys.split('\n')
    except Exception as e:
        LOG.error('Failed to split the keys: %s' % str(e))
        raise StandardError('Failed to split the keys: %s' % str(e))

    LOG.debug('Total Keys: %d' % len(keys))
    try:
        for key in keys:
            _validate_ssh_pub_key(key)
    except Exception as e:
        LOG.exception('Key validation failed')
        raise StandardError('Key validation failed: %s' % str(e))


def _write_ssh_keys(ssh_keys):
    '''
    Write the sshkeys into /home/opc/.ssh/authorized_key file
      param:
        ssh_key: public ssh key
        raise: standard error on failure
    '''
    bfo = None
    key_file = OPC_AUTHORIZED_KEY_FILE
    back_key_file = BAK_OPC_KEY_FILE
    tmp_file = TEMP_SSH_KEY_FILE
    if os.path.exists(tmp_file):
        os.remove(tmp_file)
    try:
        bfo = open(tmp_file, "w")
    except IOError:
        LOG.error("Failed to open the file for editing: %s", key_file)
        raise StandardError("Failed to open the file for editing: %s", key_file)
    try:
        bfo.write(ssh_keys)
        bfo.write("\n")
    except IOError as e:
        LOG.error("Failed to append ssh key: \n" + str(e))
        raise StandardError("Failed to append ssh key: %s" % str(e))
    finally:
        if bfo:
            bfo.close()

    if not os.path.exists(OPC_AUTHORIZED_KEY_FILE):
        try:
            open(OPC_AUTHORIZED_KEY_FILE, 'w').close()
        except IOError as e:
            LOG.error('Failed to create authorized key file: %s' % str(e))
            raise StandardError('Failed to create authorized key file: %s' % str(e))
    # create a backup file just in-case
    if os.path.exists(back_key_file):
        os.remove(back_key_file)

    shutil.copyfile(key_file, back_key_file)
    shutil.move(tmp_file, key_file)
    LOG.debug('Writing ssh keys done')


def update_ssh_keys(ssh_keys):
    '''
    Updating ssh keys to authorized_key file;
    validate keys, if validation and check
    succeeds then write all the keys to authorized key file
    parameter:
        ssh_keys: public ssh keys as string
        raise: standard standard error if validation of file operation fails
    '''
    # Remove leading and trailing new lines
    ssh_keys = ssh_keys.strip('\n')
    ssh_keys = ssh_keys.strip('\r')
    _validate_ssh_pub_keys(ssh_keys)
    _write_ssh_keys(ssh_keys)
    LOG.debug('Updating public ssh keys done')


if __name__ == '__main__':
    set_log_config()
