# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
'''
StoragePool class presents a standard interface to retrieve information regardless how the pool is created.
This class is not used to create the physical underlying storage pool itself. This is done in backend/init/storage.sh (See README for architecture details).
Usage:
object creation:
sp = StoragePool(size, free, status, devs)
or
sp = StoragePool.create_from_raid_array(ra)
or
sp = StoragePool.create_from_vg(vg_stats, disks)

serialized_obj = sp.serialize()

obj_again = StoragePool.deserialize(serialized_obj)

'''


from backend.log_util import get_logger, set_log_config
from backend.raid_util import RaidBackDev
import json

LOG = get_logger('lsa.storage_pool')


class StoragePool(object):
    '''
    size: Total size of the storage pool in bytes
    free: Total free space available in the storage pool in bytes
    status: Status of the storage pool, this status is meaningful if the disks are  mirrored with raid.
            For raid the values might be active, degraded, faulty. For block devices we set the value as active always
    devs: List of the devices present the storage pool, each item in the list is an object of RaidBackDev
    protection: data protection, for raid the value should be raid[1, 6, 10], for block storage this field is None
    num_raid_dev: number of raid devs used for initial raid creation
    '''
    def __init__(self, size, free, status, devs, protection=None, num_raid_devs=None):

        if not isinstance(size, int):
            raise ValueError('Size must be an integer')

        if not isinstance(free, int):
            raise ValueError('Free space must be an integer')

        if not devs:
            raise ValueError('List of devices must not be empty')

        self.size = size
        self.free = free
        self.status = status
        self.devices = devs
        self.protection = protection
        if num_raid_devs is not None:
            self.num_devs = num_raid_devs
        else:
            self.num_devs = len(self.devices)

    def get_devices(self):
        '''
        Get the devices of the storage pool
        Returns a list: each item in the list is an object of RaidBackDev
        '''
        return self.devices

    def _serialize_devices(self):
        '''
        Get the list of devices as json format
        format:["{"index": 0, "state": "active", "name": "sdc"}",
                "{"index": 1, "state": "active", "name": "sdd"}"
                "{"index": 2, "state": "active", "name": "sdf"}"
                "{"index": 3, "state": "active", "name": "sdf"}"
               ]
        '''
        devs = []
        for dev in self.get_devices():
            devs.append(dev.serialize())
        return devs

    def get_size(self):
        '''
        Get the size of the storage pool in bytes
        '''
        return self.size

    def get_free(self):
        '''
        Get the free space available in bytes
        '''
        return self.free

    def get_status(self):
        '''
        Get the status of the storage pool as string
        Sample values: active, faulty, degraded
        '''
        return self.status

    def get_protection(self):
        '''
        Get the data protection of the storage pool
        This function is valid in case of the storage is mirrored with raid
        '''
        return self.protection

    def get_num_devs(self):
        '''
        Get the number of devices associated with the storage pool
        '''
        return self.num_devs

    def serialize(self):
        '''
        Serialize the object as json format data
        '''
        return json.dumps({'size': self.get_size(),
                           'free': self.get_free(),
                           'status': self.get_status(),
                           'devices': self._serialize_devices(),
                           'protection': self.get_protection(),
                           'num_devs': self.num_devs
                           })

    @classmethod
    def deserialize(cls, js):
        '''
        Build the object from json format data
        Output: StoragePool object
        '''
        js = json.loads(js)
        devs = js.get('devices')
        devices = []
        for dev in devs:
            devices.append(RaidBackDev.deserialize(dev))
        return cls(js.get('size'), js.get('free'), js.get('status'), devices, js.get('protection'), js.get('num_devs'))

    @classmethod
    def create_from_raid_array(cls, ra):
        '''
        Create an object from RaidArray type object
        Returns: StoragePool object
        raise: ValueError if any one of the StoragePool construction parameters not valid
        '''
        vg_stats = ra.get_vg_stats()
        vg_size = int(vg_stats.get('vg_size'))
        vg_free = int(vg_stats.get('vg_free'))
        return cls(vg_size, vg_free, ra.get_current_raid_status(), ra.get_devices(), ra.get_raid_level(), ra.num_raid_devs)

    @classmethod
    def create_from_vg(cls, vg_stats, disks):
        '''
        Create StoragePool object from volume group information
        parameter:
            vg_stats: size and free in bytes, status as string in dictionary format: {'size': '123456789', 'free': '1234567', 'status': 'faulty/active'}
            disks: List of disks associated with the volume group. Format ['/dev/sdc', '/dev/sdd']
        raise: ValueError if any one of the StoragePool constructor parameters not valid
        '''
        devices = []
        index = 0
        for disk in disks:
            # default disk state is active , if anything wrong with the disk we mark it as faulty
            disk_state = 'active'
            if 'unknown' in str(disk):
                disk_state = 'faulty'

            dev = RaidBackDev(disk, index, disk_state)
            devices.append(dev)
            index = index + 1
        vg_size = int(vg_stats.get('vg_size'))
        vg_free = int(vg_stats.get('vg_free'))
        status = vg_stats.get('status')
        return cls(vg_size, vg_free, status, devices)


if __name__ == '__main__':
    set_log_config()
