# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
"""
module responsible to keep track of long running
tasks
Task manager will maintain a list of tasks  running or already completed
persistence is handled with a file : /opt/lsa/var/taskmgr.db

"""

import shutil
import json
import os
import os.path
import threading
from backend.log_util import get_logger, set_log_config
import backend.taskmgr
import backend.taskmgr.encoder

from backend.lsa_types import ExtJSONEncoder as LSAExtJSONEncoder


logger = get_logger('lsa.taskmgr.task_manager')

__all__ = ['TaskManager']


class TaskManager(object):
    """
    keep track of filenames as we want only one
    instance running per persistence file
    """
    __persistences = []
    __lock = threading.Lock()

    MAX_TASK_COUNT = 1000

    def __init__(self, filename):
        """
        Creates a new task manager
        parameter:
          filename: persistence file where to store historical data
        raise ValueError: one manager already exists
        """
        self.tasks = []
        with TaskManager.__lock:
            if filename in TaskManager.__persistences:
                raise ValueError('manager already exists for this persistence')
            TaskManager.__persistences.append(filename)

        self.backend_filename = filename

        self.timer = backend.ExtTimer(60.0, self.sync)
        self.timer.start()

    def __del__(self):
        """
        TaskManager destructor
        Stop the sync timer and sync to persistence
        """
        try:
            self.timer.stop()
            self.sync()
        except Exception, e:
            if logger is not None:
                # this is a destructor, some reference may be already lost
                logger.warning('sync has failed %s' % str(e))

    def load(self):
        """
        load persisted tasks from file
        raise:
          IOError: not able to access the file
          StandardError: something wrong during deserialization
        """
        logger.debug('persistence loaded from %s' % self.backend_filename)
        if not os.path.exists(self.backend_filename):
            logger.warning('no task db file')
            return
        with open(self.backend_filename, 'r') as f:
            for line in f:
                try:
                    (id, taskjsons) = line.split(':', 1)
                    _task = json.loads(taskjsons.strip(), object_hook=backend.taskmgr.encoder.task_decoder)
                    self.tasks.append(_task)
                except (KeyError, ValueError), e:
                    logger.warning('error during parsing of [%s]: %s' % (line, str(e)))
                    logger.error(' task db file seems corrupted')

        logger.debug('%d tasks loaded from persistence' % len(self.tasks))
        self.tasks.sort()

    def get_tasks(self):
        """
        Gets the list of  tasks
        Returns list of Task, can be empty.
        """
        return self.tasks

    def get_task(self, taskID):
        """
        Gets task by taskID
        parameters:
           taskID : id of the task to be retrieved
        Raise KeyError is id is not found
        """
        for t in self.tasks:
            if t.get_id() == taskID:
                return t

        raise KeyError('No such id %d' % taskID)

    def get_task_for_target(self, target_name):
        """
        Gets tasks list related to a target
        parameters:
           target_name : target name
        returns:
           list of tasks (can be empty)
        """
        return [t for t in self.tasks if t.get_target() == target_name]

    def sync(self):
        """
        Persist all tasks to file and purge all tasks
        We keep 1000 task in memory. file should not exceed 4MB
        Task are sorted by creation date.
        raise:
          IOError: not able to access the file
          StandardError: something wrong during deserialization
        """

        new_file = self.backend_filename + '.new'
        persist_ok = False
        with TaskManager.__lock:
            tasks_count = len(self.tasks)
            logger.debug('sync called for %d task(s)' % len(self.tasks))
            if tasks_count == 0:
                return
            if tasks_count > TaskManager.MAX_TASK_COUNT:
                # time to purge
                logger.info('%d registerd tasks, purging' % tasks_count)
                # tasks list is maintained sorted: keep the 100 youngest tasks
                new_task_list = self.tasks[:100]
                self.tasks = new_task_list
            f = None
            try:
                f = open(new_file, 'w')
                for t in self.tasks:
                    f.write('%d:' % t.get_id())
                    json.dump(t, f, cls=LSAExtJSONEncoder)
                    f.write('\n')
                    persist_ok = True
            except Exception, e:
                logger.warning('cannot persist tasks: %s' % e.args[0])
            finally:
                if f:
                    f.close()

        if persist_ok:
            try:
                shutil.copyfile(self.backend_filename, self.backend_filename + '.bck')
                shutil.move(new_file, self.backend_filename)
            except IOError, e:
                logger.warning('cannot update persistence file: %s' % str(e))

    def register_task(self, task):
        """
        register a new task to this manager
        raise :
           TypeError: if task not of right type
        """
        if not isinstance(task, backend.taskmgr.Task):
            raise TypeError('wrong type of object registered')

        with TaskManager.__lock:
            logger.debug('registering new task %d' % task.get_id())
            self.tasks.append(task)
            self.tasks.sort()
        # now persist the new task at the end of persistence file
        # task are persisted one per line using json encoder
        # each line is prefixed with task unique ID for fast retrieval
        f = None
        try:
            f = open(self.backend_filename, 'a')
            f.write('%d:' % task.get_id())
            json.dump(task, f, cls=LSAExtJSONEncoder)
            f.write('\n')
        except Exception, e:
            logger.warning('cannot persist a new task: %s' % e.args[0])
        finally:
            if f:
                f.close()

    def cancel_task(self, taskID):
        """
        Cancel a task
        In case task is still running , we cancel it
        parameter
           taskID : id of the task to be cancelled
        raise
           KeyError: if task not found
        """
        task = self.get_task(taskID)
        logger.debug('cancelling task %s' % str(task))
        task.abort()
        logger.debug('task %s cancelled' % str(task))


def run_all_test():

    class MyTask(backend.taskmgr.Task):
        def begin(self):
            backend.taskmgr.Task.begin(self)
            self.states.append(backend.taskmgr.TaskStateChange(backend.taskmgr.TaskState.STARTED))

    mgr = TaskManager('/tmp/taskmgr.out')

    mgr.load()
    t = MyTask('test')
    mgr.register_task(t)
    t.begin()
    mgr.sync()
    t2 = mgr.get_task(t.get_id())
    print t2
    for t in mgr.get_tasks():
        print '%s' % t
        print [str(x) for x in t.get_states()]
        print t.is_running()

    os.remove('/tmp/taskmgr.out')


if __name__ == '__main__':
    set_log_config()

    run_all_test()
