# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
"""
Utility functions(formatting numbers, size and others which can be used by
different modules.
"""
import re
import os
from backend.common import SHARE_POINT, ROOT_DEV, VG_NAME, CMD_HOSTNAME
from command_util import execute_command
import StringIO
from ConfigParser import SafeConfigParser


class CustomConfigParser(SafeConfigParser):
    '''
    Extended class of ConfigParser which deal with section or option
    within section has leading space or tabs.
    For example:
    [global]
        netbios-name = xxxx
    or
        [global]
    netbios-name = xxxx
    '''
    def read_file(self, file_name):
        if not os.path.exists(file_name):
            raise IOError('Config file(%s) not found.' % file_name)
        try:
            data = StringIO.StringIO('\n'.join(line.strip() for line in open(file_name)))
            self.readfp(data)
        except (OSError, IOError) as e:
            raise StandardError('Failed to read config file : %s' % str(e))


def get_hostname():
    """
    Get the hostname by linux hostname command
    """
    cmd = [CMD_HOSTNAME]
    out, err, code = execute_command(cmd)
    if code != 0:
        raise StandardError('Failed to get hostname: %s' % str(err))

    if out:
        out = out.strip()
        out = out.strip('\n')
        return out
    return "Hostname error"


def get_host_ip_address():
    """
    Get the host IP address. This function always returns the IP address from
    first interface.
    """
    cmd = [CMD_HOSTNAME, '-I']
    out, err, code = execute_command(cmd)
    if code != 0:
        raise StandardError('Failed to get hostname: %s' % str(err))
    if out:
        return out.split()[0]
    return "Host IP address error"


def get_hostname_short():
    """
    Get the hostname in short
    """
    cmd = [CMD_HOSTNAME, '-s']
    out, err, code = execute_command(cmd)
    if code != 0:
        raise StandardError('Failed to get hostname: %s' % str(err))
    if out:
        out = out.strip()
        out = out.strip('\n')
        return out
    return "Host name error"


def format_readable_to_bytes(readable_size, kilo=1024):
    '''
    Convert readable size to bytes.
      parameter:
        readable_size: string, example (10kb. 10kib, xxgb etc)
        output: size in bytes in integer
    '''
    # replace i with blank just in case anyone provides Mib GiB or so
    size = readable_size.replace('i', '')
    size = readable_size.lower()
    if not re.match('(\d+(?:\.\d+)?)([g|t|k|m]+[b]*)$', size, re.IGNORECASE):
        raise ValueError("Cannot interpret size, wrong format: %s" % readable_size)
    size = size.rstrip('b')

    units = ['k', 'm', 'g', 't']
    try:
        power = units.index(size[-1]) + 1
    except ValueError:
        power = 0
        size = float(size)
    else:
        size = float(size[:-1])
    powered = int(kilo) ** power
    size_in_bytes = size * float(powered)
    return int(size_in_bytes)


def format_bytes_to_readable(size):
    '''
    covert bytes to readable
      parameter:
          size: bytes
          output: readable size as string (xxxGB, xxxTB etc)
    '''
    kilo = float(1024.0)
    size = float(size)
    # don't use decimal for bytes
    if size < kilo:
        return "%d bytes" % size
    size = size / kilo

    for x in ['KB', 'MB', 'GB', 'TB']:
        if size < kilo:
            break
        size = size / kilo
    if size.is_integer():
        size = int(size)
    else:
        size = round(size, 2)
    return "%s%s" % (size, x)


def format_human_readable(readable_size):
    '''
    Format readable size to generic format. We show user on GB/MB/TB/KB format.
      parameter:
          readable_size: string format( g, G, Gib, gb, TB etc)
          output: formatted readable size ( GB, MB, TB, KB)
                  Raise exception for wrong format input
    '''
    size = readable_size.replace('i', '')
    size = size.lower()
    match_obj = re.match('(\d+(?:\.\d+)?)([g|t|k|m]+[b]*)$', size, re.IGNORECASE)
    if not match_obj:
        raise ValueError("Cannot interpret size, wrong format: %s" % readable_size)
    num = float(match_obj.groups()[0])
    # format float 1.00 as 1 1.006 as 1.01 etc
    if num.is_integer():
        num = int(num)
    else:
        num = round(num, 2)
    suffix = match_obj.groups()[1].rstrip('b')
    size = str(num) + suffix
    return size.upper() + 'B'


def get_full_mount_path(name):
    '''
    Get mount point of the share
      parameter:
        name: share/lv name
        output: mount point
    '''
    return os.path.join(os.path.sep, SHARE_POINT, name)


def get_full_lv_path(name):
    '''
    Get mount point of the share
      parameter:
        name: share/lv name
        output: lv patch ( soft link)
     '''
    return os.path.join(os.path.sep, ROOT_DEV, VG_NAME, name)


def get_proc_mount_lv_path(name):
    '''
    Get mount point of the share, typically /dev/mapper/lsa-name
      parameter:
        name: share/lv name
        output: mount point of share (actual path)
    '''
    return os.path.join(os.path.sep, ROOT_DEV, 'mapper', VG_NAME + '-' + name)


def parse_boolean(value):
    '''
    Parse the boolean  value:
    input might be Boolean, String or integer
    output: True/False
    '''
    if value is None:
        raise ValueError('Cannot parse boolean: value can not be None')
    val_lower = str(value).lower()
    if val_lower in ['1', 'true', 'yes']:
        return True
    elif val_lower in ['0', 'false', 'no']:
        return False
    else:
        raise ValueError('Invalid value: %s' % val_lower)


def round_to_multiplier(value, divider):
    '''
    Returns the closest value that can be divised by divider without remainder
    parameter:
         value: value to be rounded
         divider: divider to be used for the rounding
         output: rounded integer
    '''
    assert (divider != 0)
    return int(divider * round(float(value) / divider))
