# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
"""
validation utility:
Validate different inputs ( file, IP address, hostname etc)
"""

import string
import os
import socket
import errno
import re
from command_util import command_output_lines
from common import CMD_IP

safechars = string.letters + string.digits + "~ -_./"


def is_valid_file_name(name):
    """
    Validate file/directory name
    checks:
        - non supported character
        - invalid path
    parameters:
    name : string
         file/dir name
    """
    if not name:
        raise TypeError("Empty file")

    initial_check = ""
    try:
        initial_check = filter(lambda c: c in safechars, name)
    except:
        raise ValueError("Invalid character in the file name")
    """
    After filtering initial_check and name should be the same for
    valid file name/dir
    """
    if initial_check != name:
        raise ValueError("Invalid character in the file name:" + name)

    # check if the file is not in a directory
    if "/" not in name:
        return True
    root_dirname = os.path.sep

    root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep

    # check each part of the path
    for part in name.split(os.path.sep):
        try:
            os.lstat(root_dirname + part)
        except OSError as exc:
            if exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
                return False

    return True


def is_valid_ipv4_address(address):
    """
    check if the address ia a valid IPV4 address
    parameters:
    address : string
            IPV4 address (valid 192.168.1.1)
    """
    try:
        socket.inet_pton(socket.AF_INET, address)
    except AttributeError:
        try:
            socket.inet_aton(address)
        except socket.error:
            return False
        return address.count('.') == 3
    except socket.error:  # not a valid address
        return False

    return True


def is_valid_ipv6_address(address):
    """
    check if the address ia a valid IPV6 address
    parameters:
    address : string
            IPV6 address
    """

    try:
        socket.inet_pton(socket.AF_INET6, address)
    except socket.error:
        return False
    return True


def is_valid_ip(address):
    """
    Generic IP address validation
    this function validates both IPV4 and IPV6 address
    parameters:
    address : string
            IPV4 or IPV6 address also may contain network address
            valid address: 2001:db8:9:e54::/64, 192.168.1.0/24 etc
    """
    # assume default is AF_INET
    ip_type = socket.AF_INET
    if ":" in address:
        ip_type = socket.AF_INET6

    fragments = address.split('/')

    if len(fragments) == 2:
        ip_addr = fragments[0]
    elif len(fragments) == 1:
        ip_addr = fragments[0]
    else:
        return False

    if ip_type == socket.AF_INET:
        return is_valid_ipv4_address(ip_addr)
    elif ip_type == socket.AF_INET6:
        return is_valid_ipv6_address(ip_addr)

    return False


NETGROUP_MATCHER = re.compile('^@[a-zA-Z]+')


def is_valid_share_client(client):
    """
    validate that the parameter is valid as
    Machine Name Formats for nfs export host.
    See exports(5)
    return True for valid client, False otherwise
    raises:
       ValueError: invalid parameter type given
    """

    if client is None:
        raise ValueError("cannot be null or empty")
    if not isinstance(client, basestring):
        raise ValueError("string parameter expected")

    if client == '*':
        # anonymous
        return True
    if NETGROUP_MATCHER.match(client):
        # netgroup
        return True
    if is_valid_ip(client):
        # IP networks
        return True

    # as that point any hostname with wild card is fine
    if not client == '':
        return True

    return False


def is_valid_netif(netif):
    """
    Check whether the netif is a valid network interface
    The acceptable format of a valid netif is either a name or an ipv4 address
    Example:
        lo or 127.0.0.1
        eth0 or 10.10.10.1
    Note an interface without ip addr configurated is not a valid netif
    """
    netifs = {}
    cmd = [CMD_IP, '-4', 'addr']
    out = command_output_lines(cmd)
    for line in out:
        if (re.search('inet', line)):
            split = line.split()
            # the format of line may look like
            # inet 127.0.0.1/8 scope host lo
            # inet 10.129.200.105/24 brd 10.129.200.255 scope global eno1
            # we will save the info as
            # netifs['lo'] = 127.0.0.1, netifs['eno1'] = 10.129.200.105
            netifs[split[len(split) - 1]] = split[1].split("/")[0]
    if netif in netifs.keys() or netif in netifs.values():
        return True
    else:
        return False


if __name__ == '__main__':
    pass
