# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
"""
Module hosting helper functions for XFS administrative tasks
It uses xfsdump(8) and xfsrestore(8) CLI to respectively dump and restore filesystems
Asynchronous operation can be done using executors and queues

ex of dump of share 'foo'
# creates a xfsdump process
dumper = get_dump_command('/shares/foo')
# creates to queue to hold dump output parts
queue = StreamQueue(3)
# create the executor (thread) to consume the dumper output and fill the queue
dumperExecutor = DumpStreamExecutor(dumper, queue)
# start the dump
dumper.start()
# start the executor thread
dumperExecutor.start()
# free resource
dumper.cleanup()


# creates and start the XFS restorer
restorer = get_restore_command(os.path.join(SHARE_POINT, shareName))
queue = StreamQueue(3)
# create the executor (thread) to consume the queue and fill xfsrestore input
restorerExecutor = RestoreStreamExecutor(restorer, task, queue)
# starts the restore command
restorer.start()
# start the restore executor thread
restorerExecutor.start()
# free resource
restorer.cleanup()


"""

from backend.log_util import get_logger, set_log_config
from backend.common import (CMD_XFS_DUMP, CMD_XFS_RESTORE)
from datetime import datetime
import os
import Queue
import threading
from subprocess import Popen, PIPE
import cStringIO
from backend.lsa_types import (StreamPart, EOSPart, ErrorPart)
from backend.taskmgr import (TaskStatusCode, TaskStatus)

LOG = get_logger('lsa.xfs_util')


class XFSExecutor(object):
    """
    Wrapper around subprocess.Popen instance
    """
    def __init__(self, arguments):
        """
        Creates a new XFSExecutor
        Parameters:
          arguments: arguments to be passed to Popen
        """
        self.cmdargs = arguments
        self.proc = None
        self.err = None

    def start(self):
        """
        Starts the sub-process
        """
        try:
            self.proc = Popen(self.cmdargs, bufsize=0, stdout=PIPE, stderr=PIPE, stdin=PIPE)
        except Exception:
            for l in self.proc.stderr:
                LOG.debug('XFSExecutor STDERR: %s' % l)
            raise
        LOG.debug('XFSExecutor: cmd executed pid=%d' % self.proc.pid)

    def get_output_stream(self):
        """
        Returns sub-process output stream
        """
        return self.proc.stdout

    def get_input_stream(self):
        """
        Returns sub-process input stream
        """
        return self.proc.stdin

    def cleanup(self, force=False):
        """
        Do some cleanup (wait for the sub-process)
        Parameters:
         force: if True, sub-process is killed first
        """
        if force:
            try:
                self.proc.kill()
            except Exception:
                LOG.debug('kill has failed')

        self.proc.wait()
        if (self.proc.returncode != 0):
            # collect stderr in that case
            self.err = self.proc.stderr.readlines()

    def get_error_condition(self):
        """
        In case of sub-process execution error, return the error
        text
        """
        return self.proc.stderr

    def is_success(self):
        """
        Returns True if sub-process execution went fine
        """
        return self.proc.returncode and self.proc.returncode == 0


def get_resume_restore_command(base, subtree=None):
    """
    Spawns a sub process to resume a previous restore of XFS filesystem.
    If a previous restore command has failed it may be impossible
    to restore again the same filesystem
    parameters:
         base : the base (mountpoint) of the  XFS filesystem to be restored
         subtree : subtree of the filesystem to be restored (optional).
                    when not specified the entire filesystem is restored
    Returns a XFSExecutor instance
    """
    args = [CMD_XFS_RESTORE, '-J', '-Q', '-F']
    if subtree:
        args.extend(['-s', subtree])

    args.append('-')

    args.append(base)
    LOG.debug('resume restore command %s' % args)
    executor = XFSExecutor(args)
    return executor


def get_restore_command(base, subtree=None):
    """
    Spawns a sub process to restore XFS filesystem.
    parameters:
         base : the base (mountpoint) of the  XFS filesystem to be restored
         subtree : subtree of the filesystem to be restored (optional).
                    when not specified the entire filesystem is restored
    Returns a XFSExecutor instance
    """
    args = [CMD_XFS_RESTORE, '-F', '-J', '-o']
    if subtree:
        args.extend(['-s', subtree])

    args.append('-')

    args.append(base)
    LOG.debug('restore command %s' % args)
    executor = XFSExecutor(args)
    return executor


def get_dump_command(base, subtree=None):
    """
    Spawns a sub process to dump XFS filesystem.
    The dump is streamed on XFSExecutor output
    parameters:
          base : the base (mountpoint) of the  XFS filesystem to be dumped
          subtree : subtree of the filesystem to be dumped (optional). when not specified
                    the entire filesystem is dumped
    example:
           getDumpExecutor('/foo') => dump content of filesystem mounted on /foo
           getDumpExecutor('/foo','bar') => dump content of /foo/bar
    returns:
         a XFSExecutor instance.
    """
    _label = datetime.now().strftime("%d%m%y%H%M")
    args = [CMD_XFS_DUMP, '-F', '-J', '-L', _label, '-M', _label]
    if subtree:
            args.extend(['-s', subtree])

    args.append('-')
    args.append(base)

    LOG.debug('dump command %s' % args)
    executor = XFSExecutor(args)

    return executor


class StreamExecutor(threading.Thread):
    """
    Stream executor master-class
    Stream in and out around a XFSExecutor instance
    """

    # 128MB
    PART_SIZE_BYTES = 134217728

    def __init__(self, executor, stream_queue):
        """
        creates a new StreamExecutor
        parameters:
           executor: an instance of XFSExecutor
           stream_queue: an instance of StreamQueue
        """
        threading.Thread.__init__(self)
        self.executor = executor
        self.queue = stream_queue
        self.do_stop = False
        self.consumedIndex = 1

    def __str__(self):
        return '%s(%d)' % (self.name, id(self))

    def stop(self):
        """
        Asks the instance to stop
        """
        self.do_stop = True

    def is_success(self):
        return True


class DumpStreamExecutor(StreamExecutor):
    """
    StreamExecutor class for XFS dump operation
    This StreamExecutor will consume XFSdump executor output
    and post fragment on the stream_queue
    """

    def __init__(self, executor, stream_queue):
        StreamExecutor.__init__(self, executor, stream_queue)
        self.name = 'DumpStreamExecutor'

    def run(self):
        while True:
            if self.do_stop:
                break
            buffer = cStringIO.StringIO()
            buffer.write(self.executor.get_output_stream().read(
                DumpStreamExecutor.PART_SIZE_BYTES))
            if buffer.tell() == 0:
                _s = EOSPart()
            else:
                _s = StreamPart(self.consumedIndex, buffer.tell(), buffer)
            self.consumedIndex += 1
            shallWeStop = (isinstance(_s, EOSPart))
            # rewind the buffer
            buffer.seek(0, os.SEEK_SET)
            self.queue.put(_s, True)
            LOG.debug('new sequence posted %s' % str(_s))
            if (shallWeStop):
                # shutdown the queue
                self.queue.set_active(False)
                self.executor.cleanup()
                LOG.debug('%s:EOF reached, stop worker' % str(self))
                break


class RestoreStreamExecutor(StreamExecutor):
    """
    StreamExecutor class for XFS restore operation
    This StreamExecutor will consume fragment from the stream_queue and
    send them to XFSrestore executor input
    """

    def __init__(self, executor, task, stream_queue):
        StreamExecutor.__init__(self, executor, stream_queue)
        self.name = 'RestoreStreamExecutor'
        self.task = task

    def run(self):
        self.task.begin()

        while True:
            LOG.debug('begin of loop do_stop=%s, qempty=%s, qactive=%s' % (self.do_stop,
                                                                           self.queue.empty(),
                                                                           self.queue.is_active()))
            if self.do_stop:
                break
            if self.queue.empty() and not self.queue.is_active():
                LOG.debug('processing done, exit now')
                break
            try:
                # never block,always get out on timeout
                # with consurrency we may have entered the loop because the queue was (still) active
                # but not anymore
                part = self.queue.get(True, 30)
            except Queue.Empty, e:
                continue

            if isinstance(part, EOSPart):
                LOG.debug('EOF reached, closing input of  worker and exit')
                self.executor.get_input_stream().close()
                self.executor.cleanup()
                self.queue.task_done()
                break

            if isinstance(part, ErrorPart):
                LOG.debug('Error raised in queue, closing input of  worker and exit')
                self.executor.get_input_stream().close()
                self.executor.cleanup()
                self.task.end(TaskStatus(TaskStatusCode.FAILED,
                                         'restore failed : %s ' % part.get_buffer().getvalue()))
                self.queue.task_done()
                break
            LOG.debug('new sequence fetched %s' % str(part))
            try:
                self.executor.get_input_stream().write(part.get_buffer().getvalue())
                LOG.debug('sequence written to restorer command ')
            except IOError, e:
                LOG.warning('error while writing to executor [%s]' % str(e))
                # warn the writer that nobody wait for this
                self.queue.set_active(False)
                LOG.warning('error while writing to executor [%s]' % str(e))
                self.task.end(TaskStatus(TaskStatusCode.FAILED, 'restore failed:%s' % str(e)))
                LOG.warning('stopped task %s' % self.task)
                raise
            finally:
                part.cleanup()
                self.queue.task_done()

        # other part may have already updated the task
        # if the task has a status already, do nothing
        if self.task.get_status() is None:
            self.task.end(TaskStatus(TaskStatusCode.SUCCESS, 'restore completed'))


def run_all_test():
    pass


if __name__ == '__main__':
    set_log_config()

    run_all_test()
