/**
 * Copyright (c) 2017, 2018, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */

/**
 * Hold common helpers
 * This module serves as a repository for simple helper functions used by multiple modules
 *
 */

define(['ojs/ojcore', 'ojs/ojvalidation-number'],
       function (oj)
       {
           /**
            * byte amount in a KB
            */
           var KB_IN_BYTES = 1024;
           /**
            * byte amount in a MB
            */
           var MB_IN_BYTES = 1048576;
           /**
            * byte amount in a GB
            */
           var GB_IN_BYTES = 1073741824;
           /**
            * byte amount in a TB
            */
           var TB_IN_BYTES = 1099511627776;

           /**
            * extended versions of oj.IntlNumberConverter for bytes
            * that support parsing of string like '1GB'
            */
           function ExtendedByteConverter(options) {
               this.Init(options);
           }
           oj.Object.createSubclass(ExtendedByteConverter,
                                    oj.IntlNumberConverter,
                                    'ExtendedByteConverter');
           function X4mByteConverter(options) {
               this.Init(options);
           }
           oj.Object.createSubclass(X4mByteConverter,
                                    ExtendedByteConverter,
                                    'X4mByteConverter');

           function _isInteger(value) {
               return typeof value === "number" &&
                   isFinite(value) &&
                   Math.floor(value) === value;
           }
           /**
            * we convert all string like <number>[unit] to bytes
            * @param {string} value - input value to be parsed
            */
           ExtendedByteConverter.prototype.parse = function(value)
           {

               if (value === null || value === '') {
                   return null;
               }
               oj.Logger.log('convert ' + value);
               var _result = ExtendedByteConverter.superclass.parse.call(this, this.toBytes(value));
               oj.Logger.log('===>' + _result);
               return _result;
           };
           /**
            * return byte from string
            * parsing string as byte using multiplier
            * if string ends with T,TB,G,GB,K,KB
            * if the parameter is already an integer, we return it directly
            * @throws Error if cannot parse a number
            * @returns an integer
            */
           ExtendedByteConverter.prototype.toBytes = function (value) {
               if (_isInteger(value)) {
                   //nothing to do
                   return value;
               }

               oj.Assert.assertNonEmptyString(value);

               var _v = value.toUpperCase();
               var _m = 1;
               if (/BYTE$/.test(_v)) {
                   _m = 1;
               } else if (/K$/.test(_v)) {
                   _m = KB_IN_BYTES;
               } else if (/KB$/.test(_v)) {
                   _m = KB_IN_BYTES;
               } else if (/M$/.test(_v)) {
                   _m = MB_IN_BYTES;
               } else if (/MB$/.test(_v)) {
                   _m = MB_IN_BYTES;
               } else if (/G$/.test(_v)) {
                   _m = GB_IN_BYTES;
               } else if (/GB$/.test(_v)) {
                   _m = GB_IN_BYTES;
               } else if (/T$/.test(_v)) {
                   _m = TB_IN_BYTES;
               } else if (/TB$/.test(_v)) {
                   _m = TB_IN_BYTES;
               }

               var parsedValue = parseFloat(_v);
               if (isNaN(parsedValue)) {
                   throw new Error('Invalid byte number: ' + _v);
               }
               return Math.floor(parsedValue * _m);

           };
           /**
            * return the number of bytes, rounded up to multiple of 4MB, from string
            * parsing string as number of bytes using multiplier
            * if string ends with T,TB,G,GB,K,KB
            * if the parameter is already an integer, we round it up and return
            * @throws Error if cannot parse a number
            * @returns an integer
            */
           X4mByteConverter.prototype.toBytes = function (value) {
               var new_value = ExtendedByteConverter.prototype.toBytes(value)
               return Math.floor((new_value + 4194303) / 4194304) * 4194304;
           };


           var ExtendedByteConverterFactory = (function () {
               function _createConverter(options)
               {
                   return new ExtendedByteConverter(options);
               }

               return {
                   'createConverter': function (options) {
                       return _createConverter(options);
                   }
               };
           }());

           oj.Validation.converterFactory("LSAByteConverter",
                                          ExtendedByteConverterFactory);

           var X4mByteConverterFactory = (function () {
               function _createConverter(options)
               {
                   return new X4mByteConverter(options);
               }

               return {
                   'createConverter': function (options) {
                       return _createConverter(options);
                   }
               };
           }());

           oj.Validation.converterFactory("X4mByteConverter",
                                          X4mByteConverterFactory);

           var _MILLISECONDS_IN_A_YEAR = 31556952000;
           var _MILLISECONDS_IN_A_MONTH = 2629746000;
           var _MILLISECONDS_IN_A_WEEK = 604800000;
           var _MILLISECONDS_IN_A_DAY = 86400000;
           var _MILLISECONDS_IN_A_HOUR = 3600000;
           var _MILLISECONDS_IN_A_MINUTE = 60000;

           /**
            * get the time difference between now and given date in the past
            * @param {integer} from - the time in milliseconds since EPOC
            * @param {integer} to - the base time in milliseconds, if undefined , Date.now() is used
            * @returns {Object} - the meaningful time differences as
            *     {
            *       'years' : number of years in the difference (can be zero)
            *       'months' : number of months in the difference (can be zero)
            *       'weeks' : number of weeks in the difference (can be zero)
            *       'days' : number of days in the difference (can be zero)
            *       'hours' : number of hours  in the difference (can be zero)
            *       'minutes' : number of minutes in the difference (can be zero)
            *       'seconds' : number of minutes in the difference (can be zero)
            *     }
            */
           function getElapsedTime(from, to) {
               oj.Assert.assertNumber(from);
               var _to = to;
               if (typeof _to !== 'undefined') {
                   oj.Assert.assertNumber(_to);
               } else{
                   _to = Date.now();
               }
               var result = {};

               var fromNow = _to - from;
               oj.Assert.assert(fromNow >= 0);

               result.years = Math.floor((fromNow / _MILLISECONDS_IN_A_YEAR));
               fromNow -= result.years * _MILLISECONDS_IN_A_YEAR;

               result.months =  Math.floor((fromNow / _MILLISECONDS_IN_A_MONTH));
               fromNow -= result.months * _MILLISECONDS_IN_A_MONTH;

               result.weeks =  Math.floor((fromNow / _MILLISECONDS_IN_A_WEEK));
               fromNow -= result.weeks * _MILLISECONDS_IN_A_WEEK;

               result.days =  Math.floor((fromNow / _MILLISECONDS_IN_A_DAY));
               fromNow -= result.days * _MILLISECONDS_IN_A_DAY;

               result.hours =  Math.floor((fromNow / _MILLISECONDS_IN_A_HOUR));
               fromNow -= result.hours * _MILLISECONDS_IN_A_HOUR;

               result.minutes =  Math.floor((fromNow / _MILLISECONDS_IN_A_MINUTE));
               fromNow -= result.minutes * _MILLISECONDS_IN_A_MINUTE;

               result.seconds = Math.floor((fromNow / 1000));

               return result;
           }

           return {
               /**
                * extended byte converters
                * usual usage is to use oj.validation.converterFactory to instantiate
                * it
                */
               'byteConverter': function (options) {
                   return new ExtendedByteConverter(options);
               },
               'x4mByteConverter': function (options) {
                   return new X4mByteConverter(options);
               },
               /**
                * Generate a unique ID
                * @return {integer} - a unique (random) id
                */
               'genUid': function () {
                   return Math.round(Math.random() * 10000);
               },
               'KB_IN_BYTES':KB_IN_BYTES,
               'MB_IN_BYTES':MB_IN_BYTES,
               'GB_IN_BYTES':GB_IN_BYTES,
               'TB_IN_BYTES':TB_IN_BYTES,
               'MILLISECONDS_IN_A_YEAR':  _MILLISECONDS_IN_A_YEAR,
               'MILLISECONDS_IN_A_MONTH': _MILLISECONDS_IN_A_MONTH,
               'MILLISECONDS_IN_A_WEEK':  _MILLISECONDS_IN_A_WEEK,
               'MILLISECONDS_IN_A_DAY':   _MILLISECONDS_IN_A_DAY,
               'MILLISECONDS_IN_A_HOUR':  _MILLISECONDS_IN_A_HOUR,
               'MILLISECONDS_IN_A_MINUTE':_MILLISECONDS_IN_A_MINUTE,
               'getTimeString': function(time) {
                   return new Date(time).toUTCString();
               },
               'getElapsedTime':getElapsedTime,
               /**
                * get the time difference between now and given date in the past
                * @param {integer} from - the time in milliseconds since EPOC
                * @param {integer} to - the base time in milliseconds, if undefined , Date.now() is used
                * @returns {string} - the meaningful time differences as string, can be undefined
                *   if time is the same
                * ex :
                *    'one week ago'
                *    '2 years, 10 months, 2 weeks ago'
                *    '16 hours ago'
                */
               'getElapsedTimeString': function (from, to) {
                   oj.Assert.assertNumber(from);
                   var difference  = getElapsedTime(from, to);

                   if ((difference.years +
                        difference.months +
                        difference.weeks +
                        difference.days +
                        difference.hours +
                        difference.minutes +
                        difference.seconds ) === 0) {
                       // nothing to do
                       return undefined;
                   }

                   var result = '';
                   if (difference.years > 0) {
                       result = difference.years + ' year';
                       if (difference.years > 1) {
                           result = result.concat('s');
                       }
                   }
                   if (difference.months > 0) {
                       if (result.length !==  0) {
                           result = result.concat(', ');
                       }
                       result = result.concat(difference.months,' month');
                       if (difference.months > 1) {
                           result = result.concat('s');
                       }
                   }
                   if (difference.weeks > 0) {
                       if (result.length !==  0) {
                           result = result.concat(', ');
                       }
                       result = result.concat(difference.weeks,' week');
                       if (difference.weeks > 1) {
                           result = result.concat('s');
                       }
                   }
                   if (difference.days > 0) {
                       if (result.length !==  0) {
                           result = result.concat(', ');
                       }
                       result = result.concat(difference.days,' day');
                       if (difference.days > 1) {
                           result = result.concat('s');
                       }
                   }
                   if (difference.hours > 0) {
                       if (result.length !==  0) {
                           result = result.concat(', ');
                       }
                       result = result.concat(difference.hours,' hour');
                       if (difference.hours > 1) {
                           result = result.concat('s');
                       }
                   }
                   if (difference.minutes > 0) {
                       if (result.length !==  0) {
                           result = result.concat(', ');
                       }
                       result = result.concat(difference.minutes,' minute');
                       if (difference.minutes > 1) {
                           result = result.concat('s');
                       }
                   }
                   if (difference.seconds > 0) {
                       if (result.length !==  0) {
                           result = result.concat(', ');
                       }
                       result = result.concat(difference.seconds,' second');
                       if (difference.seconds > 1) {
                           result = result.concat('s');
                       }
                   }

                   return result + ' ago';
               },
               /**
                * TEST / internal use
                */
               'testTimeDifference': function() {
                   var _now = Date.now();
                   var diff = this.getElapsedTime(_now,_now);
                   oj.Assert.assert((diff.years === 0) && (diff.months === 0) &&
                                    (diff.weeks === 0) && (diff.days === 0) &&
                                    (diff.hours === 0) && (diff.minutes === 0));

               },
               /**
                * TEST / internal use
                */
               'testTimeDifferenceString': function() {
                   var _now = Date.now();
                   var diff = this.getElapsedTimeString(_now,_now);
                   oj.Assert.assert(typeof diff === 'undefined');

                   diff = this.getElapsedTimeString(_now - _MILLISECONDS_IN_A_MINUTE,
                                                    _now);
                   oj.Assert.assert(diff === '1 minute(s) ago');
                   diff = this.getElapsedTimeString(_now - _MILLISECONDS_IN_A_HOUR,
                                                    _now);
                   oj.Assert.assert(diff === '1 hour(s) ago');
                   diff = this.getElapsedTimeString(_now - _MILLISECONDS_IN_A_DAY,
                                                    _now);
                   oj.Assert.assert(diff === '1 day(s) ago');
                   diff = this.getElapsedTimeString(_now - _MILLISECONDS_IN_A_WEEK,
                                                    _now);
                   oj.Assert.assert(diff === '1 week(s) ago');
                   diff = this.getElapsedTimeString(_now - _MILLISECONDS_IN_A_MONTH,
                                                    _now);
                   oj.Assert.assert(diff === '1 month(s) ago');

                   diff = this.getElapsedTimeString(_now - _MILLISECONDS_IN_A_YEAR -
                                                    (2 * _MILLISECONDS_IN_A_MONTH) -
                                                    (3 * _MILLISECONDS_IN_A_WEEK) -
                                                    (6 * _MILLISECONDS_IN_A_DAY) -
                                                    (23 * _MILLISECONDS_IN_A_HOUR) -
                                                    (59 * _MILLISECONDS_IN_A_MINUTE),
                                                    _now);
                   oj.Assert.assert(diff === '1 year(s), 2 month(s), 3 week(s), 6 day(s), 23 hour(s), 59 minute(s) ago');
               }
           };
       }
      );
