/**
 * Copyright (c) 2018, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */
/**
 * Module that holds common type definitions
 */

define(['ojs/ojcore', 'jquery'], function(oj, $) {

    /**
     * @typedef {Object} Share
     * @property {string} name - name of the share
     * @property {number} size - size of the share in bytes
     * @property {number} free - free space in bytes
     * @property {Object} fs_options - filesystems options
     * @property {boolean} fs_options.inode32 - use 32bits inodes
     */
     /**
      * Factory for a new share
      * @param {string} name - name for the new share (optional)
      * @return {Share} - a new share
      */
    function createShare(name) {
        return {
            'name':name,
            'size':0,
            'free':0,
            'fs_options' : {
                'inode32': false
            },
            exports: []
        };
    }

    /**
     * @typedef {Object} ShareExport
     * @property {string} type - type of the export
     * @property {string} hosts - client host (NFS only)
     * @property {string} id - unique id o the export
     * @property {string} path - access point of the resource
     * @property {Object} options - options of the export
     * @property {string} options.hide - (NFS only)
     * @property {string} options.read_only
     * @property {string} options.root_squash - (NFS only)
     * @property {string} options.secure - (NFS only)
     * @property {string} options.subtree_check - (NFS only)
     * @property {string} options.sync - (NFS only)
     */

    function _newShareExport (exportType) {
        var _result = {
            'type':exportType,
            'id' : undefined,
            'path': undefined,
            'options': {
                'read_only' : true
            }
        };
        if (exportType === 'nfs') {
            _result.type = 'nfs';
            _result.hosts = undefined;
            _result.options.hide=false;
            _result.options.root_squash=true;
            _result.options.secure=true;
            _result.options.subtree_check=false;
            _result.options.sync=true;
            _result.get_id = function () {
                return this.hosts;
            };
        }
        if (exportType === 'smb') {
            _result.type = 'smb';
            _result.resource = undefined;
            _result.options.comment=undefined;
            _result.options.browsable=false;
            _result.options.guest_ok=false;
            _result.options.strict_sync=false;
            _result.options.sync_always=false;
            _result.options.hosts_allow=undefined;
            _result.options.hosts_deny=undefined;
            _result.get_id = function () {
                return this.resource;
            };
        }

        return _result;
    }
    function _duplicateShareExport(o) {
        var result = $.extend(true,{},o);
        result.path = undefined;
        if (result.type === 'smb') {
            result.resource = undefined;
        }
        result.id = undefined;
        return result;
    }
    /**
     * Factory to create a new share export or duplicate an existing one.
     * if an export is specified, it is used for duplication. All per export unique
     * properties are restted to default value
     * @param {string|Object} exportObj -  type of the export ('nfs' or 'smb') or an export
     * object to be duplicate
     * @return {ShareExport} - a new export
     */
    function createShareExport (exportObj) {
        oj.Assert.assert((typeof exportObj === 'string') || (typeof exportObj === 'object'));
        if (typeof exportObj === 'string') {
            oj.Assert.assert((exportObj === 'smb') || (exportObj === 'nfs'));
            return _newShareExport(exportObj);
        }
        /* duplication case 'typeof exportObj === object ' */
        return _duplicateShareExport(exportObj);
    }

    return {
        newShare: createShare,
        newShareExport: createShareExport
    };
});
