/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */
define(['ojs/ojcore', 'jquery', 'signals', 'models/storage-types'],
       function(oj, $, signals, storageTypeFactory) {
           function StoragePool() {

               var self = this;

               var getSharesOptions = {
                   async: true,
                   dataType: "json"
               };

               /**
                * Gets all exports for a share
                * @param {string} shareName - the share name
                * @return {Promise<ShareExport[]>} resolve on an array of ShareExport object, can be empty
                * @public
                */
               this.getShareExports = function (shareName) {
                   oj.Assert.assertNonEmptyString(shareName);
                   var _url = '/lsa/shares/'+shareName+'/exports';
                   return(
                       new Promise(function(resolve, reject) {
                           $.ajax($.extend({'url':_url},getSharesOptions)).done(
                               function getShareExportsAjaxDone (data) {
                                   resolve(data.list);
                               }
                           ).fail(
                               function getShareExportsAjaxFail (jqxhr, text_status, error_thown) {
                                   reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                               }
                           );
                       })
               );
               };
               /**
                * Gets a share by name
                *
                * @param {string} shareName - the share name
                * @return {Promise<Share>} resolve on a Share object
                * @public
                */
               this.getShare = function (shareName) {
                   oj.Assert.assertNonEmptyString(shareName);
                   return(
                       new Promise(function(resolve, reject) {
                           $.ajax($.extend({'url':'/lsa/shares/' + shareName},getSharesOptions)).done(
                               function getSharesAjaxDone (data) {
                                   resolve(data);
                               }
                           ).fail(
                               function getSharesAjaxFail (jqxhr, text_status, error_thown) {
                                   reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                               }
                           );
                       })
                   );
               };

               /**
                * Gets all shares
                * @return {Promise<Share[]>} resolve on an array of Share object, can be empty
                * @public
                */
               this.getShares = function () {
                   return(
                       new Promise(function(resolve, reject) {
                           $.ajax($.extend({'url':'/lsa/shares'},getSharesOptions)).done(
                               function getSharesAjaxDone (data) {
                                   resolve(data.list);
                               }
                           ).fail(
                               function getSharesAjaxFail (jqxhr, text_status, error_thown) {
                                   reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                               }
                           );
                       })
                   );
               };
               /**
                * Returns the list of tasks targeted to a specfic share
                * resolve a list of task as
                * 'ID' : unique ID as int
                * 'name': task name as string
                * 'target': the subject of the task (usually share name)
                * 'status' : {
                *       'code': status code (see backend.TaskStatusCode),
                *       'message': status message
                *      },
                * 'states' : [{
                *       'type': state type (see backend.TaskState),
                *       'message': state message ,
                *       'time': timestamp
                *      }
                */
               this.geShareTaskList = function(shareName) {
                   oj.Assert.assertNonEmptyString(shareName);
                   return(
                       new Promise(function(resolve, reject) {
                           $.ajax( {
                               'url':'/lsa/shares/'+shareName+'/tasks',
                               'async': true,
                               'dataType': "json"
                           }).done(
                               function getShareTaskListAjaxDone (data) {
                                   resolve(data.tasks.map(JSON.parse));
                               }
                           ).fail(
                               function getShareTaskListAjaxFail (jqxhr, text_status, error_thown) {
                                   reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                               }
                           );
                       })
                   );
               };
               /**
                * Returns the list of tasks targeted to a specfic share
                * of a specific type
                * resolve a list of task as
                * 'ID' : unique ID as int
                * 'name': task name as string
                * 'target': the subject of the task (usually share name)
                * 'status' : {
                *       'code': status code (see backend.TaskStatusCode),
                *       'message': status message
                *      },
                * 'states' : [{
                *       'type': state type (see backend.TaskState),
                *       'message': state message ,
                *       'time': timestamp
                *      }
                */
               this.geShareTaskTypeList = function(shareName, taskType) {
                   oj.Assert.assertNonEmptyString(shareName);
                   oj.Assert.assertNonEmptyString(taskType);
                   return(
                       new Promise(function(resolve, reject) {
                           $.ajax( {
                               'url':'/lsa/shares/'+shareName+'/tasks',
                               'method':'GET',
                               'async': true,
                               'dataType': "json",
                               'data': {
                                   'taskName': taskType
                               }
                           }).done(
                               function geShareTaskTypeList (data) {
                                   resolve(data.tasks.map(JSON.parse));
                               }
                           ).fail(
                               function geShareTaskTypeList (jqxhr, text_status, error_thown) {
                                   reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                               }
                           );
                       })
                   );
               };

               /**
                * Removes a share
                * @param name: string representing the name of the share
                *
                * @returns {Promise} a promise which resolve on nothing
                * In case of error response text is returned
                */
               this.deleteShare = function (name) {
                   oj.Assert.assertNonEmptyString(name);
                   var ajaxCallOption = {
                       'url':'/lsa/shares/'+ name,
                       'method': 'DELETE'
                   };
                   return (
                       new Promise(function (resolve,reject) {
                           $.ajax(ajaxCallOption).done(function deleteShareAjaxDone( data ) {
                               self.shareDeletionEventchannel.dispatch(
                                   {'name':name});
                               resolve(data);
                           }).fail(function deleteShareAjaxFail(jqxhr, text_status, error_thown) {
                               //TODO : find a way to only send back the message
                               //      not the entire HTML
                               reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                           });
                       }));
               };
               /**
                * Removes a new share export
                * @param {string} shareName - name of the share
                * @param {ShareExport} exportData - exports object name
                * @return {Promise} a promise which resolve on nothing
                */
               this.deleteShareExport = function (shareName, exportData) {
                   oj.Assert.assertNonEmptyString(shareName);
                   oj.Assert.assertObject(exportData);
                   return (new Promise(function (resolve,reject) {
                           $.ajax( {
                               'url':"/lsa/shares/" + shareName + '/exports/' + exportData.get_id(),
                               'method': 'DELETE',
                               'async': true,
                               'contentType': "application/json",
                               'data': JSON.stringify({
                                       'type' : exportData.type
                               })
                           }).done(function deleteShareExportAjaxDone(data) {
                               self.shareModificationEventchannel.dispatch({'name':shareName});
                               resolve(data);
                           }).fail(function deleteShareExportAjaxFail (jqxhr, text_status, error_thown) {
                               reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                           });
                       }));
               };
               /**
                * Adds a new share export
                * @returns a promise with a data payload containing the share
                * path on success and the error message on failure
                * @param shareName: string, name of the new share
                * @param {ShareExport} exportData - exports object
                */
               this.addShareExport = function (shareName, exportData) {
                   oj.Assert.assertNonEmptyString(shareName);
                   oj.Assert.assertObject(exportData);
                   return (new Promise(function(resolve,reject) {
                           $.ajax( {
                               'url':"/lsa/shares/" + shareName + '/exports',
                               'method': 'POST',
                               'async': true,
                               'contentType': "application/json",
                               'data': JSON.stringify(exportData)
                           }).done(function addShareExportAjaxDone (data) {
                               self.shareModificationEventchannel.dispatch({'name':shareName});
                               resolve(data);
                           }).fail(function addShareExportAjaxFail (jqxhr, text_status, error_thown) {
                               reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                           });
                       }));
               };

               /**
                * Adds a new share
                * @returns a promise with a data payload containing the share
                * name on success and the error message on failure
                * @param name: string, name of the new share
                * @param size: int , size of the share in bytes
                * accessible hosts
                * @param inode32: boolean, the inode32 option need to be set to
                * workaround legacy application limitation 26879349
                * Response text is returned
                */
               this.addShare = function (name, size, inode32) {
                   oj.Assert.assertNonEmptyString(name);
                   return (
                       new Promise(function(resolve,reject) {
                           $.ajax( {
                               'url':"/lsa/shares",
                               'method': 'POST',
                               'async': true,
                               'contentType': "application/json",
                               'data': JSON.stringify({
                                       'name' : name,
                                       'size' : size,
                                       'fs_options': {
                                           'inode32': inode32
                                       }
                               })
                           }).done(function addShareAjaxDone(data) {
                               self.shareCreationEventchannel.dispatch(data);
                               oj.Logger.log('task id = ' + data.task_id);
                               resolve(data);
                           }).fail(function addShareAjaxFail(jqxhr, text_status, error_thown) {
                               reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                           });
                       }));
               };

               /**
                * Modifies an existing share
                * @returns a promise with a data payload containing the share
                * path on success and the error message on failure
                * @param name: string, name of the new share
                * @param size: int , size of the share in bytes
                * Response text is returned
                */
               this.modifyShare = function (name, size) {
                   oj.Assert.assertNonEmptyString(name, 'missing name parameter');
                   return (
                       new Promise(function(resolve, reject) {
                           $.ajax ({method: 'PUT',
                                    url: '/lsa/shares/' + name,
                                    'contentType': "application/json",
                                    'data':
                                    JSON.stringify({
                                        'size' : size
                                    }),
                                    success: function( data ) {
                                       self.shareModificationEventchannel.dispatch(data);
                                       resolve(data);
                                   }}).fail(function(jqxhr, text_status, error_thown) {
                                       reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                                   });
                       }));
               };
               /**
                * Modifies an existing share export
                * @returns a promise with a data payload containing the share
                * path on success and the error message on failure
                * @param {string} shareName - name of the new share
                * @param {ShareExport} exportData - exports object
                * @return {Promise}
                */
               this.modifyShareExport = function (shareName,  exportData) {
                   oj.Assert.assertNonEmptyString(shareName);
                   oj.Assert.assertObject(exportData);
                   return (
                       new Promise(function(resolve, reject) {
                           $.ajax ({method: 'PUT',
                                    url: '/lsa/shares/' + shareName+ '/exports/' + exportData.get_id(),
                                    'contentType': "application/json",
                                    'data': JSON.stringify(exportData),
                                    success: function( data ) {
                                        self.shareModificationEventchannel.dispatch({'name':shareName});
                                        resolve(data);
                                   }}).fail(function(jqxhr, text_status, error_thown) {
                                       reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                                   });
                       }));
               };
               this.getPoolUsage = function () {
                   return(
                       new Promise(function(resolve, reject) {
                           $.ajax( {
                               'url':'/lsa/pool/utilization',
                               'async': true,
                               'dataType': "json"
                           }).done(
                               function getPoolUsageAjaxDone (data) {
                                   resolve(data);
                               }
                           ).fail(
                               function getPoolUsageAjaxFail (jqxhr, text_status, error_thown) {
                                   reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                               }
                           );
                       })
                   );
               };

               this.getPoolStatus = function () {
                   return(
                       new Promise(function(resolve, reject) {
                           $.ajax( {
                               'url':'/lsa/pool/status',
                               'async': true,
                               'dataType': "json"
                           }).done(
                               function getPoolStatusAjaxDone (data) {
                                   var all_devices=data.devices;
                                   $.ajax({
                                       'url':'/lsa/platform/disks',
                                       'async': true,
                                       'dataType': "json"
                                   }).done(
                                       function (pdisks) {
                                           all_devices.forEach(function(device) {
                                               pdisks.disks.forEach(function (disk){
                                                   if (device.dev === '/dev/'  + disk.name) {
                                                       // found it : add wwn info
                                                       device.wwn = disk.wwn;
                                                   }
                                               });
                                           });
                                           resolve(data);
                                       }
                                   ).fail(function (jqxhr, text_status, error_thown) {
                                       reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                                   });
                               }
                           ).fail(
                               function getPoolStatusAjaxFail (jqxhr, text_status, error_thown) {
                                   reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                               }
                           );
                       })
                   );
               };

           }
           //assume singleton for StoragePool.
           /**
            * share creation event channel.
            * after a share is successfully created the signal is dispatched
            * with the following object (in order).
            * {
            *   name : name of the new share
            * }
            *
            */
           StoragePool.prototype.shareCreationEventchannel = new signals.Signal();
           /**
            * share modification event channel.
            * after a share is successfully modified the signal is dispatched
            * with the following object (in order).
            * {
            *   name : name of the new share
            * }
            *
            */
           StoragePool.prototype.shareModificationEventchannel = new signals.Signal();
           /**
            * share deletion event channel.
            * after a share is successfully deleted the signal is dispatched
            * with the following object (in order).
            * {
            *   name : name of the new share
            * }
            *
            */
           StoragePool.prototype.shareDeletionEventchannel = new signals.Signal();
           /**
            * share backup & restore event channel.
            * after background task for share backup or restore is spawned, the signal is dispatched
            * with the following object (in order).
            * {
            *   taskId : task id of the task to do share backup or restore
            *
            * }
            *
            */
           StoragePool.prototype.shareBackupRestoreEventchannel = new signals.Signal();

           return new StoragePool();
       }
      );
