/**
 * Copyright (c) 2017, 2018, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */

/**
 * Add share dialog viewModel
 * bind against add-share-dialog.html views.
 * all exports are realised by applying add-share-nfs-export or
 * add-share-cifs-export views as tile within the masonry layout
 */

define(['ojs/ojcore',
        'knockout',
        'jquery',
        'ojs/ojknockout',
        'models/storagepool',
        'models/storage-types',
        'viewModels/common_vm_def',
        'viewModels/confirmation-dialog',
        'viewModels/client-commands-dialog',
        'helpers',
        'ojs/ojbutton',
        'ojs/ojmasonrylayout',
        'ojs/ojinputtext',
        'ojs/ojinputnumber',
        'ojs/ojslider',
        'ojs/ojswitch',
        'ojs/ojdialog',
        'ojs/ojvalidation-base'],
       function (oj, ko, $, ojk, storagePoolSVC, storageTypeFactory,
                 CommonVMDefinitons, confirmation, clientCommandsDialog, Helpers)
       {

           function AddShareViewModel()
           {
               var self = this;
               const size_4mb = 4194304;
               const size_16mb = 16777216;

               this.shareName =  ko.observable('');

               this.shareSize = ko.observable(0);
               this.shareMinSize = ko.observable(0);
               this.shareMaxSize = ko.observable(1);
               // backend requires multiple of 4MB
               this.shareSizeStep = ko.observable(size_4mb);
               this.shareSizeInstruction = ko.pureComputed(function () {
                   return 'Remaining available space: ' + self.byteConverter.format(self.shareMaxSize());
               });

               this.isShareModified = false;
               this.isShareDialogOpened = false;
               this.shareDialogDefaultWidth = undefined;
               this.shareDialogDefaultHeight = undefined;

               this.nameReadOnly = ko.observable(false);
               this.inode32Option = ko.observable(false);
               this.inode32OptReadOnly = ko.observable(false);
               this.dialogTitle = ko.observable('');

               this.actionType = undefined;

               this.existingShareNames = [];
               this.actionLabel = ko.observable(oj.Translations.getTranslatedString('Create'));

               this.shouldDisableCreateButton = ko.pureComputed(function () {
                   /* share related checks */
                   if (self.shareName() === '') {
                       return true;
                   }
                   if (self.shareSize() === 0) {
                       return true;
                   }
                   if (self.shareNameValidator.errorRaised() === true) {
                       return true;
                   }
                   if (self.duplicateNameOptionValidator.errorRaised() === true) {
                       return true;
                   }
                   /*****************************/

                   if (self.nfs.NFSExportIdentifierOptionValidator.errorRaised().length  !== 0) {
                       return true;
                   }
                   if (self.nfs.NFSIPsOptionValidator.errorRaised().length !== 0) {
                       return true;
                   }
                   if (self.cifs.CIFSHostOptionValidator.errorRaised().length !== 0) {
                       return true;
                   }
                   if (self.cifs.CIFSResourceOptionValidator.errorRaised().length !== 0) {
                       return true;
                   }
                   if (self.cifs.CIFSexportIdentifierOptionValidator.errorRaised().length  !== 0) {
                       return true;
                   }

                   self.exportIdentifierOptionChanged();

                   // if we have SMB export , resource name must not be empty
                   var found = false;
                   ko.utils.arrayForEach(self.allExports(), function(exp) {
                       if (exp.type === 'smb' && (exp.resource === '' || exp.resource === undefined)) {
                            found = true;
                        }
                   });
                   return found;
               });

               this.cancelButtonDisabled =  ko.observable(false);

               this.allExports = ko.observableArray([]);
               this.allDisplayableExports = ko.pureComputed(function() {
                   return self.allExports().filter(function (e) {
                       return (e.state === 'existing' || e.state === 'modified' || e.state === 'added');
                   });
               });
               /**
                * Reset UI components
                * @param {string} actionType - type of action being performed [create|modify|duplicate]
                * @param {Share} shareObject - the share we acting on
                */
               function resetDialogComponents(actionType, shareObject) {
                   switch(actionType) {
                   case 'create':
                       self.dialogTitle(oj.Translations.getTranslatedString('Add New Share'));
                       self.actionLabel(oj.Translations.getTranslatedString('Create'));
                       self.nameReadOnly(false);
                       self.shareName('');
                       self.inode32Option(false);
                       // backend requires minimum of 16MB
                       self.shareMinSize(size_16mb);
                       self.shareSize(size_16mb);
                       break;
                   case 'modify':
                       oj.Assert.assertObject(shareObject);
                       self.dialogTitle(oj.Translations.getTranslatedString('View or Modify '+shareObject.name +' Share'));
                       self.actionLabel(oj.Translations.getTranslatedString('Modify'));
                       self.shareName(shareObject.name);
                       self.inode32Option(shareObject.fs_options.inode32);
                       self.shareMinSize(shareObject.size);
                       self.shareSize(shareObject.size);
                       self.nameReadOnly(true);
                       break;
                   case 'duplicate':
                       oj.Assert.assertObject(shareObject);
                       self.dialogTitle(oj.Translations.getTranslatedString('Duplicate of ' +shareObject.name + ' Share'));
                       self.actionLabel(oj.Translations.getTranslatedString('Create'));
                       self.nameReadOnly(false);
                       self.shareName('');
                       self.inode32Option(shareObject.fs_options.inode32);
                       self.shareMinSize(size_16mb);
                       if (shareObject.size > self.shareMaxSize()) {
                           self.shareSize(self.shareMaxSize());
                       } else {
                           self.shareSize(shareObject.size);
                       }
                       break;
                   }
               }
               /**
                * method called to open new share creation dialog
                *
                * @param {number} availableSize - available space in byte of the storage pool
                * @param {string[]} shareNameList - list of existing share name. i.e share attribute which define its unicity
                * @param {string} actionType - either 'create', 'modify' or 'duplicate'
                * @param {Share} entry - entry object to be used to prepopulate dialog box.
                *               expect to be null if action is 'create'
                * @public
                */
               this.open = function (availableSize, shareNameList, actionType, entry)
               {
                   oj.Assert.assert(((actionType === 'create') ||
                                     (actionType === 'modify') ||
                                     (actionType === 'duplicate')));

                   /*
                    * first time we enter here save the height and width
                    * of the dialog
                    */
                   if (self.shareDialogDefaultWidth === undefined) {
                       self.shareDialogDefaultWidth = $(self.dialogSelector).css('width');;
                       self.shareDialogDefaultHeight = $(self.dialogSelector).css('height');
                   }

                   self.isShareModified = false;

                   self.allExports([]);

                   self.shareMaxSize(availableSize);

                   if (entry !== null) {
                       resetDialogComponents(actionType, entry);
                   } else {
                       resetDialogComponents(actionType);
                   }

                   self.actionType = actionType;

                   /* The inode 32 option cannot be modified afterwards */
                   /*  When action is modification, switch is set to readonly */
                   if (actionType === 'modify') {
                       self.inode32OptReadOnly(true);
                   } else {
                       self.inode32OptReadOnly(false);
                   }

                   self.existingShareNames = shareNameList;

                   if (entry !== null && entry !== undefined) {
                       storagePoolSVC.getShareExports(entry.name).then(
                           function (list) {
                               list.forEach(
                                   function(e) {
                                       var newExport = storageTypeFactory.newShareExport(e.type);
                                       // add the state property and the id used
                                       // in tiles
                                       var _merged;
                                       if (actionType === 'duplicate') {
                                           /* in case of duplication, all exports must be seen as new */
                                           _merged = $.extend(newExport, e,
                                                              {
                                                                  'state':'added',
                                                                  'exportId' : Helpers.genUid()
                                                              });
                                       } else {
                                           _merged = $.extend(newExport, e,
                                                              {
                                                                  'state':'existing',
                                                                  'exportId' : Helpers.genUid()
                                                              });
                                       }


                                       $.extend( e,_merged);

                                       // keep a carbon copy of hosts attribute
                                       // as this lead to deletion of it change
                                       if (e.type === 'nfs') {
                                           e.previousHosts = e.hosts;
                                       }
                                       if (e.type === 'smb') {
                                           e.previousResource = e.resource;
                                       }
                                   });
                               if (actionType === 'duplicate') {
                                   self.allExports(list.map(storageTypeFactory.newShareExport));
                               } else {
                                   /* add exports without transformation */
                                   self.allExports(list);
                               }
                           }).catch(
                               function(why){
                                   // enough for now
                                   oj.Logger.warn('cannot get share\'s exports : '+ why);
                               });
                   }

                   oj.Logger.log('size min:'+self.shareMinSize() + ' size max:' +
                                 self.shareMaxSize() + ' size :' + self.shareSize());
                   /* reset dialog size */
                   $(self.dialogSelector).css('width', self.shareDialogDefaultWidth);
                   $(self.dialogSelector).css('height', self.shareDialogDefaultHeight);

                   self.currentSelectedExportTileId = undefined;

                   $(self.dialogSelector).ojDialog('open');
               };

               this.handleCancelClick = function ()
               {
                   $(self.dialogSelector).ojDialog("close");
               };

               /**
                * check out for default values and other placeholder
                * according to export type type to send correct input
                * to service layer
                */
               function assignMandatoryValues(exportData) {
                   if (exportData.type === 'nfs') {
                       if (exportData.hosts === undefined) {
                           // in UI undefined means 'all'
                           // in service layer this is '*'
                           exportData.hosts = '*';
                       }
                   }
                   return exportData;
               }
               /**
                * check if the export of the identifeir has changed.
                * i.e host for nfs , resource for samba
                * return true if changed false otherwsie
                */
               function isExportIdentifierChanged(exportData) {
                   if (exportData.type === 'nfs') {
                       if (exportData.previousHosts !== exportData.hosts) {
                           return true;
                       }
                   }
                   if (exportData.type === 'smb') {
                       if (exportData.previousResource!== exportData.resource) {
                           return true;
                       }
                   }
                   return false;
               }
               /**
                * build a share export object
                * ready for deletion. In case of user change identifier
                * of the export we use the previous value
                *
                */
               function getExportDataForDeletion(exportData) {
                   if (exportData.type === 'nfs') {
                       // overwrite hosts with previous value
                       return $.extend({},exportData,{'hosts' : exportData.previousHosts});
                   }
                   if (exportData.type === 'smb') {
                       return $.extend({},exportData,{'resource' : exportData.previousResource});
                   }
                   return {};
               }
               this.submitShareForm = function () {
                   oj.Logger.log('submitShareForm for ' + self.actionType);
                   if (self.actionType === 'create' || self.actionType === 'duplicate') {
                       return storagePoolSVC.addShare(self.shareName(), self.shareSize(),
                                                      self.inode32Option()).then(
                                                          function (data) {
                                                              return Promise.all(
                                                                  ko.utils.arrayMap(self.allExports(),
                                                                                    function(exp) {
                                                                                        return storagePoolSVC.addShareExport(
                                                                                            self.shareName(),
                                                                                            assignMandatoryValues(exp));
                                                                                    }));
                                                          }).catch(function (message) {
                                                              oj.Logger.warn(message);
                                                          });
                   }

                   /** existing share modification **/

                   /** first check for removed and modified exports **/
                   /** construct and array of promise with modification type **/
                   var exportsPromises = ko.utils.arrayMap(self.allExports(),function(exp) {
                       if (exp.modified !== undefined && exp.modified === true) {
                           if (isExportIdentifierChanged(exp)) {
                               // modifying identifier lead to deletion first
                               return storagePoolSVC.deleteShareExport(
                                   self.shareName(),getExportDataForDeletion(exp)).then(
                                       storagePoolSVC.addShareExport(self.shareName(),assignMandatoryValues(exp)));
                           } else {
                               return storagePoolSVC.modifyShareExport(self.shareName(),assignMandatoryValues(exp));
                           }
                       } else if (exp.state === 'removed') {
                           return storagePoolSVC.deleteShareExport(self.shareName(),
                                                                   assignMandatoryValues(exp));
                       } else if (exp.state === 'added') {
                           return storagePoolSVC.addShareExport(self.shareName(),assignMandatoryValues(exp));
                       }
                       return Promise.resolve(true);
                   });
                   if (self.isShareModified === false) {
                       /* only modify export of the share*/
                       return Promise.all(exportsPromises);
                   }
                   return storagePoolSVC.modifyShare(self.shareName(),self.shareSize()).then (
                       function () {
                           return Promise.all(exportsPromises);
                       });
               };

               /**
                * modify, create or duplicate a share depending on the
                * current dialog object actionType
                */
               this.handleActionClick = function (event)
               {
                   self.submitShareForm().then(function() {
                       //$(self.dialogSelector).ojDialog("close");
                   }).catch(function(why) {
                       confirmation.showDialog({
                           'title' :  self.shareOperationFailedTitle,
                           'message': self.shareOperationFailedMessage,
                           'longMsg' : why,
                           'notifyOnly': true
                       });
                       oj.Logger.warn('Cannot submit form : ' + why);
                   });
                   // TODO : this should be within 'then' statement
                   //        until calls to backend become asynchronous
                   //        move the close of the window here.
                   $(self.dialogSelector).ojDialog("close");
                   return true;
               };
               this.handleCommandsClick = function (event)
               {
                    oj.Logger.log('inside handleCommandsClick');
                    // get the event target : the button
                    var target = event.target;
                    // get the tile i.e the closest element with oj-panel class
                    var tile = $(target).closest(".storage-add-share-export-tile");
                    // get the layout i.e the closest oj-masonrylayout
                    var masonryLayout = tile.closest(":oj-masonrylayout");
                    oj.Assert.assert((masonryLayout[0] !== undefined), 'missing css class to locate tile !');

                    var exportID = parseInt(tile.attr("id").substring('share_export_tile_'.length));
                    var exp = undefined;
                    self.allExports().forEach(function (item) {
                        if (item.exportId === exportID) {
                           exp = item;
                        }
                    });

                    clientCommandsDialog.open(self.shareName(), exp);
               };
               /**
                * callback when 'dismiss' button is clicked on the tile
                * grab the tile id and ask the layout to do the remove
                */
               this.exportTileRemove = function (event)
               {
                   oj.Logger.log('inside exportTileRemove');
                   // get the event target : the button
                   var target = event.target;
                   // get the tile i.e the closest element with oj-panel class
                   var tile = $(target).closest(".storage-add-share-export-tile");
                   // get the layout i.e the closest oj-masonrylayout
                   var masonryLayout = tile.closest(":oj-masonrylayout");
                   oj.Assert.assert((masonryLayout[0] !== undefined), 'missing css class to locate tile !');

                   // flush all validation errors for that tile
                   // ID on tile is prefixed by 'share_export_tile_<id>' see add-share-dialog.html
                   var exportID = parseInt(tile.attr("id").substring('share_export_tile_'.length));
                   self.nfs.NFSExportIdentifierOptionValidator.errorRaised.remove(exportID);
                   self.cifs.CIFSexportIdentifierOptionValidator.errorRaised.remove(exportID);
                   self.nfs.NFSIPsOptionValidator.errorRaised.remove(exportID);
                   self.cifs.CIFSHostOptionValidator.errorRaised.remove(exportID);
                   self.cifs.CIFSResourceOptionValidator.errorRaised.remove(exportID);

                   masonryLayout[0].removeTile("#" + tile.attr("id"));
               };
               /**
                * handle the event raised when a tile within the masonry layout is removed
                * remove the export form the list
                */
               this.exportsLayoutHandleRemove = function (event)
               {
                   var tile = event.detail.tile;
                   var context = ko.contextFor(tile);
                   var exportData = context.$data;
                   var removedExport = undefined;
                   self.allExports().forEach(function (item) {
                       if (item.exportId === exportData.exportId) {
                           removedExport = item;
                       }
                   });
                   /* check out the state of the export */
                   /* if it was a new one just forget about it*/
                   if (removedExport.state === 'added') {
                       self.allExports.remove(removedExport);
                   } else {
                       /* tag it for removal */
                       removedExport.state = 'removed';
                   }
               };

               /**
                * called by 'add' button : add a new export into the layout
                */
               this.exportsLayoutAdd = function (event)
               {
                   oj.Assert.assert((event.target.value === 'smb') || (event.target.value === 'nfs'));
                   var newExport = storageTypeFactory.newShareExport(event.target.value);
                   newExport.exportId = Helpers.genUid();
                   newExport.state = 'added';
                   self.allExports.unshift(newExport);
               };
               /**
                * validate that id is unique across all NFS exports
                */
               this.nfs.NFSExportIdentifierOptionValidator = {
                   getHint: function () {
                       return self.nfs.IPsOptionHint;
                   },
                   errorRaised: ko.observableArray([]),
                   validate: function(value) {
                       return exportIdentifierOptionValidate(this.errorRaised,'nfs', 'hosts', value);
                   }
               };

               /**
                * validate that id is unique across all NFS exports
                */
               this.cifs.CIFSexportIdentifierOptionValidator = {
                   getHint: function () {
                       return self.cifs.CIFSResourceOptionHint;
                   },
                   errorRaised: ko.observableArray([]),
                   validate:function(value) {
                       return exportIdentifierOptionValidate(this.errorRaised,'smb', 'resource', value);
                   }
               };

               /**
                * check all exports for id collision by export type
                * @param {ko.observableArray}  errors -  observable array to be filled with current export tile id in case of error
                * @param {string} exportType - type of export ('nfs or 'smb')
                * @param {string} identifierKey - the key to be check within export JSON object
                * @param {string} value - the id value to be searched for. The validator will raise the error if 'value' is found
                * @throws {oj.ValidationError} - if 'value' is found within other export object.
                */
               function exportIdentifierOptionValidate (errors, exportType, identifierKey, value) {

                   var exp = undefined;
                   self.allExports().forEach(function (item) {
                       if (item[identifierKey] === value) {
                           exp = item;
                       }
                   });
                   if (exp !== undefined) {
                       errors.push(self.currentSelectedExportTileId);
                       throw new oj.ValidatorError(self.exportIdentifierOptionValidatorHint);
                   } else {
                       errors.remove(self.currentSelectedExportTileId);
                   }
               }
               /**
                * check that share name respect naming convention
                */
               this.shareNameValidator = {
                   getHint: function () {
                       return self.nameOptionValidationCharacterMessageDetails;
                   },
                   errorRaised: ko.observable(false),
                   validate: function(value) {
                       if (/[a-zA-Z0-9_-]{1,}/.test(value)) {
                           this.errorRaised(false);
                           return true;
                       }
                       this.errorRaised(true);
                       throw new oj.ValidatorError(self.nameOptionValidationCharacterMessageDetails);
                   }
               };
               /**
                * checks that share name is unique
                */
               this.duplicateNameOptionValidator = {
                   errorRaised : ko.observable(false),
                   validate: function(value) {
                       if (self.existingShareNames.indexOf(value) === -1) {
                           this.errorRaised(false);
                           return true;
                       }
                       this.errorRaised(true);
                       throw new oj.ValidatorError(self.nameOptionValidationDuplicateMessageDetails);
                   }
               };

               /**
                * value change event listener for share size slider
                */
               this.shareSizeOptionChangeListener = function (event) {
                   oj.Logger.log('value-changed on '+ event.target.id + ' ' + event.detail.previousValue +
                                 ' => ' + event.detail.value);
                   oj.Logger.log('size min:'+self.shareMinSize() + ' size max:' +
                                 self.shareMaxSize() + ' size :' + self.shareSize());
                   if (self.dialogOpened) {
                       // only if coming from user input
                       // enough as we do not update this value from VM
                       self.isShareModified = true;
                   }
               };
               this.shareSizeInputValidator = ko.computed(
                   function()
                   {
                       return [{
                           type: 'numberRange',
                           options: {
                               min: self.shareMinSize(),
                               max: self.shareMaxSize()
                               }}];
                   });
               /**
                * value change event listener for share options
                */
               this.shareOptionChangeListener = function (event) {

                   if (self.dialogOpened) {
                       // only if coing from user input
                       // enough as we do not update this value from VM
                       self.isShareModified = true;
                   }
               };
               this.dialogOpened = false;
               /**
                * listener to dialog open/close
                * we only consider change made to components when window is opened (i.e coming from user action)
                */
               this.shareDialogOpenCloseListener = function (event) {
                   if (event.type === 'ojClose') {
                       self.dialogOpened = false;
                   }
                   if (event.type === 'ojOpen') {
                       self.dialogOpened = true;
                   }
               };
               /**
                * listener to sync always option. If the user enables the sync always option, strict  sync
                * must be enabled. So we automatically enable strict sync when sync always is enabled.
                * this listener used by 'strict sync option' in add-share-cifs-export.html
                */
               this.exportSAChangeListener = function(event) {
                   if (event.detail.value === true) {
                       // take care about dependents
                       var _form = event.target.closest('.oj-form');
                       oj.Assert.assert(_form !== null);
                       // get the specific export
                       var exportId = parseInt(_form.getAttribute('exportId'));
                       // enable the strict sync option for that particular export
                       self.allExports().forEach(function (item) {
                           if (item.exportId === exportId) {
                               item.options.strict_sync = true;
                           }
                       });
                       // traverse all the switches and enable the option(strict sync), identified by class
                       var child_switches = $(event.target.closest('.oj-form')).find('.oj-switch');
                       for ( var i = 0; i < child_switches.length; i++) {
                           if ($(child_switches[i]).hasClass('sa_dependent')) {
                               child_switches[i].value = true;
                           }
                       }
                   }
                   self.exportOptionChangeListener(event);
               };
               /**
                * called when a component in export tile changed
                * the associated form hold for us the export ID
                * we mark it as "changed"
                */
               this.exportOptionChangeListener = function(event) {
                   var _form = event.target.closest('.oj-form');
                   oj.Assert.assert(_form !== null);
                   var exportId = parseInt(_form.getAttribute('exportId'));
                   self.allExports().forEach(function (item) {
                       if (item.exportId === exportId) {
                           if (item.state !== 'added') {
                               item.modified = true;
                           }
                       }
                   });
                   self.currentSelectedExportTileId = exportId;
               };

               /* validation must be within tile context */
               /* I.e validation on one tile must not affect overall dialog validation */
               /* or other tiles validation */
               /* so we keep alwasy the selected tile context */
               this.currentSelectedExportTileId = undefined;

               this.exportTileComponentListener = function(event) {
                   var exportId = parseInt(event.target.getAttribute('exportId'));
                   self.currentSelectedExportTileId = exportId;
               };
               this.cifs.CIFSResourceOptionValidator = {
                   /* errorRaised is within the context of one export tile: keep an array of exportTile IDs */
                   errorRaised:ko.observableArray([]),
                   validate: function(value) {
                       if (/[a-zA-Z\\[0-9]]?/.test(value)) {
                           this.errorRaised.remove(self.currentSelectedExportTileId);
                           return true;
                       }
                       this.errorRaised.push(self.currentSelectedExportTileId);
                       throw new oj.ValidatorError(AddShareViewModel.prototype.cifs.resourceOptionMessageDetails);
                   }
               };
               this.cifs.CIFSHostOptionValidator = {
                   errorRaised: ko.observableArray([]),
                   validate: function(value) {
                       var trimmedValue = value.trim();
                       if (trimmedValue === '') {
                           // nothing to do
                           this.errorRaised.remove(self.currentSelectedExportTileId);
                           return true;
                       }
                       // ip address ?
                       if (/^(?!0)(?!.*\.$)((1?\d?\d|25[0-5]|2[0-4]\d)(\.|$)){4}$/.test(trimmedValue)) {
                           this.errorRaised.remove(self.currentSelectedExportTileId);
                           return true;
                       }
                       //any kind o hostname
                       if (/[a-zA-Z\.[0-9]]?/.test(trimmedValue)) {
                           this.errorRaised.remove(self.currentSelectedExportTileId);
                           return true;
                       }
                       this.errorRaised.push(self.currentSelectedExportTileId);
                       throw new oj.ValidatorError(AddShareViewModel.prototype.cifs.resourceOptionMessageDetails);
                   }
           };



               /**
                * special case for 'host' option as this lead to delete/create operation
                */
               this.exportIdentifierOptionChangeListener = function(event) {
                   var exportId = parseInt(event.target.getAttribute('exportId'));
                   self.allExports().forEach(function (item) {
                       if(item.exportId === exportId) {
                           if (item.state !== 'added') {
                               item.modified = true;
                           }
                       }
                   });

                   /* we must trigger an update, assigning the same value */
                   /* do not trigger an update -> check the current value first */
                   if (self.exportIdentifierOptionChanged.peek()) {
                       self.exportIdentifierOptionChanged(false);
                   } else {
                       self.exportIdentifierOptionChanged(true);
                   }
               };

               /* for export identifier ghange we need to re-evaluate */
               /* the state of 'create' button (see shouldDisableCreateButton) */
               /* for new export creation we need an observalbe to trigger the computeObservable */
               /* (re)evaluation */
               this.exportIdentifierOptionChanged = ko.observable(false);

               this.nfs.NFSIPsOptionValidator = {
               errorRaised: ko.observableArray([]),
               validate: function(value) {
                   if (value === '' || value === '*') {
                       //everyone
                       this.errorRaised.remove(self.currentSelectedExportTileId);
                       return true;
                   }
                   // ip address ?
                   if (/^(?!0)(?!.*\.$)((1?\d?\d|25[0-5]|2[0-4]\d)(\.|$)){4}$/.test(value)) {
                       this.errorRaised.remove(self.currentSelectedExportTileId);
                       return true;
                   }
                   //netgroup
                   if (/^@[a-zA-Z]+/.test(value)) {
                       this.errorRaised.remove(self.currentSelectedExportTileId);
                       return true;
                   }
                   //any kind o hostname
                   if (/[a-zA-Z\.[0-9]]?/.test(value)) {
                       this.errorRaised.remove(self.currentSelectedExportTileId);
                       return true;
                   }
                   this.errorRaised.push(self.currentSelectedExportTileId);
                   throw new oj.ValidatorError(AddShareViewModel.prototype.nfs.IPsOptionMessageDetails);
               }
           };


           }

           $.extend(AddShareViewModel.prototype,CommonVMDefinitons.prototype.definitions);
           $.extend(AddShareViewModel.prototype,CommonVMDefinitons.prototype.converters);
           AddShareViewModel.prototype.dialogSelector='#addShareDialog';


           AddShareViewModel.prototype.nameOptionLabel = oj.Translations.getTranslatedString('Share name');

           AddShareViewModel.prototype.nameOptionTip = oj.Translations.getTranslatedString('Name of the new share. For NFS exports, it will be used to form the exported path /shares/<share name>.');
           AddShareViewModel.prototype.nameOptionValidationCharacterMessageDetails = oj.Translations.getTranslatedString('Any alphanumeric character. Cannot contain \'/\'.');
           AddShareViewModel.prototype.nameOptionValidationDuplicateMessageDetails = oj.Translations.getTranslatedString('Share name already exists');
           AddShareViewModel.prototype.nameOptionValidationDuplicateHint = oj.Translations.getTranslatedString('Share name must be unique.');
           AddShareViewModel.prototype.sizeOptionLabel = oj.Translations.getTranslatedString('Share size');

           AddShareViewModel.prototype.sizeOptionTip = oj.Translations.getTranslatedString('Size of the share (for example 10G, 5T, 10gb or 2.5tb)');

           AddShareViewModel.prototype.exportIdentifierOptionValidatorHint = oj.Translations.getTranslatedString('Export of the same type cannot share the same identifier');

           AddShareViewModel.prototype.nfs = {
               'IPsOptionLabel' :  oj.Translations.getTranslatedString('Hosts'),
               'IPsOptionHint' :  oj.Translations.getTranslatedString('Hosts string must be unique between NFS exports for this share'),
               'IPsOptionPlaceHolder' : oj.Translations.getTranslatedString('*'),
               'IPsOptionTip' : oj.Translations.getTranslatedString('Specify the authorized clients to access this share as specified in exports(5). * for any host or hostname or IP specification. Default value is * (any host).'),
               'IPsOptionMessageDetails' : oj.Translations.getTranslatedString('Must follow client specification as described in exports(5)'),
               'syncOptionLabel' : oj.Translations.getTranslatedString('Synchronous operations'),
               'syncOptionTip' : oj.Translations.getTranslatedString('Reply to requests only after the changes have been committed to stable storage'),
               'subtreeCheckOptionLabel' : oj.Translations.getTranslatedString('Subtree checking'),
               'subtreeCheckOptionTip' : oj.Translations.getTranslatedString('See exports(5) man page for details'),
               'secureOptionLabel' : oj.Translations.getTranslatedString('Require privileged port'),
               'secureOptionTip' : oj.Translations.getTranslatedString('Requires that requests originate on an Internet port less than IPPORT_RESERVED (1024)'),
               'hideOptionLabel' : oj.Translations.getTranslatedString('Hide nested filesystems'),
               'hideOptionTip' : oj.Translations.getTranslatedString('See exports(5) man page for details'),
               'rootSquashOptionLabel' : oj.Translations.getTranslatedString('Root squashing'),
               'rootSquashOptionTip' : oj.Translations.getTranslatedString('Map requests from uid/gid 0 to the anonymous uid/gid')
           };
           AddShareViewModel.prototype.cifs = {
               'CIFSResourceOptionPlaceHolder' : oj.Translations.getTranslatedString('Enter name of the SMB export'),
               'CIFSResourceOptionHint' : oj.Translations.getTranslatedString('Name must be unique between SMB exports for this share'),
               'resourceOptionLabel' :  oj.Translations.getTranslatedString('SMB share name'),
               'resourceOptionTip' :  oj.Translations.getTranslatedString('Folder name to be used by this Samba share'),
               'resourceOptionMessageDetails' : oj.Translations.getTranslatedString('Must follow client specification as described in cifs(5)'),
               'commentOptionLabel' : oj.Translations.getTranslatedString('Comment'),
               'commentOptionTip' : oj.Translations.getTranslatedString('Textual description of the share'),
               'browsableOptionLabel' : oj.Translations.getTranslatedString('Browsable'),
               'browsableOptionTip' : oj.Translations.getTranslatedString('Allow share to be seen in the list of available shares in a net view and in the browse list'),
               'strictSyncOptionLabel' : oj.Translations.getTranslatedString('Strict sync'),
               'strictSyncOptionTip' : oj.Translations.getTranslatedString('Honors applications requests for sync calls. Enabling this option may slow down the throughput, use with caution. See smb.conf(5) man page for details.'),
               'syncAlwaysOptionLabel' : oj.Translations.getTranslatedString('Sync always'),
               'syncAlwaysOptionTip' : oj.Translations.getTranslatedString('Enforce writes to stage storage before the write call returns. See smf.conf(5) man page for details'),
               'guestOkOptionLabel' : oj.Translations.getTranslatedString('Guest ok'),
               'guestOkOptionTip' : oj.Translations.getTranslatedString('Allow connecting to this export without password. Privileges will be those of the guest account. see smb.conf(5) man page for details.'),
               'hostsAllowOptionLabel' : oj.Translations.getTranslatedString('Hosts allowed'),
               'hostsAllowOptionTip' : oj.Translations.getTranslatedString('Client systems that can connect to the share. Expected format: comma separated list of hosts (specified by name or IP address).'),
               'hostsDenyOptionLabel' : oj.Translations.getTranslatedString('Hosts denied'),
               'hostsDenyOptionTip' : oj.Translations.getTranslatedString('Client systems that CANNOT connect to the share. Expected format: comma separated list of hosts (specified by name or IP address).')
           };


           AddShareViewModel.prototype.readOnlyOptionLabel = oj.Translations.getTranslatedString('Read only');
           AddShareViewModel.prototype.readOnlyOptionTip= oj.Translations.getTranslatedString('Allow only read requests on this volume');
           AddShareViewModel.prototype.shareOptionLabel = oj.Translations.getTranslatedString('Export options');

           AddShareViewModel.prototype.optionsLabel = oj.Translations.getTranslatedString('Options');

           AddShareViewModel.prototype.exportsSectionLabel = oj.Translations.getTranslatedString('Export Protocols');
           AddShareViewModel.prototype.addExportButtonLabel = oj.Translations.getTranslatedString('Add export...');
           AddShareViewModel.prototype.addExportNFSButtonLabel = oj.Translations.getTranslatedString('NFS export');
           AddShareViewModel.prototype.addExportCIFSButtonLabel = oj.Translations.getTranslatedString('SMB export');



           AddShareViewModel.prototype.inode32Label = 'Use 32 bits inodes only';
           AddShareViewModel.prototype.inode32OptionTip = 'Needed to run legacy 32 bits applications on filesystem bigger than 1T. See man xfs(5) for details.';

           AddShareViewModel.prototype.shareOperationFailedMessage = oj.Translations.getTranslatedString('Failed to modify share');
           AddShareViewModel.prototype.shareOperationFailedTitle = oj.Translations.getTranslatedString('Operation has failed');


           return new AddShareViewModel();
       }
      );
