/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */
define([
        'knockout',
        'ojs/ojcore',
        'ojs/ojfilmstrip',
        'ojs/ojknockout',
        'ojs/ojbutton',
        'ojs/ojdialog'
    ],
    function (ko, oj)
    {
        'use strict';

        /* Translatable strings */
        var msgDialogTitle = ('Help');


        function HelpDialog()
        {

            var self = this;
            /* Merge opts with default values to create model for dialog */
            this.title = msgDialogTitle;
            /**
             * Shows the help dialog.
             * @param {integer} [pageId] the current help item to be displayed
             * if not specified the welcome page is shown
             */
            this.show = function (pageId)
            {
                var _pageToBeOpened = _getCurrentItemId(pageId);
                $('#helpDialog').on('ojopen',
                                    function (event, ui)
                                    {
                                        self.currDetailPage(_pageToBeOpened);
                                    });
                $('#helpDialog').ojDialog('open');
            };


            self.currDetailPage = ko.observable(0);

            /**
             * Returns an index (into topics array) for the help page that
             * should initially be displayed.
             * @param {integer} [pageId] the id of the page to be displayed.
             * @returns the currentItem of master details to be displayed
             */
            function _getCurrentItemId(pageId)
            {
                var i, routeRegExp,
                    longestMatch = undefined,
                    longestMatchIndex = undefined;

                if (pageId !== undefined) {
                    oj.Assert.assertNumber(pageId, 'Invalid page id');
                }
                var _topicLength = self.topics.length;

                /*
                 * If we were called with a specific page Id,
                 * then check that first
                 */
                if (pageId) {
                    for (i = 0; i < _topicLength; i++)
                    {
                        if (self.topics[i].id === pageId)
                        {
                            return (i);
                        }
                    }
                    /* if we are here the page is wrong */
                    oj.Assert.assert(false, 'page ID ' + pageId + ' not found');
                }

                /* Check for a page whose route matches the current route */
                var _currentRoute = window.location.pathname.replace(/\/$/, '');

                /* Check for an exact match first */
                for (i = 0; i < _topicLength; i++)
                {
                    if (_currentRoute === self.topics[i].route)
                    {
                        return (i);
                    }
                }

                /*
                 * If we haven't found a match yet, try a looser match against
                 * the current route using wildcards
                 */
                for (i = 0; i < _topicLength; i++)
                {
                    if ('route' in self.topics[i]) {
                        routeRegExp =
                            new RegExp('^' + self.topics[i].route + '*');
                        if (routeRegExp.test(_currentRoute))
                        {
                            if (longestMatch === undefined ||
                                self.topics[i].route.length >
                                longestMatch.length)
                            {
                                longestMatch = self.topics[i].route;
                                longestMatchIndex = i;
                            }
                        }
                    }
                }

                if (longestMatchIndex)
                {
                    return (longestMatchIndex);
                }

                /* If all else fails, just start with the first page */
                return (0);
            }



            this.handleDetailCreate = function ()
            {
                var pagingModel =
                    $('#helpTopicDetail').ojFilmStrip('getPagingModel');
                pagingModel.on('beforePage', self.handleDetailBeforePage);
                self.handleDetailBeforePage();
            };



            this.handleDetailBeforePage = function (event)
            {
                var pagingModel =
                    $('#helpTopicDetail').ojFilmStrip('getPagingModel'),
                    pageCount = pagingModel.getPageCount(),
                    pageIndex = event ? event.page : 0;

                /*
                 * wrap in try..catch block because the detail filmstrip is
                 * initializedfirst and the master filmstrip is not yet
                 * available.
                 */
                try
                {
                    var topicList = $('#helpTopicList'),
                        masterPagingModel =
                        topicList.ojFilmStrip('getPagingModel'),
                        masterPage = masterPagingModel.getPage(),
                        masterItemsPerPage =
                        topicList.ojFilmStrip('getItemsPerPage'),
                        newMasterPage = Math.floor(pageIndex /
                            masterItemsPerPage);

                    /*
                     * change master page in detail's beforePage event so that
                     * the master page change transitions at the same
                     * time as the detail page
                     */
                    if (newMasterPage !== masterPage)
                    {
                        masterPagingModel.setPage(newMasterPage);
                    }
                }
                catch (e)
                {
                    /* Do nothing */
                }
            };

            this.topicClick = function (data, event)
            {

                var context = ko.contextFor(event.target),
                    index = context.$index();
                self.currDetailPage(index);
            };

            this.getDetailItemInitialDisplay = function (index)
            {
                return (index < 1 ? '' : 'none');
            };

            this.getMasterItemInitialDisplay = function (index)
            {
                return (index < 4 ? '' : 'none');
            };

            this.getMasterItemLabelId = function (index)
            {
                return ('masterLabel' + index);
            };

            this.getMasterItemSelected = function (index)
            {
                return (self.currDetailPage() === index ? 'true' :
                    'false');
            };

            this.isMasterItemSelected = function (index)
            {
                return (self.getMasterItemSelected(index) === 'true');
            };

        }
        HelpDialog.prototype.topics = [
        {
            name: ('Welcome'),
            subName: undefined,
            id: 'welcome'
        },
        {
            name: ('Dashboard'),
            subName: undefined,
            id: 'dash'
        },
        {
            name: ('Storage'),
            subName: undefined,
            id: 'storage'
        },
        {
            name: ('Backup & Restore'),
            subName: undefined,
            id: 'appliance'
        },
        {
            name: ('Appliance'),
            subName: undefined,
            id: 'appliance'
        },
        {
            name: ('Administration'),
            subName: undefined,
            id: 'Administration'
        },
        {
            name: ('Shape Migration'),
            subName: undefined,
            id: 'migration'
        },
        {
            name: ('Release Notes'),
            subName: undefined,
            id: undefined
        }];

        /* Used also by helpContent view */
        /* Must be a singleton    */
        return (new HelpDialog());
    });
