# Copyright (c) 2018, Oracle and/or its affiliates. All rights reserved.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
# flask based web application
#  backend calls are delegated to:
#  - backend if they don't require privileges
#  - priv_exec if they require privileged execution
#  system calls shouldn't be made in this file

from flask import jsonify, Blueprint, render_template, request
from flask.ext.login import login_required, current_user

from flask import current_app

import psutil
import time
import multiprocessing
import backend_test as web_backend
import logging
from backend.common import LV_DEFAULT_PHYSICAL_EXTENT_SIZE, RAID_LEVEL_NONE, SMB_EXPORT_CONF, LSA_STATE_BIN

from backend.lsa_types import ExtJSONEncoder as LSAExtJSONEncoder
import json
from backend.share_util import ShareEntry
from backend.config_manager import NfsConfigManager, SambaConfigManager
import subprocess
from ..doc.views import auto
import backend.utility

lsa = Blueprint('lsa', __name__,
                template_folder='templates',
                static_folder='static')


def createExpResp(str):
    ''' create a standard exception response, str is used to pass additional
    information '''
    current_app.logger.error(str)
    resp = jsonify({'msg': str})
    resp.status_code = 500
    return resp


@lsa.route('/pool/utilization')
@auto.doc(groups=['private', 'internal'])
@login_required
def pool_utilization():
    """
    Storage pool utilization
    Returns storage pool utilization as JSON Object
    Object properties are
       total: total space in bytes as number
       free: free space in bytes as number
    example:
       {
        'total': 12884901888,
        'free' : 4294967296
       }
    """
    try:
        pool = web_backend.get_storage_pool()
        return jsonify(total=pool.get_size(), free=pool.get_free(), status=pool.get_status())
    except StandardError, e:
        return createExpResp('/pool/utilization failed: %s' % str(e))


@lsa.route('/pool', methods=['POST'])
@auto.doc(groups=['private', 'internal'])
@login_required
def create_pool():
    """
    Creates the storage pool from a list of devices
    expected arguments within the request from are:
        raid_level: raid level ('raid0', 'raid1', 'raid6', 'raid10', or 'none')
        devices: list of comma-separated paths (e.g., '/dev/sdb,/dev/sdc')
    raise:
         KeyError: missing/unsupported parameter
    """
    # retrieve parameters
    raid_level = request.form.get('raid_level')
    if not raid_level:
        raise KeyError('missing parameter: raid_level')
    if raid_level == 'raid0':
        int_raid_level = 0
    elif raid_level == 'raid1':
        int_raid_level = 1
    elif raid_level == 'raid6':
        int_raid_level = 6
    elif raid_level == 'raid10':
        int_raid_level = 10
    elif raid_level == 'none':
        int_raid_level = RAID_LEVEL_NONE
    else:
        raise KeyError('unsupported raid_level: %s' % raid_level)
    devices = request.form.get('devices')
    if not devices:
        raise KeyError('missing parameter: devices')
    try:
        web_backend.create_pool(int_raid_level, devices)
    except Exception, e:
        return createExpResp('/pool failed: %s' % str(e))

    return ('', 201)


@lsa.route('/pool/status')
@auto.doc(groups=['private', 'internal'])
@login_required
def pool_status():
    """
    Gets the storage pool status/information
    Returns storage pool information  as JSON Object
    Object properties are
       status: pool status 'active' 'degraded', 'recovery', 'unknown', 'unavailable', or 'faulty'
       raid_level: raid level of the pool 6 or 10 etc..
       devices: JSON object array of RAID array devices
    RAID array devices are
       dev: device name as string (ex /dev/sdb)
       state: state of the device as string : faulty,active,spare,removed

    example :
      {
        'status': 'active',
        'num_devs': '4'
        'protection': 'raid10', null for block based volumes.
         'devices': [
                     {
                       'dev': '/dev/sda'
                       'state': 'active'
                     },
                     {
                       'dev': '/dev/sdb'
                       'state': 'active'
                     },
                     {
                       'dev': '/dev/sdc'
                       'state': 'spare'
                     },
                    ]
      }


    """
    current_app.logger.debug('Inside pool_status')
    try:
        pool = web_backend.get_storage_pool()
        raidStatus = {
            'status': pool.get_status(),
            'num_devs': pool.get_num_devs(),
            'protection': pool.get_protection(),
            'devices': [{'dev': d.name, 'state': d.state} for d in pool.get_devices()]
        }
        return jsonify(raidStatus)
    except Exception, e:
        return createExpResp('/pool/status failed: %s' % str(e))


@lsa.route('/pool/nfs/clients')
@auto.doc(groups=['internal'])
@login_required
def get_nfs_clients():
    """
    Gets the (possibly empty) list of connected NFS clients

    Object properties are
       list: array of connected client IP as string

    example:
       {
        'list': ['10.166.104.66','10.166.104.6']
       }
    """
    try:
        res = web_backend.get_nfs_clients()
    except Exception, e:
        return createExpResp('/pool/nfs/clients failed: %s' % str(e))

    return jsonify(list=res)


@lsa.route('/storageUsages')
@auto.doc(groups=['private', 'internal'])
@login_required
def storageUsages():
    """
    Gets all storages usages
    Returns all storages usages as JSON Object
    Object property is storage name 'root' or 'pool'
    Values are float representing usages as percentage
    Example:
       {
        'root': 14.6,
        'pool': 80.4
       }

    """
    try:
        usage = web_backend.storageUsages()
    except Exception, e:
        return createExpResp('/storageUsages failed: %s' % str(e))

    # TODO : should be per storage verb like we have in /pool/utilization
    # something like /storage/root and /storage/pool /storages
    #  and return {'name','mnt','size',used}
    return jsonify({'root': usage.get('/'),
                    'pool': usage.get('data')})


def get_uptime():
    """
    returns the number of seconds the system has been up
    """
    with open('/proc/uptime', 'r') as f:
        uptime_seconds = float(f.readline().split()[0])

    return uptime_seconds


@lsa.route('/sys')
@auto.doc(groups=['private', 'internal'])
@login_required
def sysInfo():
    """
    Returns system information and configuration

    Returns a JSON object as follow
       {
         'hostname' : hostname as string,
         'ip' : IP address as string,
         'cpu' : cpu count as int
         'mem' : memory count in bytes as int
         'uptime' : EPOC uptime in ms as float
         'os_version': platform OS version as string
         'kernel_version': platform kernel version as string
         'platform_version': platform version as string
         'root_disk_size' : size in bytes as int of root filesystem
       }
    Example
       {
         'cpu' : 2,
         'root_disk_size' : 42811318272,
         'hostname' : 'foo.bar.com',
         'ip' : '10.166.175.251',
         'mem' : 3,
         'os_version' : 'Oracle Linux Server release 7.3',
         'kernel_version' : '4.1.12-61.1.18.el7uek.x86_64',
         'platform_version' : '0.0.3'
         'uptime' : 3708316.4
       }
    """
    hostname = backend.utility.get_hostname()
    ip = web_backend.get_host_ip_address()
    kernel_v = web_backend.get_kernel_version()
    os_v = web_backend.get_os_version()
    platform_v = web_backend.get_platform_version()
    return jsonify({'hostname': hostname, 'ip': ip,
                    'cpu': multiprocessing.cpu_count(),
                    'mem': (psutil.virtual_memory().total),
                    'root_disk_size': psutil.disk_usage('/').total,
                    'os_version': os_v,
                    'kernel_version': kernel_v,
                    'platform_version': platform_v,
                    'uptime': get_uptime()})


@lsa.route('/sys/usages')
@auto.doc(groups=['private', 'internal'])
@login_required
def allUsages():
    """
    Gets usages of system resources
    Returns a JSON Object of all system usages
    Returns a JSON object as follow
       {
         cpu: a float representing the current overall CPU usage percentage
         mem: a float representing the current memory usage
         mem_avail: available memory in bytes as int
         swap: the percentage of sawp used as float
         disk: root filesystem usage expressed as percentage
         disk_avail: root filesystem free space in bytes
         bytes_sent: system-wide network I/O statistics, number of bytes sent
         bytes_recv: system-wide network I/O statistics, number of bytes received
       }
    Example
      {
       'bytes_recv' : 7075879445,
       'bytes_sent' : 12808652318,
       'cpu'        : 1.1,
       'disk'       : 43.5,
       'disk_avail' : 24189227008,
       'mem'        : 15.9,
       'mem_avail'  : 3255631872,
       'swap'       : 0
      }
    """
    try:
        usage = web_backend.systemUsages()
    except Exception, e:
        return createExpResp('/sys/usages failed: %s' % str(e))

    # TODO : should not having percent directly as we loose information
    #        model should give us min,max and value
    return jsonify({'cpu': usage['cpu'],
                    'mem': usage['mem'],
                    'mem_avail': usage['mem_avail'],
                    'swap': usage['swap'],
                    'disk': usage['disk'],
                    'disk_avail': usage['disk_avail'],
                    'bytes_sent': usage['bytes_sent'],
                    'bytes_recv': usage['bytes_recv']})


@lsa.route('/shares', methods=['GET'])
@auto.doc(groups=['private', 'internal'])
@login_required
def get_shares():
    '''
    Get list of shares
    each share has the following format:
        {
            name: 'test1',
            size: 121151651,
            free: 1001515
            fs_options: {inode32: 'False'}
        }
    '''
    try:
        slist = web_backend.get_shares()
        return jsonify({'list': slist})
    except Exception, e:
        return createExpResp('/shares failed: %s' % str(e))


@lsa.route('/shares/<string:name>', methods=['GET'])
@auto.doc(groups=['private', 'internal'])
@login_required
def get_share(name):
    '''
    Get the share by name
    each share has the following format:
        {
            name: 'test1',
            size: '121151651',
            free: '1001515'
            fs_options: {inode32: 'False'}
        }
    '''
    try:
        shares = web_backend.get_shares([name])
        if not shares:
            return createExpResp('No share found in the system with name %s' % name)
        return jsonify(shares[0])
    except Exception, e:
        return createExpResp('/shares/%s failed: %s' % (name, str(e)))


@lsa.route('/shares/<string:name>/exports', methods=['GET'])
@auto.doc(groups=['private', 'internal'])
@login_required
def get_exports(name):
    '''
    Get the list of exports for a particular share.
    each export has the following format dictionary items

        format nfs export:
        {
            id: 'host'
            type: 'nfs'
            name: 'name of the share'
            hosts: host name or IP address
            path: NFS resource access point (path of the export for NFS clients)
            options: {
                read_only: boolean,
                sync: boolean,
                root_squash: boolean,
                secure: boolean,
                hide: boolean
            }
        }
        format samba export:
        {
            id: 'resource name'
            type: 'smb',
            resource: 'resource name as string'
            options: {
                path: 'path of the share: resource name',
                comment: 'comment as string',
                browsable: 'boolean',
                guest_ok: 'boolean',
                hosts_allow: 'single or comma separated IP addresses/hosts/network'
                hosts_deny: 'single or comma separated IP addresses/hosts/network'
            }
        }
        list might be empty.

    '''
    try:
        share = ShareEntry(name)
        nfs_manager = NfsConfigManager()
        smb_conf_manager = SambaConfigManager(SMB_EXPORT_CONF)
        nfs_exports = nfs_manager.get_exports(share)
        smb_exports = smb_conf_manager.get_exports(share)
        elist = []
        for export in nfs_exports:
            elist.append(export.serialize())
        for export in smb_exports:
            elist.append(export.serialize())
        return jsonify({'list': elist})
    except Exception, e:
        return createExpResp('/shares/%s/exports failed: %s' % (name, str(e)))


@lsa.route('/shares/<string:name>/exports/<path:export_id>', methods=['GET'])
@auto.doc(groups=['private', 'internal'])
@login_required
def get_export(name, export_id):
    '''
    Get an export for a particular share by export_ID.
    Format:
        format nfs export:
        {
            id: 'host'
            type: 'nfs'
            name: 'name of the share'
            hosts: host name or IP address
            path: path of the share
            options: {
                read_only: 'True/False'
                sync: True/False
                root_squash: True/False
                hide: True/False
                secure: True/False
            }
        }
        format samba export:
        {
            id: 'resource name'
            type: 'smb',
            resource: 'resource name as string'
            options: {
                read only: 'boolean'
                path: 'path of the share: resource name',
                comment: 'comment as string',
                browsable: 'boolean',
                guest_ok: 'boolean',
                hosts_allow: 'single or comma separated IP addresses/hosts/network'
                hosts_deny: 'single or comma separated IP addresses/hosts/network'
            }
        }
    '''
    try:
        share = ShareEntry(name)
        export = NfsConfigManager().get_export(share, export_id)
        return jsonify(export.serialize())
    except Exception, e:
        return createExpResp('/shares/%s/exports failed: %s' % (name, str(e)))


@lsa.route('/shares', methods=['POST'])
@auto.doc(groups=['private', 'internal'])
@login_required
def create_share():
    '''
    Creates a new share.
    expected argument within the request from are:
        'name': The name for the new share
        'size': size of the share in human readable format
        'inode32': workaround for legacy application limitation for 26879349
    raise:
         KeyError: missing parameter
         StandardError: error raised by backend
    return the new share name and the id of the long-running task for share creation
    '''
    # retrieve parameter
    request_json = request.get_json()
    share_name = request_json.get('name')
    size = request_json.get('size')
    if request_json.get('fs_options') is None:
        raise KeyError('missing \'fs_options\' parameter')
    inode32 = request_json.get('fs_options').get('inode32')

    if not share_name:
        raise KeyError('missing name parameter')
    if not size:
        raise KeyError('missing size parameter')
    if inode32 is None:
        raise KeyError('missing inode32 parameter')
    # size in bytes must be multiple of 4MB, as lvcreate requires multiple of 512 and rounds up to multiple of 4MB
    if size % LV_DEFAULT_PHYSICAL_EXTENT_SIZE:
        raise ValueError('Size of the share in bytes must be multiple of %d' % LV_DEFAULT_PHYSICAL_EXTENT_SIZE)
    try:
        __id = web_backend.create_share(share_name, size, inode32)
    except Exception, e:
        return createExpResp('/shares POST failed: %s' % str(e))
    return jsonify({'name': share_name, 'task_id': __id})


@lsa.route('/shares/<string:name>', methods=['PUT'])
@auto.doc(groups=['private', 'internal'])
@login_required
def modify_share(name):
    '''
    Modify an existing share.
    expected argument within the request from are:
      'size': size of the share in bytes
    return the modified share name as {name:<name>}
    raise:
         KeyError: missing parameter
         StandardError: error raised by backend
    '''
    # retrieve parameter
    request_json = request.get_json()
    size = request_json.get('size')
    if not size:
        raise KeyError('missing size parameter')
    # size in bytes must be multiple of 512, otherwise lvm raises error
    # size in bytes should be multiple of 4MB, as lvcreate rounds up to it anyway
    if size % LV_DEFAULT_PHYSICAL_EXTENT_SIZE:
        raise ValueError('Size of the share in bytes must be multiple of %d' % LV_DEFAULT_PHYSICAL_EXTENT_SIZE)
    try:
        web_backend.resize_share(name, size)
    except Exception, e:
        return createExpResp('/shares/%s PUT failed: %s' % (name, str(e)))
    return jsonify(name=name)


@lsa.route('/shares/<string:shareName>/tasks')
@auto.doc(groups=['internal'])
@login_required
def get_share_tasks_list(shareName):
    """
    Gets task list targeted to a specific share
    url param:
     shareName : the name of the share
    param:
      taskName : task name (type) if specified , only task of that type will be returned
    returns:
       {'tasks': [task JSON object]}  (see get_task())
    """
    current_app.logger.debug('Inside get_tasks_list_for_share, name=%s' % shareName)
    current_app.logger.debug('    %s' % str(request.args))
    t_name = request.args.get('taskName', None)
    current_app.logger.debug('Inside get_tasks_list_for_share, name=%s, task name=%s' % (shareName, t_name))
    try:
        result = web_backend.list_background_tasks(target_name=shareName, task_name=t_name)
        jsoned_list = [json.dumps(jt, cls=LSAExtJSONEncoder) for jt in result['tasks']]
        return jsonify({'tasks': jsoned_list})
    except StandardError, e:
        current_app.logger.exception('/shares/%s/tasks' % shareName)
        return createExpResp('/shares/%s/tasks failed: %s' % (shareName, str(e)))


@lsa.route('/shares/<string:name>', methods=['DELETE'])
@auto.doc(groups=['private', 'internal'])
@login_required
def delete_share(name):
    """
    Delete share by name
    Parameter:
       name : the name of the share
    """
    current_app.logger.debug('Inside delete_share, name=%s' % (name))

    try:
        web_backend.delete_share(name)
    except StandardError, e:
        return createExpResp('/shares/%s DELETE failed: %s' % (name, str(e)))

    return ('', 201)


@lsa.route('/shares/<string:name>/exports/<path:export_id>', methods=['DELETE'])
@auto.doc(groups=['private', 'internal'])
@login_required
def delete_export(name, export_id):
    """
    Delete an export by export ID
    Parameter:
       name : the name of the share
       export_id:  export identifier, for nfs it's hosts, for samba it's the resource name
    """
    current_app.logger.debug('Inside delete_export, name=%s, export id =\"%s\"' % (name, export_id))
    current_app.logger.debug('Inside delete_export, request_args=%s' % request.args)
    request_json = request.get_json()
    e_type = request_json.get('type', None)
    if not e_type:
        return createExpResp('export type argument not found')
    try:
        if e_type == 'nfs':
            web_backend.delete_nfs_export(name, export_id)
        elif e_type == 'smb':
            web_backend.delete_samba_export(name, export_id)
    except StandardError, e:
        return createExpResp('/shares/%s/exports/ type: %s DELETE failed: %s' % (name, e_type, str(e)))

    return ('', 201)


@lsa.route('/admin/ssh/keys', methods=['POST'])
@auto.doc(groups=['private', 'internal'])
@login_required
def update_ssh_keys():
    '''
    Add a public ssh key to the server
    form_data:
        sshkey: public ssh key
    '''

    openssh_pubkeys = request.form.get('sshkey')

    try:
        web_backend.update_ssh_keys(openssh_pubkeys)
    except StandardError, e:
        return createExpResp('Public key addition failed: %s' % str(e))

    return ('', 201)


@lsa.route('/admin/ssh/keys', methods=['GET'])
@auto.doc(groups=['private', 'internal'])
@login_required
def get_ssh_keys():
    '''
    Add a public ssh key to the server
    form_data:
        sshkey: public ssh key
    '''
    try:
        keys = web_backend.get_ssh_keys()
    except StandardError, e:
        return createExpResp('/admin/ssh/keys failed: %s' % str(e))
    return jsonify({'ssh_keys': keys})


@lsa.route('/admin/samba/options', methods=['POST'])
@auto.doc(groups=['private', 'internal'])
@login_required
def set_samba_options():
    '''
    set samba global options to the server
    form_data:
        options: dictionary of options. supported options are 'workgroup', 'netbios name', 'interfaces', 'ad_enabled', 'ip_address', 'user_name', 'user_password'
    example:
        options['workgroup']: 'workgroup_name'
        options['netbios name']: 'netbios_name'
        options['interfaces']: 'lo, eno1'
        options['ad_enabled']: True False
        options['fqdn']: host.domain.com
        options['ip_address']: x.x.x.x
        options['user_name']: 'user_name'
        options['user_password']: '********'
    '''
    options = request.form.get('options')

    try:
        web_backend.set_samba_options(options)
    except StandardError, e:
        return createExpResp('set samba options failed: %s' % str(e))

    return ('', 201)


@lsa.route('/admin/samba/options', methods=['GET'])
@auto.doc(groups=['private', 'internal'])
@login_required
def get_samba_options():
    '''
    get samba global settings
    '''
    try:
        options = web_backend.get_samba_options()
    except Exception, e:
        return createExpResp('/admin/samba/options failed: %s' % str(e))
    return jsonify(options)


@lsa.route('/platform/instance', methods=['GET'])
@auto.doc(groups=['private', 'internal'])
@login_required
def get_instance_details():
    '''
    get oci instance details
    Output JSON format:
        {'region': location of the instance(string)
         'domain': availability domain(string)
         'shape': type of shape(string)
         'instance_name': instance/display name(string)
         'creation_time': instance creation time as string using local time representation (ex: 19:00:01 08/14/18 CEST)
        }
    '''
    try:
        inst_dtls = web_backend.get_instance_details()
    except Exception, e:
        return createExpResp('/platform/instance failed: %s' % str(e))
    return jsonify(inst_dtls)


@lsa.route('/admin/ksplice/state', methods=['GET'])
@auto.doc(groups=['private', 'internal'])
@login_required
def get_ksplice_autoupdate_state():
    '''
    Get ksplice autoupdate state.
    format: {'conf_key': value(yes/no)}
    '''
    try:
        state = web_backend.get_ksplice_autoinstall_state()
    except Exception, e:
        return createExpResp('/admin/ksplice/state: %s' % str(e))
    return jsonify({'ksplice_enable_state': state})


@lsa.route('/platform', methods=['GET'])
@auto.doc(groups=['private', 'internal'])
def get_platform_info():
    '''
    Get platform information as dictionary.
    Output:
        format: {version': platform version(string), '': configuration state(CONFIGURED|UNCONFIGURED),
                 status: True if we are in good shape otherwise False,
                 shape: OCI shape( eg. BM.HighIO1.36)
                }
    '''
    try:
        version = web_backend.get_platform_version()
        state = web_backend.get_appliance_configuration_state()
        shape = web_backend.get_oci_shape()
        # TODO For now we are assigning hard coded status i.e default is True.
        # We will add more checkings in future
        status = True
    except Exception, e:
        return createExpResp('/platform GET: %s' % str(e))
    return jsonify({'version': version, 'status': status, 'state': state, 'shape': shape})


@lsa.route('/admin/ksplice/setting', methods=['POST'])
@auto.doc(groups=['private', 'internal'])
@login_required
def toggle_ksplice_autoupdate_setting():
    '''
    Toggle ksplice autoupdate setting
    form_data:
        cur_state: current state of autoupdate(yes or no)
    '''
    cur_state = request.form.get('cur_state')
    if not cur_state:
        raise KeyError('Missing cur_state parameter')

    try:
        new_state = web_backend.toggle_ksplice_autoupdate_setting(cur_state)
    except StandardError, e:
        return createExpResp('Updating ksplice setting failed: %s' % str(e))
    return jsonify({'ksplice_enable_state': new_state})


@lsa.route('/shares/<string:name>/exports', methods=['POST'])
@auto.doc(groups=['private', 'internal'])
@login_required
def create_export(name):
    """
    Creates new export to the server
    Returns the new share identifier  host for NFS and resource name for samba
    Parameter:
    form_data for NFS share :
        'type' : type of the export must be 'nfs'
        'name' : the name to be used for the new share
        'hosts' : authorized client for the new share
        'options' : option for the new share
    form_data for SMB share:
        'type': type of the export(must be 'smb')
        'name': name of the share
        'resource': resource name
        'options': options for samba exports
    """
    request_json = request.get_json()
    if request_json is None:
        return createExpResp('missing request parameters i.e. can not be None')

    e_type = request_json.get('type')
    if e_type is None:
        return createExpResp('Type argument not found')

    try:
        if e_type == 'nfs':
            return _create_nfs_export(name, request_json)
        elif e_type == 'smb':
            return _create_samba_export(name, request_json)
        else:
            # this is unusual.
            return createExpResp('Invalid export type')
    except StandardError, e:
        current_app.logger.exception('export creation failed')
        return createExpResp('Share(%s) export(%s) creation failed: %s' % (name, e_type, str(e)))


def _create_nfs_export(name, request_json):

    """
    Creates a new NFS export.
    name: name of the share
    expected properties within the request JSON are:
        'name': The name for the new nfs share
        'hosts': authorized client for the new share
        'options': the share options as expected by NFS( dictionary item
                    example {sync: true, hide: false, root_squash: true} )
    raise:
         KeyError: missing parameter
         StandardError: error raised by backend
    """

    hosts = request_json.get('hosts')

    if not hosts:
        raise KeyError('missing hosts parameter')

    options = request_json.get('options')

    if not options:
        raise KeyError('missing options parameter')

    current_app.logger.debug('calling backend for NFS export creation')

    # exceptions are caught in create nfs export
    res = web_backend.add_nfs_export(name, hosts, options)
    return jsonify(id=res.get('id'), path=res.get('path'), hosts=res.get('hosts'), options=res.get('options'))


def _create_samba_export(name, request_json):
    '''
    Create a samba export on the system
    parameter:
        request_json: {resource: "resource name",
                            options: {"read only": "boolean", "browsable": "boolean", "guest_ok": "boolean",
                                      "strict_sync": "", "sync_always": "boolean",
                                       "comment": "blah blah", "hosts_allow": "x.x.x.x", "hosts_deny": "x.x.x.x"
                                     }
                       }
                       mandatory keys: read only, browsable, guest_ok, comment
                       *** If guest_ok is true then hosts_allow and hosts_deny should be empty
                           hosts_allow and hosts_deny might have comma separated IP addresses or
                           network addresses
        name: name of the share as string
    raise:
        KeyError: missing parameter
        StandardError: error raised by backend
    return: share name
    '''
    resource = request_json.get('resource')
    if resource is None:
        raise KeyError('missing resource parameter')
    options = request_json.get('options')
    if options is None:
        raise KeyError('missing options parameter')
    res = web_backend.create_samba_export(name, resource, options)
    return jsonify(id=res.get('id'))


@lsa.route('/shares/<string:name>/exports/<path:export_id>', methods=['PUT'])
@auto.doc(groups=['private', 'internal'])
@login_required
def modify_export(name, export_id):
    '''
    Updates an existing  nfs export/samba export
    Returns the export ID(host for NFS and resource name for Samba)
    export_id:  export identifier, for nfs it's hosts, for samba it's the resource name
    expected argument within the request JSON data:
        'name': The name for the new nfs share
        'export_id' : unique identifier of the export
        'type' : 'nfs' | 'smb'
        'options': export options or attributes
    raise:
         KeyError: missing parameter
         StandardError: error raised by backend
    '''
    request_json = request.get_json()
    e_type = request_json.get('type')
    if e_type is None:
        return createExpResp('export type argument not found')
    try:
        if e_type == 'nfs':
            if request_json.get('options') is None:
                return createExpResp('options argument not found')
            return _modify_nfs_export(name, export_id, request_json.get('options'))
        elif e_type == 'smb':
            return _modify_samba_export(name, export_id, request_json)
        else:
            # this is unusual.
            return createExpResp('Invalid export type')
    except StandardError, e:
        return createExpResp('Share(%s) export (%s) modification failed: %s' % (name, e_type, str(e)))


def _modify_nfs_export(name, hosts, options):
    '''
    Updates an existing NFS export.
    parameter:
        'name': name of the share
        hosts: authorized clients for the nfs export, IP/network address
        'options': the share options as expected by NFS( dictionary item
                    example {sync: true, hide: false, root_squash: true} )
     raise:
         KeyError: missing parameter
         StandardError: error raised by backend

    '''

    current_app.logger.debug('calling backend for NFS share creation')

    # exceptions are caught in create_share
    res = web_backend.modify_nfs_export(name, hosts, options)
    return jsonify(id=res.get('id'))


def _modify_samba_export(name, resource, request_json):
    '''
    Modify a samba export on the system
    parameter:
        name: name of the share
        resource: resource name of the samba export(ID)
        request_json: {options: {"read only": "boolean", "browsable": "boolean", "guest_ok": "boolean",
                                 "strict_sync": "", "sync_always": "boolean",
                                 "comment": "blah blah", "hosts_allow: "x.x.x.x", "hosts_deny": "x.x.x.x"
                                }
                       }
                       mandatory keys: read only, browsable, guest_ok, comment
                       *** If guest_ok is true then hosts_allow and hosts_deny should be empty
                           hosts_allow and hosts_deny might have comma separated IP addresses or
                           network addresses
       raise:
        KeyError: missing parameter
        StandardError: error raised by backend
    '''
    options = request_json.get('options')
    if options is None:
        raise KeyError('missing options parameter')
    res = web_backend.modify_samba_export(name, resource, options)
    return jsonify(id=res.get('id'), resource=res.get('resource'))


@lsa.route('/log/<string:type>', methods=['GET'])
@auto.doc(groups=['private', 'internal'])
@login_required
def logs(type):
    """
    Gets the logs related to given category
    parameter:
      type : the type of logs. can be 'system', 'boot', 'services', 'storage'
      keyword: regExp used to filter the content of logs
    Returns
      {
        'entries':<logs a multi line text>
      }
    """
    current_app.logger.debug('calling backend for logs')
    try:
        keyword = request.args.get('keyword', None)
        res = web_backend.get_logs(type, keyword)
        return jsonify({'entries': res})
    except StandardError, e:
        return createExpResp('/log/%s failed: %s' % (type, str(e)))


@lsa.route('/services')
@auto.doc(groups=['private', 'internal'])
@login_required
def getServices():
    """
    returns the list of services
    As an appliance context. only enabled services are returned
    returned list limited to appliance related services. I.e: nfs-idmapd,nfs-mountd,nfs-server,httpd
    returns
     [
       {
        'name'   : <systemd service name>,
        'desc'   : <systemd service description>,
        'status' : <systemd service status as reported by systemctl(1)>,
        'logs'   : <systemd service logs as string array>
       },
       ...
     ]

    example:
     [
      {
        'name': 'httpd',
        'desc' : 'The Apache HTTP Server',
        'status' : 'running',
        'logs' : [
                   'Starting The Apache HTTP Server...',
                   'httpd.service: main process exited, code=exited, status=1/RUNNING'
                   ...
                 ]
      }
     ]

    """
    current_app.logger.debug('calling backend for service list')
    try:
        services = web_backend.getServices()
        if current_app.logger.isEnabledFor(logging.DEBUG):
            for service in services:
                current_app.logger.debug(service)

        return jsonify({'list': services})
    except Exception, e:
        return createExpResp('/pool/services failed: %s' % str(e))


@lsa.route('/shares/<string:shareName>/restore', methods=['POST'])
@auto.doc(groups=['internal'])
@login_required
def restore_share_as_object(shareName):
    """
    restore a share from OCI object store
    expected parameters:
        shareName: name of the share to be backed up
    expected form parameter:
        namespace-name : namespace name of the remote object store service
        bucket-name  : bucket name of the remote bucket
        object-name : backup archive name to object store
    returns:
        task id of the spawned background task doing the restore

    remote path is 'backups/<shareName>/<archive name>'

    see backend_test:restore_share_backup() for details

    """
    current_app.logger.debug('Inside restore_share_as_object, name=%s' % shareName)
    namespace_name = request.form.get('namespace-name')
    if not namespace_name:
        return createExpResp('missing namespace-name parameter')
    obj_name = request.form.get('object-name')
    if not obj_name:
        return createExpResp('missing object-name parameter')
    bucket_name = request.form.get('bucket-name')
    if not bucket_name:
        return createExpResp('missing bucket_name parameter')
    result = None

    try:
        result = web_backend.restore_share_backup(shareName, namespace_name, bucket_name, obj_name)
    except StandardError, e:
        return createExpResp('/shares/%s/restore failed: %s' % (shareName, str(e)))

    return jsonify(taskId=result['taskId'])


@lsa.route('/shares/<string:shareName>/backup', methods=['POST'])
@auto.doc(groups=['internal'])
@login_required
def backup_share_as_object(shareName):
    """
    Backup a share to remote object store
    expected parameters:
        shareName: name of the share to be backed up
    expected form parameters
        namespace-name : namespace name of the remote object store service
        bucket-name  : bucket name of the remote bucket on object store service
        object-name : backup archive name to object store
    returns:
        task id of the spawned background task doing the backup

    remote path is 'backups/<shareName>/<archive name>'

    see backend_test:backup_share_on_object_store() for details
    """

    current_app.logger.debug('Inside backup_share_as_object, name=%s' % shareName)

    namespace_name = request.form.get('namespace-name')
    if not namespace_name:
        return createExpResp('missing namespace-name parameter')
    obj_name = request.form.get('object-name')
    if not obj_name:
        return createExpResp('missing object-name parameter')
    bucket_name = request.form.get('bucket-name')
    if not bucket_name:
        return createExpResp('missing bucket_name parameter')
    result = None
    try:
        result = web_backend.backup_share_on_object_store(shareName,
                                                          namespace_name,
                                                          bucket_name,
                                                          obj_name)
    except StandardError, e:
        return createExpResp('/shares/%s/backup failed: %s' % (shareName, str(e)))

    return jsonify(taskId=result['taskId'])


@lsa.route('/admin/reboot')
@auto.doc(groups=['private', 'internal'])
@login_required
def reboot():
    ''' reboots the system '''
    current_app.logger.debug('calling backend for reboot')
    try:
        web_backend.system_reboot()
    except Exception, e:
        return createExpResp('/admin/reboot failed: %s' % str(e))

    return ('', 201)


@lsa.route('/admin/update')
@auto.doc(groups=['private', 'internal'])
@login_required
def update():
    ''' updates the system '''
    current_app.logger.debug('calling backend for update')
    try:
        res = web_backend.system_update()
    except Exception, e:
        return createExpResp('/admin/update failed: %s' % str(e))

    return jsonify(res)


@lsa.route('/admin/nfs/restart')
@auto.doc(groups=['private', 'internal'])
@login_required
def restart_nfs_stack():
    ''' restart all NFS shares required services '''

    current_app.logger.debug('calling backend for nfs stack')
    try:
        web_backend.restartNFSStack()
    except StandardError, e:
        return createExpResp('/admin/nfs/restart failed: %s' % str(e))

    return ('', 201)


@lsa.route('/admin/smb/restart')
@auto.doc(groups=['private', 'internal'])
@login_required
def restart_smb_services():
    ''' restart all SMB related services '''

    current_app.logger.debug('calling backend for smb stack')
    try:
        web_backend.restartSambaServices()
    except StandardError, e:
        return createExpResp('/admin/smb/restart failed: %s' % str(e))

    return ('', 201)


@lsa.route('/username')
@auto.doc(groups=['internal'])
@login_required
def username():
    ''' return the username of the currently logged in user '''
    return jsonify(name=current_user.name)


@lsa.route('/platform/status')
@auto.doc(groups=['internal'])
@login_required
def platform_status():
    """
    Checks platform status, in terms of whether the OCI shape is supported
    according to a white list, such as
        ['BM.HighIO1.36',
         'BM.DenseIO1.36',
         'VM.DenseIO1.4',
         'VM.DenseIO1.8',
         'VM.DenseIO1.16',
         'BM.DenseIO2.52',
         'VM.DenseIO2.8',
         'VM.DenseIO2.16',
         'VM.DenseIO2.24']
    """
    try:
        # #TODO delete this API. cleanup, need to clean up front end calls and management codes
        # we show red banner if the shape is not supported on the UI, since we are supporting all the shapes, UI codes need
        # to be re-factored.
        # This API is not needed anymore as we support all platforms
        # see webapp/static/js/models/platform.js  line 108
        return jsonify(web_backend.get_platform_status())
    except StandardError, e:
        return createExpResp('/platform/status failed: %s' % str(e))


@lsa.route('/platform/state', methods=['GET'])
@auto.doc(groups=['internal'])
@login_required
def platform_get_state():
    """
    Returns the current platform state
    The state is retrieved from helper script state.sh
    Returns
       {'state':<state number>}
    """
    try:
        state_code = subprocess.call(LSA_STATE_BIN, shell=True)
        current_app.logger.debug('current lsa state : %s' % state_code)
        return jsonify({'state': state_code})
    except StandardError, e:
        return createExpResp('/platform/state failed: %s' % str(e))


@lsa.route('/platform/unconfigure', methods=['POST'])
@auto.doc(groups=['internal'])
@login_required
def unconfigure_platform():
    '''
    Unconfigure the current platform.
    Following actions are taken:
       - stop and disable NFS server
       - stop and disable samba server
       - stop and disable http server
       - revert appliance state
    '''
    current_app.logger.debug('unconfigure_platform called')
    try:
        web_backend.unconfigure()
    except StandardError, e:
        return createExpResp('/platform/unconfigure failed: %s' % str(e))
    return ('', 201)


@lsa.route('/platform/disks', methods=['GET'])
@auto.doc(groups=['internal'])
@login_required
def platform_list_disks():
    '''
    Input:
        none
    Return:
        a JSON object as follows,
        {
          disks:[
                  {
                  name: name of the disk as string
                  id: /dev/disk/by-id entry of /dev/name
                  type: type of the disk as string, ['iscsi' | 'fc' | 'sas' | 'nvmf' |...]
                  wwn: world wide name as string
                  size: size of disk in bytes as int
                  state: state of disk as string, ['running' | 'not running']
                  status: status of disk as dict
                    {
                        partitioned: status as boolean of whether the disk contains partitions
                        mounted: status as boolean of whether the disk or at least one of the partitions(if any) is mounted
                    }
                  }
                ]
        }
        eg.
        [
            {
            'name': sdb
            'id': scsi-3605d0faaf5b04f249b65dfc3781b8d3f
            'type': 'iscsi'
            'wwn: 'iqn.2015-12.com.ora...'
            'size': 107374182400
            'state': 'running'
            'status'
                {
                    partitioned: True
                    mounted: False
                }
            },
            {
            'name': sdc
            ...
            }
        ]
    '''
    try:
        return jsonify({'disks': web_backend.get_platform_disks()})
    except StandardError, e:
        return createExpResp('/platform/disks failed: %s' % str(e))


@lsa.route('/task/<int:task_id>')
@auto.doc(groups=['internal'])
@login_required
def get_task(task_id):
    '''
    Returns the task by task's id
    Returns a JSON object as follow
    {
       'ID' : unique ID as int
       'name': task name as string
       'target': the subject of the task (usually share name)
       'status' : {
                   'code': status code (see backend.TaskStatusCode),
                   'message': status message
                  },
       'states' : [{
                   'type': state type (see backend.TaskState),
                   'message': state message ,
                   'time': timestamp
                  },...]
      }
    '''
    current_app.logger.debug('get_task called')
    try:
        result = web_backend.get_background_task(task_id)
        return jsonify({'task': json.dumps(result['task'], cls=LSAExtJSONEncoder)})
    except StandardError, e:
        current_app.logger.exception('/tasks failed')
        return createExpResp('/tasks failed: %s' % str(e))


@lsa.route('/tasks')
@auto.doc(groups=['internal'])
@login_required
def list_tasks():
    '''
    Returns the list of ongoing tasks
    Returns list of JSON object as follow
    [
      {
       'ID' : unique ID as int
       'name': task name as string
       'target': the subject of the task (usually share name)
       'status' : {
                   'code': status code (see backend.TaskStatusCode),
                   'message': status message
                  },
       'states' : [{
                   'type': state type (see backend.TaskState),
                   'message': state message ,
                   'time': timestamp
                  },...]
      },...
    ]
    request parameter:
       running: boolean , return only running tasks ?
    '''
    # current_app.logger.debug('list_tasks called')
    running_tasks_only = request.args.get('running', type=bool)
    # current_app.logger.debug('running only ? %s' % running_tasks_only)
    try:
        if running_tasks_only:
            result = web_backend.list_background_tasks(running=running_tasks_only)
        else:
            result = web_backend.list_background_tasks()
        jsoned_list = [json.dumps(jt, cls=LSAExtJSONEncoder) for jt in result['tasks']]
        return jsonify({'tasks': jsoned_list})
    except StandardError, e:
        current_app.logger.exception('/tasks failed')
        return createExpResp('/tasks failed: %s' % str(e))


@lsa.route('/admin/oci/config', methods=['DELETE'])
@auto.doc(groups=['internal'])
@login_required
def delete_oci_service_config():
    """
    Deletes configuration for OCI service access
    """
    current_app.logger.debug('Inside delete_oci_service_config')
    try:
        web_backend.delete_oci_config()
    except StandardError, e:
        return createExpResp('/admin/oci/config failed: %s' % str(e))
    return ('', 201)


@lsa.route('/admin/oci/config', methods=['GET'])
@auto.doc(groups=['internal'])
@login_required
def get_oci_service_config():
    """
    Gets configuration for OCI service access
    Returns a dict with key/value as described
    in OCI service configuration documentation
    """
    try:
        return jsonify(web_backend.get_oci_config())
    except StandardError, e:
        return createExpResp('/admin/oci/config failed: %s' % str(e))


@lsa.route('/admin/oci/config', methods=['POST'])
@auto.doc(groups=['internal'])
@login_required
def set_oci_service_config():
    """
    Sets configuration for OCI service access
    form data:
        user: user name used for OCI configuration
        tenancy: user tenancy
        region: tenancy region
        passphrase: passphrase of private key file (optional)
        key file : stored as 'keyfile' in request.files
    """
    oci_user = request.form.get('user')
    if not oci_user:
        return createExpResp('missing user parameter')
    oci_tenancy = request.form.get('tenancy')
    if not oci_tenancy:
        return createExpResp('missing tenancy parameter')
    oci_region = request.form.get('region')
    if not oci_region:
        return createExpResp('missing region parameter')

    # these are optional
    oci_keyfile = request.files['keyfile']
    oci_keyfile_passphrase = request.form.get('passphrase')

    try:
        web_backend.set_oci_config(oci_user,
                                   oci_tenancy,
                                   oci_region,
                                   oci_keyfile,
                                   oci_keyfile_passphrase)
    except StandardError, e:
        return createExpResp('/admin/oci/config failed: %s' % str(e))

    return ('', 201)


@lsa.route('/platform/localtime', methods=['GET'])
@auto.doc(groups=['internal'])
@login_required
def platform_local_time():
    '''
    Gets the platform local time.
    Input:
        none
    Return: returns 'ltime' set as  local time expressed like
      11:57:29 07/20/18 CEST
    '''
    return jsonify({'ltime': time.strftime('%X %x %Z')})


@lsa.route('/')
@login_required
def index():
    return render_template('index.html')
